//
// C++ Interface: engine
//
// Description: Asynchronous Database Query Engine
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef ENGINE_H
#define ENGINE_H

#include "sysdeps.h"

#include "fplan.h"
#ifdef HAVE_PILOTLINK
#include "palm.h"
#endif
#include "dbobj.h"

class Preferences;

class Preferences {
        private:
                template <typename T, typename LT = T> class PreferencesValueInfo {
                        public:
                                typedef T datatype_t;
                                typedef LT loadtype_t;
                                PreferencesValueInfo(const std::string& n, const T& dflt) : m_name(n), m_default(dflt) {}
                                const std::string& get_name(void) const { return m_name; }
                                const T& get_default(void) const { return m_default; }

                        private:
                                std::string m_name;
                                datatype_t m_default;
                };

                template <typename T> class PreferencesValue {
                        public:
                                typedef typename T::datatype_t datatype_t;

                                PreferencesValue(const T& info) : m_info(info), m_value(info.get_default()), m_dirty(false) {}
                                void operator=(datatype_t v);
                                operator datatype_t() { return m_value; }
                                sigc::signal<void,datatype_t>& signal_change(void) { return m_change; }
                                const sigc::signal<void,datatype_t>& signal_change(void) const { return m_change; }
                                const std::string& name(void) { return m_info.get_name(); }

                        private:
                                typedef typename T::loadtype_t loadtype_t;
                                typedef T infotype_t;

                                const infotype_t& m_info;
                                datatype_t m_value;
                                sigc::signal<void,datatype_t> m_change;
                                bool m_dirty;

                                bool is_dirty(void) const { return m_dirty; }
                                void clear_dirty(void) { m_dirty = false; }

                                friend class Preferences;
                };

                static const PreferencesValueInfo<FPlanRoute::id_t, long long> curplan_info;
                static const PreferencesValueInfo<double> vcruise_info;
                static const PreferencesValueInfo<double> windspeed_info;
                static const PreferencesValueInfo<double> winddir_info;
                static const PreferencesValueInfo<double> qnh_info;
                static const PreferencesValueInfo<double> temp_info;
                static const PreferencesValueInfo<bool, long long> gps_info;
                static const PreferencesValueInfo<bool, long long> autowpt_info;
                static const PreferencesValueInfo<double> vblock_info;
                static const PreferencesValueInfo<int, long long> maxalt_info;
                static const PreferencesValueInfo<int, long long> mapflags_info;
                static const PreferencesValueInfo<double> mapscale_info;
                static const PreferencesValueInfo<double> mapscaleairportdiagram_info;
                static const PreferencesValueInfo<Glib::ustring> globaldbdir_info;
                static const PreferencesValueInfo<Glib::ustring> gpshost_info;
                static const PreferencesValueInfo<Glib::ustring> gpsport_info;

                static const std::string curplan_name;
                static const std::string vcruise_name;
                static const std::string windspeed_name;
                static const std::string winddir_name;
                static const std::string qnh_name;
                static const std::string temp_name;
                static const std::string gps_name;
                static const std::string autowpt_name;
                static const std::string vblock_name;
                static const std::string maxalt_name;
                static const std::string mapflags_name;
                static const std::string mapscale_name;
                static const std::string mapscaleairportdiagram_name;
                static const std::string globaldbdir_name;
                static const std::string gpshost_name;
                static const std::string gpsport_name;

                static const FPlanRoute::id_t curplan_dflt;
                static const double vcruise_dflt;
                static const double windspeed_dflt;
                static const double winddir_dflt;
                static const double qnh_dflt;
                static const double temp_dflt;
                static const bool gps_dflt;
                static const bool autowpt_dflt;
                static const double vblock_dflt;
                static const int maxalt_dflt;
                static const int mapflags_dflt;
                static const double mapscale_dflt;
                static const double mapscaleairportdiagram_dflt;
                static const Glib::ustring globaldbdir_dflt;
                static const Glib::ustring gpshost_dflt;
                static const Glib::ustring gpsport_dflt;

        public:
                Preferences(void);
                ~Preferences(void);
                void commit(void);

                PreferencesValue<PreferencesValueInfo<FPlanRoute::id_t, long long> > curplan;
                PreferencesValue<PreferencesValueInfo<double> > vcruise;
                PreferencesValue<PreferencesValueInfo<double> > windspeed;
                PreferencesValue<PreferencesValueInfo<double> > winddir;
                PreferencesValue<PreferencesValueInfo<double> > qnh;
                PreferencesValue<PreferencesValueInfo<double> > temp;
                PreferencesValue<PreferencesValueInfo<bool, long long> > gps;
                PreferencesValue<PreferencesValueInfo<bool, long long> > autowpt;
                PreferencesValue<PreferencesValueInfo<double> > vblock;
                PreferencesValue<PreferencesValueInfo<int, long long> > maxalt;
                PreferencesValue<PreferencesValueInfo<int, long long> > mapflags;
                PreferencesValue<PreferencesValueInfo<double> > mapscale;
                PreferencesValue<PreferencesValueInfo<double> > mapscaleairportdiagram;
                PreferencesValue<PreferencesValueInfo<Glib::ustring> > globaldbdir;
                PreferencesValue<PreferencesValueInfo<Glib::ustring> > gpshost;
                PreferencesValue<PreferencesValueInfo<Glib::ustring> > gpsport;

        private:
                sqlite3x::sqlite3_connection db;

                static const std::string insert_command;
                static const std::string select_command;

                void save(const std::string& name, long long v);
                void save(const std::string& name, const std::string& v);
                //void save(const std::string& name, const Glib::ustring& v) { save(name, (std::string)v); }
                void save(const std::string& name, double v);
                void load(const std::string& name, long long& v);
                void load(const std::string& name, std::string& v);
                void load(const std::string& name, Glib::ustring& v) { std::string v1; load(name, v1); v = v1; }
                void load(const std::string& name, double& v);
                template<class T> void load(T& t) { typename T::loadtype_t x(t); load(t.name(), x); t = x; t.clear_dirty(); }
                template<class T> void save(T& t) { if (t.is_dirty()) save(t.name(), (typename T::loadtype_t)t); }
};

class Engine {
        private:
                class DbThread {
                        private:
                                class ResultBase {
                                        public:
                                                ResultBase(const sigc::slot<void>& dbaction, const sigc::slot<void>& dbcancel = sigc::slot<void>());
                                                ~ResultBase();

                                                bool is_done(void) const { return m_done; }
                                                bool is_error(void) const { return m_error; }
                                                sigc::connection connect(const sigc::signal<void>::slot_type& slot);
                                                void disconnect_all(void);
                                                void cancel(void);
                                                template<class T> static void cancel(Glib::RefPtr<T>& q);

                                                void reference(void);
                                                void unreference(void);

                                                void execute(void);
                                                void seterror(void);

                                        protected:
                                                Glib::Mutex m_mutex;
                                                sigc::slot<void> m_dbcancel;
                                                sigc::slot<void> m_dbaction;
                                                sigc::signal<void> m_donecb;
                                                int m_refcount;
                                                bool m_done;
                                                bool m_error;
                                                bool m_executing;
                                };

                        public:
                                template <class Db> class Result : public ResultBase {
                                        public:
                                                typedef typename Db::element_t element_t;
                                                typedef typename Db::elementvector_t elementvector_t;

                                                Result(const sigc::slot<elementvector_t>& dbaction, const sigc::slot<void>& dbcancel = sigc::slot<void>());
                                                ~Result();
                                                const elementvector_t& get_result(void) const { return m_result; }
                                                elementvector_t& get_result(void) { return m_result; }

                                                void set_result(const elementvector_t& ev) { m_result = ev; }

                                        protected:
                                                elementvector_t m_result;
                                };

                                typedef Result<AirportsDb> AirportResult;
                                typedef Result<AirspacesDb> AirspaceResult;
                                typedef Result<NavaidsDb> NavaidResult;
                                typedef Result<WaypointsDb> WaypointResult;
                                typedef Result<AirwaysDb> AirwayResult;
                                typedef Result<MapelementsDb> MapelementResult;

                                class ElevationResult : public ResultBase {
                                        public:
                                                typedef TopoDb30::elev_t elev_t;

                                                ElevationResult(const Point& pt, TopoDb30& topodb);
#ifdef HAVE_PILOTLINK
                                                ElevationResult(const Point& pt, PalmGtopo& topodb);
#endif
                                                ~ElevationResult();
                                                elev_t get_result(void) const { return m_result; }

                                        protected:
                                                elev_t m_result;

                                                void execute(TopoDb30& topodb, const Point& pt);
#ifdef HAVE_PILOTLINK
                                                void execute_palm(PalmGtopo& topodb, const Point& pt);
#endif
                                };

                                class ElevationMinMaxResult : public ResultBase {
                                        public:
                                                typedef TopoDb30::elev_t elev_t;
                                                typedef TopoDb30::minmax_elev_t minmax_elev_t;

                                                ElevationMinMaxResult(const Rect& r, TopoDb30& topodb);
                                                ElevationMinMaxResult(const Polygon& p, TopoDb30& topodb);
#ifdef HAVE_PILOTLINK
                                                ElevationMinMaxResult(const Rect& r, PalmGtopo& topodb);
                                                ElevationMinMaxResult(const Polygon& p, PalmGtopo& topodb);
#endif
                                                ~ElevationMinMaxResult();
                                                minmax_elev_t get_result(void) const { return m_result; }

                                        protected:
                                                minmax_elev_t m_result;

                                                void execute_rect(TopoDb30& topodb, const Rect& r);
                                                void execute_poly(TopoDb30& topodb, const Polygon& p);
#ifdef HAVE_PILOTLINK
                                                void execute_rect_palm(PalmGtopo& topodb, const Rect& r);
                                                void execute_poly_palm(PalmGtopo& topodb, const Polygon& p);
#endif
                                };

                                class ElevationMapResult : public ResultBase {
                                        public:
                                                ElevationMapResult(const Rect& bbox, TopoDb30& topodb);
#ifdef HAVE_PILOTLINK
                                                ElevationMapResult(const Rect& bbox, PalmGtopo& topodb);
#endif
                                                ~ElevationMapResult();
                                                Rect get_bbox(void) const { return m_bbox; }
                                                unsigned int get_width(void) const { return m_width; }
                                                unsigned int get_height(void) const { return m_height; }
                                                TopoDb30::elev_t operator()(unsigned int x, unsigned int y);
                                                Glib::RefPtr<Gdk::Drawable> get_pixmap(void) const { return m_pixmap; }

                                        protected:
                                                Rect m_bbox;
                                                Glib::RefPtr<Gdk::Drawable> m_pixmap;
                                                unsigned int m_width;
                                                unsigned int m_height;
                                                TopoDb30::elev_t *m_elev;

                                                void execute(TopoDb30& topodb);
#ifdef HAVE_PILOTLINK
                                                void execute_palm(PalmGtopo& topodb);
#endif
                                };

                                class ElevationMapCairoResult : public ResultBase {
                                        public:
                                                ElevationMapCairoResult(const Rect& bbox, TopoDb30& topodb);
#ifdef HAVE_PILOTLINK
                                                ElevationMapCairoResult(const Rect& bbox, PalmGtopo& topodb);
#endif
                                                ~ElevationMapCairoResult();
                                                Rect get_bbox(void) const { return m_bbox; }
                                                Cairo::RefPtr<Cairo::ImageSurface> get_surface(void) const { return m_surface; }

                                        protected:
                                                Rect m_bbox;
                                                Cairo::RefPtr<Cairo::ImageSurface> m_surface;

                                                void execute(TopoDb30& topodb);
#ifdef HAVE_PILOTLINK
                                                void execute_palm(PalmGtopo& topodb);
#endif
                                };

                                typedef ResultBase AirportSaveResult;
                                typedef ResultBase AirspaceSaveResult;
                                typedef ResultBase NavaidSaveResult;
                                typedef ResultBase WaypointSaveResult;
                                typedef ResultBase AirwaySaveResult;
                                typedef ResultBase MapelementSaveResult;

                                DbThread(const std::string& dir_main = "", const std::string& dir_aux = "");
                                ~DbThread();

                                Glib::RefPtr<AirportResult> airport_find_by_icao(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirportsDb::comp_t comp = AirportsDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirportResult> airport_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirportsDb::comp_t comp = AirportsDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirportResult> airport_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirportResult> airport_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirportSaveResult> airport_save(const AirportsDb::Airport& e);

                                Glib::RefPtr<AirspaceResult> airspace_find_by_icao(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirspacesDb::comp_t comp = AirspacesDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirspaceResult> airspace_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirspacesDb::comp_t comp = AirspacesDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirspaceResult> airspace_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirspaceResult> airspace_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirspaceSaveResult> airspace_save(const AirspacesDb::Airspace& e);

                                Glib::RefPtr<NavaidResult> navaid_find_by_icao(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, NavaidsDb::comp_t comp = NavaidsDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<NavaidResult> navaid_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, NavaidsDb::comp_t comp = NavaidsDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<NavaidResult> navaid_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<NavaidResult> navaid_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0);
                                Glib::RefPtr<NavaidSaveResult> navaid_save(const NavaidsDb::Navaid& e);

                                Glib::RefPtr<WaypointResult> waypoint_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, WaypointsDb::comp_t comp = WaypointsDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<WaypointResult> waypoint_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<WaypointResult> waypoint_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0);
                                Glib::RefPtr<WaypointSaveResult> waypoint_save(const WaypointsDb::Waypoint& e);

                                Glib::RefPtr<AirwayResult> airway_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirwaysDb::comp_t comp = AirwaysDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirwayResult> airway_find_by_text(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirwaysDb::comp_t comp = AirwaysDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirwayResult> airway_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirwayResult> airway_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0);
                                Glib::RefPtr<AirwaySaveResult> airway_save(const AirwaysDb::Airway& e);

                                Glib::RefPtr<MapelementResult> mapelement_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, MapelementsDb::comp_t comp = MapelementsDb::comp_contains, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<MapelementResult> mapelement_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0);
                                Glib::RefPtr<MapelementResult> mapelement_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0);
                                Glib::RefPtr<MapelementSaveResult> mapelement_save(const MapelementsDb::Mapelement& e);

                                Glib::RefPtr<ElevationResult> elevation_point(const Point& pt);
                                Glib::RefPtr<ElevationMapResult> elevation_map(const Rect& bbox);
                                Glib::RefPtr<ElevationMapCairoResult> elevation_map_cairo(const Rect& bbox);
                                Glib::RefPtr<ElevationMinMaxResult> elevation_minmax(const Rect& r);
                                Glib::RefPtr<ElevationMinMaxResult> elevation_minmax(const Polygon& p);

                        private:
                                AirportsDb m_airportdb;
                                NavaidsDb m_navaiddb;
                                WaypointsDb m_waypointdb;
                                AirwaysDb m_airwaydb;
                                AirspacesDb m_airspacedb;
                                MapelementsDb m_mapelementdb;
                                TopoDb30 m_topodb;
#ifdef HAVE_PILOTLINK
                                PalmAirports m_palmairportdb;
                                PalmNavaids m_palmnavaiddb;
                                PalmWaypoints m_palmwaypointdb;
                                PalmAirspaces m_palmairspacedb;
                                PalmMapelements m_palmmapelementdb;
                                PalmGtopo m_palmgtopodb;
#endif
                                bool m_terminate;
                                Glib::Mutex m_mutex;
                                Glib::Cond m_cond;
                                Glib::Thread *m_thread;
                                std::list<Glib::RefPtr<ResultBase> > m_tasklist;

                                void thread(void);
                                template<class T> void queue(Glib::RefPtr<T>& p);
                };

        public:
                typedef enum {
                        auxdb_none = 0,
                        auxdb_prefs = 1,
                        auxdb_override = 2
                } auxdb_mode_t;

                Engine(const std::string& dir_main = "", auxdb_mode_t auxdbmode = auxdb_prefs, const std::string& dir_aux = "");

                Preferences& get_prefs(void) { return m_prefs; }
                FPlan& get_fplan_db(void) { return m_fplandb; }
                TracksDb& get_tracks_db(void) { return m_tracksdb; }

        public:
                // Asynchronous interface
                typedef DbThread::AirportResult AirportResult;
                typedef DbThread::AirspaceResult AirspaceResult;
                typedef DbThread::NavaidResult NavaidResult;
                typedef DbThread::WaypointResult WaypointResult;
                typedef DbThread::AirwayResult AirwayResult;
                typedef DbThread::MapelementResult MapelementResult;
                typedef DbThread::AirportSaveResult AirportSaveResult;
                typedef DbThread::AirspaceSaveResult AirspaceSaveResult;
                typedef DbThread::NavaidSaveResult NavaidSaveResult;
                typedef DbThread::WaypointSaveResult WaypointSaveResult;
                typedef DbThread::AirwaySaveResult AirwaySaveResult;
                typedef DbThread::MapelementSaveResult MapelementSaveResult;
                typedef DbThread::ElevationResult ElevationResult;
                typedef DbThread::ElevationMinMaxResult ElevationMinMaxResult;
                typedef DbThread::ElevationMapResult ElevationMapResult;
                typedef DbThread::ElevationMapCairoResult ElevationMapCairoResult;

                Glib::RefPtr<AirportResult> async_airport_find_by_icao(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirportsDb::comp_t comp = AirportsDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.airport_find_by_icao(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<AirportResult> async_airport_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirportsDb::comp_t comp = AirportsDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.airport_find_by_name(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<AirportResult> async_airport_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0) { return m_dbthread.airport_find_bbox(rect, limit, loadsubtables); }
                Glib::RefPtr<AirportResult> async_airport_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0) { return m_dbthread.airport_find_nearest(pt, limit, rect, loadsubtables); }
                Glib::RefPtr<AirportSaveResult> async_airport_save(const AirportsDb::Airport& e) { return m_dbthread.airport_save(e); }

                Glib::RefPtr<AirspaceResult> async_airspace_find_by_icao(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirspacesDb::comp_t comp = AirspacesDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.airspace_find_by_icao(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<AirspaceResult> async_airspace_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirspacesDb::comp_t comp = AirspacesDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.airspace_find_by_name(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<AirspaceResult> async_airspace_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0) { return m_dbthread.airspace_find_bbox(rect, limit, loadsubtables); }
                Glib::RefPtr<AirspaceResult> async_airspace_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0) { return m_dbthread.airspace_find_nearest(pt, limit, rect, loadsubtables); }
                Glib::RefPtr<AirspaceSaveResult> async_airspace_save(const AirspacesDb::Airspace& e) { return m_dbthread.airspace_save(e); }

                Glib::RefPtr<NavaidResult> async_navaid_find_by_icao(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, NavaidsDb::comp_t comp = NavaidsDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.navaid_find_by_icao(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<NavaidResult> async_navaid_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, NavaidsDb::comp_t comp = NavaidsDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.navaid_find_by_name(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<NavaidResult> async_navaid_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0) { return m_dbthread.navaid_find_bbox(rect, limit, loadsubtables); }
                Glib::RefPtr<NavaidResult> async_navaid_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0) { return m_dbthread.navaid_find_nearest(pt, limit, rect, loadsubtables); }
                Glib::RefPtr<NavaidSaveResult> async_navaid_save(const NavaidsDb::Navaid& e) { return m_dbthread.navaid_save(e); }

                Glib::RefPtr<WaypointResult> async_waypoint_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, WaypointsDb::comp_t comp = WaypointsDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.waypoint_find_by_name(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<WaypointResult> async_waypoint_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0) { return m_dbthread.waypoint_find_bbox(rect, limit, loadsubtables); }
                Glib::RefPtr<WaypointResult> async_waypoint_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0) { return m_dbthread.waypoint_find_nearest(pt, limit, rect, loadsubtables); }
                Glib::RefPtr<WaypointSaveResult> async_waypoint_save(const WaypointsDb::Waypoint& e) { return m_dbthread.waypoint_save(e); }

                Glib::RefPtr<AirwayResult> async_airway_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirwaysDb::comp_t comp = AirwaysDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.airway_find_by_name(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<AirwayResult> async_airway_find_by_text(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, AirwaysDb::comp_t comp = AirwaysDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.airway_find_by_text(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<AirwayResult> async_airway_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0) { return m_dbthread.airway_find_bbox(rect, limit, loadsubtables); }
                Glib::RefPtr<AirwayResult> async_airway_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0) { return m_dbthread.airway_find_nearest(pt, limit, rect, loadsubtables); }
                Glib::RefPtr<AirwaySaveResult> async_airway_save(const AirwaysDb::Airway& e) { return m_dbthread.airway_save(e); }

                Glib::RefPtr<MapelementResult> async_mapelement_find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, MapelementsDb::comp_t comp = MapelementsDb::comp_contains, unsigned int loadsubtables = ~0) { return m_dbthread.mapelement_find_by_name(nm, limit, skip, comp, loadsubtables); }
                Glib::RefPtr<MapelementResult> async_mapelement_find_bbox(const Rect& rect, unsigned int limit = ~0, unsigned int loadsubtables = ~0) { return m_dbthread.mapelement_find_bbox(rect, limit, loadsubtables); }
                Glib::RefPtr<MapelementResult> async_mapelement_find_nearest(const Point& pt, unsigned int limit = ~0, const Rect& rect = Rect(), unsigned int loadsubtables = ~0) { return m_dbthread.mapelement_find_nearest(pt, limit, rect, loadsubtables); }
                Glib::RefPtr<MapelementSaveResult> async_mapelement_save(const MapelementsDb::Mapelement& e) { return m_dbthread.mapelement_save(e); }

                Glib::RefPtr<ElevationResult> async_elevation_point(const Point& pt) { return m_dbthread.elevation_point(pt); }
                Glib::RefPtr<ElevationMapResult> async_elevation_map(const Rect& bbox) { return m_dbthread.elevation_map(bbox); }
                Glib::RefPtr<ElevationMapCairoResult> async_elevation_map_cairo(const Rect& bbox) { return m_dbthread.elevation_map_cairo(bbox); }
                Glib::RefPtr<ElevationMinMaxResult> async_elevation_minmax(const Rect& r) { return m_dbthread.elevation_minmax(r); }
                Glib::RefPtr<ElevationMinMaxResult> async_elevation_minmax(const Polygon& p) { return m_dbthread.elevation_minmax(p); }

                std::string get_aux_dir(auxdb_mode_t auxdbmode = auxdb_prefs, const std::string& dir_aux = "");
                static std::string get_default_aux_dir(void);

        protected:
#ifdef HAVE_PILOTLINK
                static PalmDbBase::PalmStringCompare::comp_t to_palm_compare(DbBaseCommon::comp_t comp);
#endif

        private:
                Preferences m_prefs;
                DbThread m_dbthread;
                FPlan m_fplandb;
                TracksDb m_tracksdb;
};

#endif /* ENGINE_H */
