/***************************************************************************
 *   Copyright (C) 2007, 2009 by Thomas Sailer                             *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"

#include <limits>
#include <sqlite3x.hpp>
#include <cstdio>
#include "geom.h"

const Point::coord_t Point::lon_min;
const Point::coord_t Point::lon_max;
const Point::coord_t Point::lat_min;
const Point::coord_t Point::lat_max;
const Point Point::invalid = Point(0, 0x80000000);

const float Point::to_deg;
const float Point::to_rad;
const float Point::from_deg;
const float Point::from_rad;

const double Point::to_deg_dbl;
const double Point::to_rad_dbl;
const double Point::from_deg_dbl;
const double Point::from_rad_dbl;

const int32_t Point::pole_lat;

const float Point::earth_radius;
const float Point::km_to_nmi;
const float Point::m_to_ft;
const float Point::ft_to_m;
const float Point::mps_to_kts;

const double Point::earth_radius_dbl;
const double Point::km_to_nmi_dbl;
const double Point::m_to_ft_dbl;
const double Point::ft_to_m_dbl;
const double Point::mps_to_kts_dbl;

/*
 * cf. Joseph O'Rourke, Computational Geometry in C
 */

/* 
 * softSurfer's (www.softsurfer.com) variant uses less multiplications,
 * but computes the same value
 * http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm
 */

/* twice the signed area of the triangle p0, p1, p2 */

int64_t Point::area2(const Point& p0, const Point& p1, const Point& p2)
{
#if 1
        int64_t p2x = p2.get_lon();
        int64_t p2y = p2.get_lat();
        int64_t p0x = p0.get_lon() - p2x;
        int64_t p0y = p0.get_lat() - p2y;
        int64_t p1x = p1.get_lon() - p2x;
        int64_t p1y = p1.get_lat() - p2y;
        return p0x * (p1y - p0y) - p0y * (p1x - p0x);
#else
        return p0.get_lon() * (int64_t)p1.get_lat() - p0.get_lat() * (int64_t)p1.get_lon() +
                        p2.get_lon() * (int64_t)p0.get_lat() - p2.get_lat() * (int64_t)p0.get_lon() +
                        p1.get_lon() * (int64_t)p2.get_lat() - p1.get_lat() * (int64_t)p2.get_lon();
#endif
}

inline bool Point::is_between_helper(const Point& p0, const Point& p1) const
{
        if (p0.get_lon() != p1.get_lon())
                return ((get_lon() >= p0.get_lon() && get_lon() <= p1.get_lon()) ||
                                (get_lon() >= p1.get_lon() && get_lon() <= p0.get_lon()));
        return ((get_lat() >= p0.get_lat() && get_lat() <= p1.get_lat()) ||
                        (get_lat() >= p1.get_lat() && get_lat() <= p0.get_lat()));
}

inline bool Point::intersect_precheck(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        int32_t xmin = l0a.get_lon();
        int32_t xmax = l0b.get_lon();
        int32_t ymin = l0a.get_lat();
        int32_t ymax = l0b.get_lat();
        if (xmin > xmax) {
                int32_t t = xmax;
                xmax = xmin;
                xmin = t;
        }
        if (ymin > ymax) {
                int32_t t = ymax;
                ymax = ymin;
                ymin = t;
        }
        if (l1a.get_lon() > xmax && l1b.get_lon() > xmax)
                return false;
        if (l1a.get_lon() < xmin && l1b.get_lon() < xmin)
                return false;
        if (l1a.get_lat() > ymax && l1b.get_lat() > ymax)
                return false;
        if (l1a.get_lat() < ymin && l1b.get_lat() < ymin)
                return false;
        return true;
}

/*
 * "proper" intersection, i.e. not one endpoint collinear with the other line
 */

bool Point::is_proper_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        if (!intersect_precheck(l0a, l0b, l1a, l1b))
                return false;
        int64_t abc = area2(l0a, l0b, l1a);
        int64_t abd = area2(l0a, l0b, l1b);
        int64_t cda = area2(l1a, l1b, l0a);
        int64_t cdb = area2(l1a, l1b, l0b);
        if (!abc || !abd || !cda || !cdb)
                return false;
        return ((abc > 0) ^ (abd > 0)) && ((cda > 0) ^ (cdb > 0));
}

/*
 * intersection, including one endpoint lying on the other line
 */

bool Point::is_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        if (!intersect_precheck(l0a, l0b, l1a, l1b))
                return false;
        int64_t abc = area2(l0a, l0b, l1a);
        int64_t abd = area2(l0a, l0b, l1b);
        int64_t cda = area2(l1a, l1b, l0a);
        int64_t cdb = area2(l1a, l1b, l0b);
        if (abc && abd && cda && cdb && ((abc > 0) ^ (abd > 0)) && ((cda > 0) ^ (cdb > 0)))
                return true;
        if (!abc && l1a.is_between_helper(l0a, l0b))
                return true;
        if (!abd && l1b.is_between_helper(l0a, l0b))
                return true;
        if (!cda && l0a.is_between_helper(l1a, l1b))
                return true;
        if (!cdb && l0b.is_between_helper(l1a, l1b))
                return true;
        return false;
}

bool Point::is_between(const Point& p0, const Point& p1) const
{
        if (area2(p0, p1, *this))
                return false;
        return is_between_helper(p0, p1);
}

bool Point::is_convex(const Point& vprev, const Point& vnext) const
{
        return vnext.is_left_or_on(vprev, *this);
}

bool Point::is_in_cone(const Point& vprev, const Point& v, const Point& vnext) const
{
        if (v.is_convex(vprev, vnext))
                return vprev.is_strictly_left(v, *this) && 
                                vnext.is_strictly_left(*this, v);
        return !(vnext.is_left_or_on(v, *this) && 
                        vprev.is_left_or_on(*this, v));
}

/* 
 * based on an algorithm by softSurfer (www.softsurfer.com)
 * http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm
 */

/*
 * left_of: tests if a point is Left|On|Right of an infinite line.
 *    Input:  three points P0, P1, and P2
 *    Return: >0 for P2 left of the line through P0 and P1
 *            =0 for P2 on the line
 *            <0 for P2 right of the line
 *    See: the January 2001 Algorithm "Area of 2D and 3D Triangles and Polygons"
 */

inline bool Point::is_left_of(const Point& p0, const Point& p1) const
{
        int64_t lon = area2(p0, p1, *this);
        if (lon < 0)
                return -1;
        if (lon > 0)
                return 1;
        return 0;
}

bool Point::left_of(const Point& p0, const Point& p1) const
{
        int64_t p0x = p0.get_lon() - get_lon();
        int64_t p0y = p0.get_lat() - get_lat();
        int64_t p1x = p1.get_lon() - get_lon();
        int64_t p1y = p1.get_lat() - get_lat();
        int64_t lon = p0x * (p1y - p0y) - p0y * (p1x - p0x);
        if (lon < 0)
                return -1;
        if (lon > 0)
                return 1;
        return 0;
}

bool Point::in_box(const Point& p0, const Point& p1) const
{
        int32_t xmin = p0.get_lon();
        int32_t xmax = p1.get_lon();
        int32_t ymin = p0.get_lat();
        int32_t ymax = p1.get_lat();
        if (xmin > xmax) {
                int32_t t = xmax;
                xmax = xmin;
                xmin = t;
        }
        if (ymin > ymax) {
                int32_t t = ymax;
                ymax = ymin;
                ymin = t;
        }
        return (get_lon() >= xmin && get_lon() <= xmax) && (get_lat() >= ymin && get_lat() <= ymax);
}

/* http://mathworld.wolfram.com/Line-LineIntersection.html */

bool Point::intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        int64_t x1 = l0a.get_lon();
        int64_t y1 = l0a.get_lat();
        int64_t x2 = l0b.get_lon();
        int64_t y2 = l0b.get_lat();
        int64_t x3 = l1a.get_lon();
        int64_t y3 = l1a.get_lat();
        int64_t x4 = l1b.get_lon();
        int64_t y4 = l1b.get_lat();
        int64_t detd = (x1 - x2) * (y3 - y4) - (x3 - x4) * (y1 - y2);
        if (!detd)
                return false;
        int64_t det1 = (x1 * y2) - (x2 * y1);
        int64_t det2 = (x3 * y4) - (x4 * y3);
        set_lon((det1 * (x3 - x4) - det2 * (x1 - x2) + detd / 2) / detd);
        set_lat((det1 * (y3 - y4) - det2 * (y1 - y2) + detd / 2) / detd);
        return true;
}

Point Point::simple_unwrap(const Point& p2) const
{
        float r1 = cosf(((*this) + p2).get_lat_rad() * 0.5);
        Point p(p2 - *this);
        p.set_lon((coord_t)(r1 * p.get_lon()));
        return p;
}

float Point::simple_true_course(const Point& p2) const
{
        Point p(simple_unwrap(p2));
        float v = atan2f(p.get_lon(), p.get_lat()) * (180.0 / M_PI);
        if (v < 0)
                v += 360;
        return v;
}

uint64_t Point::simple_distance_rel( const Point & p2 ) const
{
        Point p(simple_unwrap(p2));
        return p.length2();
}

float Point::simple_distance_km(const Point& p2) const
{
        Point p(simple_unwrap(p2));
        return p.length() * (earth_radius * to_rad);
}

Point Point::simple_course_distance_nmi(float course, float dist) const
{
        float r1 = cosf(get_lat_rad());
        float r2 = course * (M_PI / 180.0);
        float x = sinf(r2);
        float y = cosf(r2);
        r2 = r1 * r1 * x * x + y * y;
        if (r2 < 0.001)
                return *this;
        r2 = dist * (1.0 / km_to_nmi / earth_radius / to_rad) / sqrt(r2);
        x *= r2;
        y *= r2;
        return (*this) + Point((coord_t)x, (coord_t)y);
}

Rect Point::simple_box_km( float centerdist ) const
{
        centerdist *= (from_rad / earth_radius);
        float distlon = centerdist;
        if (abs(get_lat()) < pole_lat)
                distlon /= cosf(get_lat_rad());
        Point pdist((int32_t)distlon, (int32_t)centerdist);
        Point sw(*this - pdist), ne(*this + pdist);
        if (sw.get_lat() < -pole_lat)
                sw.set_lat(-pole_lat);
        if (ne.get_lat() > pole_lat)
                ne.set_lat(pole_lat);
        return Rect(sw, ne);
}

Rect Point::simple_box_nmi( float centerdist ) const
{
        return simple_box_km(centerdist * (1.0 / km_to_nmi));
}

/*
 * Spheric Geometry
 * http://de.wikipedia.org/wiki/Orthodrome
 * http://williams.best.vwh.net/avform.htm
 */

float Point::spheric_true_course(const Point& p2) const
{
        Point pdiff(p2 - *this);
        float lat1 = get_lat_rad(), lat2 = p2.get_lat_rad();
        float sinlat1 = sinf(lat1), coslat1 = cosf(lat1);
        float sinlat2 = sinf(lat2), coslat2 = cosf(lat2);
        float londiff = pdiff.get_lon_rad();
        float crs = atan2f(sinf(londiff) * coslat2, coslat1 * sinlat2 - sinlat1 * coslat2 * cosf(londiff));
        crs *= (180.0 / M_PI);
        if (crs < 0)
                crs += 360;
        return crs;
}

float Point::spheric_distance_km(const Point& p2) const
{
        Point pdiff(p2 - *this);
#if 0
        float slatd = sinf(pdiff.get_lat_rad() * 0.5f);
        float slond = sinf(pdiff.get_lon_rad() * 0.5f);
        return 2.0f * earth_radius * asinf(sqrtf(slatd * slatd + cosf(get_lat_rad()) * cosf(p2.get_lat_rad()) * slond * slond));
#else
        float lat = pdiff.get_lat_rad();
        float lon = pdiff.get_lon_rad() * cosf(get_lat_rad());
        float dist = acosf(cosf(lat) * cosf(lon));
        //double dist = acos(cos(lat) * cos(lon));
        return dist * earth_radius;
#endif
}

Point Point::spheric_course_distance_nmi(float course, float dist) const
{
        //std::cerr << "spheric_course_distance_nmi: crs " << course << " d " << dist << " pt " << (*this) << std::endl;
        dist *= (1.0f / km_to_nmi / earth_radius);
        course *= (M_PI / 180.f);
        float sintc = sinf(course), costc = cosf(course);
        float lat1 = get_lat_rad();
        float sinlat1 = sinf(lat1), coslat1 = cosf(lat1);
        float cosd = cosf(dist), sind = sinf(dist);
        float sinlat = sinlat1 * cosd + coslat1 * sind * costc;
        float lat = asinf(sinlat);
#if 1
        float dlon = atan2f(sintc * sind * coslat1, cosd - sinlat1 * sinlat);
#else
        float dlon = 0;
        float coslat = cosf(lat);
        if (fabsf(coslat) > 1e-10)
                dlon = asinf(sintc * sind / coslat);
#endif
        Point p;
        p.set_lat_rad(lat);
        p.set_lon_rad(dlon);
        p.set_lon(get_lon() + p.get_lon());
        return p;
}

double Point::spheric_distance_km_dbl(const Point & p2) const
{
        Point pdiff(p2 - *this);
#if 0
        double slatd = sin(pdiff.get_lat_rad_dbl() * 0.5);
        double slond = sin(pdiff.get_lon_rad_dbl() * 0.5);
        return 2.0 * earth_radius_dbl * asin(sqrt(slatd * slatd + cos(get_lat_rad_dbl()) * cos(p2.get_lat_rad_dbl()) * slond * slond));
#else
        double lat = pdiff.get_lat_rad_dbl();
        double lon = pdiff.get_lon_rad_dbl() * cos(get_lat_rad_dbl());
        double dist = acos(cos(lat) * cos(lon));
        return dist * earth_radius_dbl;
#endif
}

Point Point::halfway( const Point & pt2 ) const
{
        Point pd(pt2 - *this);
        pd = Point(pd.get_lon() / 2, pd.get_lat() / 2);
        return *this + pd;
}

Glib::ustring Point::get_lon_str( void ) const
{
        float t = get_lon_deg(), t1;
        char sgn = (t < 0) ? 'W' : 'E';
        t = modff(fabsf(t), &t1);
        t *= 60;
        if (t >= 59.9995) {
                t = 0.0;
                t1 += 1.0;
        }
        char buf[16];
        snprintf(buf, sizeof(buf), "%c%03.0f\302\260%08.5f'", sgn, t1, t);
        return buf;
}

Glib::ustring Point::get_lat_str( void ) const
{
        float t = get_lat_deg(), t1;
        char sgn = (t < 0) ? 'S' : 'N';
        t = modff(fabsf(t), &t1);
        t *= 60;
        if (t >= 59.9995) {
                t = 0.0;
                t1 += 1.0;
        }
        char buf[16];
        snprintf(buf, sizeof(buf), "%c%02.0f\302\260%08.5f'", sgn, t1, t);
        return buf;
}

Glib::ustring Point::get_lon_str2( void ) const
{
        float t = get_lon_deg(), t1;
        char sgn = (t < 0) ? 'W' : 'E';
        t = modff(fabsf(t), &t1);
        t *= 60;
        if (t >= 59.9995) {
                t = 0.0;
                t1 += 1.0;
        }
        char buf[16];
        snprintf(buf, sizeof(buf), "%c%03.0f %08.5f", sgn, t1, t);
        return buf;
}

Glib::ustring Point::get_lat_str2( void ) const
{
        float t = get_lat_deg(), t1;
        char sgn = (t < 0) ? 'S' : 'N';
        t = modff(fabsf(t), &t1);
        t *= 60;
        if (t >= 59.9995) {
                t = 0.0;
                t1 += 1.0;
        }
        char buf[16];
        snprintf(buf, sizeof(buf), "%c%02.0f %08.5f", sgn, t1, t);
        return buf;
}

std::ostream& Point::print(std::ostream& os) const
{
        os << '(' << (std::string)get_lon_str() << ',' << (std::string)get_lat_str() << ')';
        return os;
}

bool Point::set_lon_str( const Glib::ustring & str )
{
        char hemi;
        double deg, min, sec;
        Point::coord_t c(0);
        bool sok = sscanf(str.c_str(), "%c%lf\302\260%lf'%lf\"", &hemi, &deg, &min, &sec) == 4;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf %lf", &hemi, &deg, &min, &sec) == 4;
        if (sok) {
                if (hemi != 'E' && hemi != 'e' && hemi != 'W' && hemi != 'w')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min + (1.0 / 60.0 / 60.0) * sec) * Point::from_deg);
                if (hemi == 'W' || hemi == 'w')
                        c = -c;
                goto ok;
        }
        sok = sscanf(str.c_str(), "%c%lf\302\260%lf'", &hemi, &deg, &min) == 3;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf", &hemi, &deg, &min) == 3;
        if (sok) {
                if (hemi != 'E' && hemi != 'e' && hemi != 'W' && hemi != 'w')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min) * Point::from_deg);
                if (hemi == 'W' || hemi == 'w')
                        c = -c;
                goto ok;
        }
err:
        return false;

ok:
        lon = c;
        return true;
}

bool Point::set_lat_str( const Glib::ustring & str )
{
        char hemi;
        double deg, min, sec;
        Point::coord_t c(0);
        bool sok = sscanf(str.c_str(), "%c%lf\302\260%lf'%lf\"", &hemi, &deg, &min, &sec) == 4;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf %lf", &hemi, &deg, &min, &sec) == 4;
        if (sok) {
                if (hemi != 'N' && hemi != 'n' && hemi != 'S' && hemi != 's')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min + (1.0 / 60.0 / 60.0) * sec) * Point::from_deg);
                if (hemi == 'S' || hemi == 's')
                        c = -c;
                goto ok;
        }
        sok = sscanf(str.c_str(), "%c%lf\302\260%lf'", &hemi, &deg, &min) == 3;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf", &hemi, &deg, &min) == 3;
        if (sok) {
                if (hemi != 'N' && hemi != 'n' && hemi != 'S' && hemi != 's')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min) * Point::from_deg);
                if (hemi == 'S' || hemi == 's')
                        c = -c;
                goto ok;
        }
err:
        return false;

ok:
        lat = c;
        return true;
}

int64_t Rect::get_east_unwrapped(void ) const
{
        int64_t r(get_east());
        if (get_east() < get_west())
                r += 0x100000000LL;
        return r;
}

bool Rect::is_inside( const Point & pt ) const
{
        Point p0(pt - get_southwest());
        Point p1(get_northeast() - get_southwest());
        return ((uint32_t)p0.get_lon() <= (uint32_t)p1.get_lon()) &&
                        ((uint32_t)p0.get_lat() <= (uint32_t)p1.get_lat());
}

bool Rect::is_lon_inside( const Point & pt ) const
{
        Point p0(pt - get_southwest());
        Point p1(get_northeast() - get_southwest());
        return ((uint32_t)p0.get_lon() <= (uint32_t)p1.get_lon());
}

bool Rect::is_lat_inside( const Point & pt ) const
{
        Point p0(pt - get_southwest());
        Point p1(get_northeast() - get_southwest());
        return ((uint32_t)p0.get_lat() <= (uint32_t)p1.get_lat());
}

uint64_t Rect::simple_distance_rel( const Point & pt ) const
{
        if (is_inside(pt))
                return 0;
        uint64_t r0 = pt.simple_distance_rel(get_southwest());
        uint64_t r1 = pt.simple_distance_rel(get_northeast());
        uint64_t r2 = pt.simple_distance_rel(get_southeast());
        uint64_t r3 = pt.simple_distance_rel(get_northwest());
        uint64_t r = std::min(std::min(r0, r1), std::min(r2, r3));
        if (is_lon_inside(pt)) {
                uint64_t r0 = pt.simple_distance_rel(Point(pt.get_lon(), get_north()));
                uint64_t r1 = pt.simple_distance_rel(Point(pt.get_lon(), get_south()));
                r = std::min(r, std::min(r0, r1));
        }
        if (is_lat_inside(pt)) {
                uint64_t r0 = pt.simple_distance_rel(Point(get_east(), pt.get_lat()));
                uint64_t r1 = pt.simple_distance_rel(Point(get_west(), pt.get_lat()));
                r = std::min(r, std::min(r0, r1));
        }
        return r;
}

float Rect::simple_distance_km( const Point & pt ) const
{
        return sqrt(simple_distance_rel(pt)) * (Point::earth_radius * Point::to_rad);
}

float Rect::spheric_distance_km( const Point & pt ) const
{
        if (is_inside(pt))
                return 0;
        float r0 = pt.spheric_distance_km(get_southwest());
        float r1 = pt.spheric_distance_km(get_northeast());
        float r2 = pt.spheric_distance_km(get_southeast());
        float r3 = pt.spheric_distance_km(get_northwest());
        float r = std::min(std::min(r0, r1), std::min(r2, r3));
        if (is_lon_inside(pt)) {
                float r0 = pt.spheric_distance_km(Point(pt.get_lon(), get_north()));
                float r1 = pt.spheric_distance_km(Point(pt.get_lon(), get_south()));
                r = std::min(r, std::min(r0, r1));
        }
        if (is_lat_inside(pt)) {
                float r0 = pt.spheric_distance_km(Point(get_east(), pt.get_lat()));
                float r1 = pt.spheric_distance_km(Point(get_west(), pt.get_lat()));
                r = std::min(r, std::min(r0, r1));
        }
        return r;
}

bool Rect::is_intersect( const Rect & r ) const
{
        return (is_inside(r.get_northeast()) ||
                is_inside(r.get_northwest()) ||
                is_inside(r.get_southeast()) ||
                is_inside(r.get_southwest()) ||
                r.is_inside(get_northeast()) ||
                r.is_inside(get_northwest()) ||
                r.is_inside(get_southeast()) ||
                r.is_inside(get_southwest()));
}

uint64_t Rect::simple_distance_rel( const Rect & r ) const
{
        if (is_intersect(r))
                return 0;
        uint64_t r0 = simple_distance_rel(get_southwest());
        uint64_t r1 = simple_distance_rel(get_northeast());
        uint64_t r2 = simple_distance_rel(get_southeast());
        uint64_t r3 = simple_distance_rel(get_northwest());
        return std::min(std::min(r0, r1), std::min(r2, r3));
}

float Rect::simple_distance_km( const Rect & r ) const
{
        return sqrt(simple_distance_rel(r)) * (Point::earth_radius * Point::to_rad);
}

float Rect::spheric_distance_km( const Rect & r ) const
{
        if (is_intersect(r))
                return 0;
        float r0 = spheric_distance_km(get_southwest());
        float r1 = spheric_distance_km(get_northeast());
        float r2 = spheric_distance_km(get_southeast());
        float r3 = spheric_distance_km(get_northwest());
        return std::min(std::min(r0, r1), std::min(r2, r3));
}

std::ostream & Rect::print( std::ostream & os ) const
{
        return os << get_southwest() << '-' << get_northeast();
}

std::ostream& operator<<(std::ostream& os, const Point& p)
{
        return p.print(os);
}

std::ostream & operator<<(std::ostream & os, const Rect & r)
{
        return r.print(os);
}

Polygon::reference Polygon::operator[](size_type x)
{
        size_type y(size());
        if (y > 0)
                x %= y;
        return base_t::operator[](x);
}

Polygon::const_reference Polygon::operator[](size_type x) const
{
        size_type y(size());
        if (y > 0)
                x %= y;
        return base_t::operator[](x);
}

bool Polygon::is_on_boundary( const Point & pt ) const
{
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, i++)
                if (pt.is_between(operator[](i), operator[](j)))
                        return true;
        return false;
}

bool Polygon::is_intersection(const Point & la, const Point & lb) const
{
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, i++)
                if (Point::is_intersection(la, lb, operator[](i), operator[](j)))
                        return true;
        return false;
}

bool Polygon::is_intersection(const Rect & r) const
{
        return is_intersection(r.get_southwest(), r.get_northwest())
            || is_intersection(r.get_northwest(), r.get_northeast())
            || is_intersection(r.get_northeast(), r.get_southeast())
            || is_intersection(r.get_southeast(), r.get_southwest());
}

int Polygon::windingnumber( const Point & pt ) const
{
        int wn = 0; /* the winding number counter */

        /* loop through all edges of the polygon */
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, i++) {
                int64_t lon = Point::area2(operator[](j), operator[](i), pt);
                /* check if pt lies on edge from V[j] to V[i] */
                if (!lon && pt.is_between_helper(operator[](j), operator[](i)))
                        return std::numeric_limits<int>::max();
                /* edge from V[j] to V[i] */
                if (operator[](j).get_lat() <= pt.get_lat()) {
                        if (operator[](i).get_lat() > pt.get_lat())
                                /* an upward crossing */
                                if (lon > 0)
                                        /* P left of edge: have a valid up intersect */
                                        wn++;
                } else {
                        if (operator[](i).get_lat() <= pt.get_lat())
                                /* a downward crossing */
                                if (lon < 0)
                                        /* P right of edge: have a valid down intersect */
                                        wn--;
                }
        }
        return wn;
}

/*
 * negative, if polygon is clockwise oriented
 */
int64_t Polygon::area2( void ) const
{
        int64_t area = 0;

        /* http://mathworld.wolfram.com/PolygonArea.html */
        /* loop through all edges of the polygon */
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, i++) {
                /* edge from V[j] to V[i]  (j = (i + 1) mod size) */
                int64_t x1 = operator[](j).get_lon();
                int64_t y1 = operator[](j).get_lat();
                int64_t x2 = operator[](i).get_lon();
                int64_t y2 = operator[](i).get_lat();
                area += (x1 * y2 - x2 * y1);
        }
        return area;
}

void Polygon::remove_redundant_polypoints( void )
{
        unsigned int sz = size();
        if (sz >= 2) {
                if (operator[](0) == operator[](sz-1)) {
                        sz--;
                        resize(sz);
                }
        }
        bool first = true;
        for (unsigned int i = 0; i < sz;) {
                unsigned int j = (i + 1) % sz;
                unsigned int k = (j + 1) % sz;
                if (!operator[](j).is_between(operator[](i), operator[](k))) {
                        i++;
                        continue;
                }
#if 0
                if (first) {
                        std::cerr << "remove_redundant_polypoints: poly:";
                        for (unsigned int m = 0; m < sz; m++)
                                operator[](m).print(std::cerr << ' ');
                        std::cerr << std::endl;
                }
                first = false;
                operator[](j).print(std::cerr << "remove_redundant_polypoints: removing: ");
                operator[](i).print(std::cerr << " between: ");
                operator[](k).print(std::cerr << ' ') << std::endl;
#endif
                erase(begin() + j);
                sz--;
        }
        if (sz < 3)
                clear();
}

Rect Polygon::get_bbox( void ) const
{
        if (!size())
                return Rect();
        Point pt(operator[](0));
        Point sw(0, 0), ne(0, 0);
        for (unsigned int i = 1; i < size(); i++) {
                Point p(operator[](i) - pt);
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        return Rect(sw + pt, ne + pt);
}

void Polygon::bindblob(sqlite3x::sqlite3_command & cmd, int index) const
{
        unsigned int sz(size() * 8);
        if (!sz) {
                cmd.bind(index);
                return;
        }
        uint8_t data[sz];
        for (size_type i(0); i < size(); i++) {
                Point pt(operator[](i));
                data[8*i+0] = pt.get_lon() >> 0;
                data[8*i+1] = pt.get_lon() >> 8;
                data[8*i+2] = pt.get_lon() >> 16;
                data[8*i+3] = pt.get_lon() >> 24;
                data[8*i+4] = pt.get_lat() >> 0;
                data[8*i+5] = pt.get_lat() >> 8;
                data[8*i+6] = pt.get_lat() >> 16;
                data[8*i+7] = pt.get_lat() >> 24;
        }
        cmd.bind(index, data, sz);
}

void Polygon::getblob(sqlite3x::sqlite3_cursor & cursor, int index)
{
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        if (sz <= 7)
                return;
        sz >>= 3;
        reserve(sz);
        for (size_type i(0); i < sz; i++) {
                uint32_t lon, lat;
                lon = data[8*i+0] | (data[8*i+1] << 8) | (data[8*i+2] << 16) | (data[8*i+3] << 24);
                lat = data[8*i+4] | (data[8*i+5] << 8) | (data[8*i+6] << 16) | (data[8*i+7] << 24);
                push_back(Point(lon, lat));
        }
}

bool PolygonHole::is_on_boundary(const Point & pt) const
{
        if (m_exterior.is_on_boundary(pt))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                if (ii->is_on_boundary(pt))
                        return true;
        return false;
}

bool PolygonHole::is_intersection(const Point & la, const Point & lb) const
{
        if (m_exterior.is_intersection(la, lb))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                if (ii->is_intersection(la, lb))
                        return true;
        return false;
}

bool PolygonHole::is_intersection(const Rect & r) const
{
        if (m_exterior.is_intersection(r))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                if (ii->is_intersection(r))
                        return true;
        return false;
}

int PolygonHole::windingnumber(const Point & pt) const
{
#if 1
        int wn(m_exterior.windingnumber(pt));
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                wn += ii->windingnumber(pt);
        return wn;
#else
        int wn(m_exterior.windingnumber(pt));
        if (!wn)
                return wn;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                if (ii->windingnumber(pt))
                        return 0;
        return wn;
#endif
}

int64_t PolygonHole::area2(void ) const
{
        int64_t r(m_exterior.area2());
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                r += ii->area2();
        return r;
}

void PolygonHole::remove_redundant_polypoints(void )
{
        m_exterior.remove_redundant_polypoints();
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                ii->remove_redundant_polypoints();
}

void PolygonHole::normalize(void )
{
        bool es(m_exterior.area2() < 0);
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++) {
                bool is(ii->area2() < 0);
                if (is == es)
                        std::reverse(ii->begin(), ii->end());
        }
}

unsigned int PolygonHole::bloblength(const Polygon & p)
{
        return p.size() * 8 + 4;
}

uint8_t * PolygonHole::blobencode(const Polygon & p, uint8_t * data)
{
        unsigned int sz(p.size());
        data[0] = sz;
        data[1] = sz >> 8;
        data[2] = sz >> 16;
        data[3] = sz >> 24;
        data += 4;
        for (unsigned int i = 0; i < sz; ++i) {
                Point pt(p[i]);
                data[0] = pt.get_lon() >> 0;
                data[1] = pt.get_lon() >> 8;
                data[2] = pt.get_lon() >> 16;
                data[3] = pt.get_lon() >> 24;
                data[4] = pt.get_lat() >> 0;
                data[5] = pt.get_lat() >> 8;
                data[6] = pt.get_lat() >> 16;
                data[7] = pt.get_lat() >> 24;
                data += 8;
        }
        return data;
}

uint8_t const *PolygonHole::blobdecode(Polygon& p, uint8_t const *data, uint8_t const *end)
{
        p.clear();
        if (data + 4 > end)
                return end;
        uint32_t sz(data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24));
        data += 4;
        if (sz >= 0x01000000 || data + 8 * sz > end)
                return end;
        p.reserve(sz);
        for (unsigned int i = 0; i < sz; ++i) {
                uint32_t lon, lat;
                lon = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
                lat = data[4] | (data[5] << 8) | (data[6] << 16) | (data[7] << 24);
                data += 8;
                p.push_back(Point(lon, lat));
        }
        return data;
}

void PolygonHole::getblob(sqlite3x::sqlite3_cursor & cursor, int index)
{
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        uint8_t const *end(data + sz);
        data = blobdecode(m_exterior, data, end);
        while (data < end) {
                m_interior.push_back(Polygon());
                data = blobdecode(m_interior.back(), data, end);
        }
}

void PolygonHole::bindblob(sqlite3x::sqlite3_command & cmd, int index) const
{
        unsigned int sz(bloblength(m_exterior));
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                sz += bloblength(*ii);
        if (sz <= 4) {
                cmd.bind(index);
                return;
        }
        uint8_t data[sz];
        uint8_t *d = blobencode(m_exterior, data);
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ii++)
                d = blobencode(*ii, d);
        cmd.bind(index, data, sz);
}

CartesianVector3D & CartesianVector3D::operator +=( const CartesianVector3D & v )
{
        m_v[0] += v.m_v[0];
        m_v[1] += v.m_v[1];
        m_v[2] += v.m_v[2];
        return *this;
}

CartesianVector3D & CartesianVector3D::operator -=( const CartesianVector3D & v )
{
        m_v[0] -= v.m_v[0];
        m_v[1] -= v.m_v[1];
        m_v[2] -= v.m_v[2];
        return *this;
}

CartesianVector3D & CartesianVector3D::operator *=( float v )
{
        m_v[0] *= v;
        m_v[1] *= v;
        m_v[2] *= v;
        return *this;
}

float CartesianVector3D::scalarprod( const CartesianVector3D& v ) const
{
        return m_v[0] * v.m_v[0] + m_v[1] * v.m_v[1] + m_v[2] * v.m_v[2];
}

float CartesianVector3D::length2( void ) const
{
        return m_v[0] * m_v[0] + m_v[1] * m_v[1] + m_v[2] * m_v[2];
}

float CartesianVector3D::nearest_fraction( const CartesianVector3D & v0, const CartesianVector3D & v1 )
{
        CartesianVector3D n(v1 - v0);
        float t = n.length2();
        if (t < 0.001)
                return 0.5;
        t = n.scalarprod((*this) - v0) / t;
        return std::min(std::max(t, 0.0f), 1.0f);
}

void CartesianVector3D::rotate_xz( float ang )
{
        float c = cos(ang), s = sin(ang);
        float t = m_v[0] * c - m_v[2] * s;
        m_v[2] = m_v[0] * s + m_v[2] * c;
        m_v[0] = t;
}

void CartesianVector3D::rotate_xy( float ang )
{
        float c = cos(ang), s = sin(ang);
        float t = m_v[0] * c - m_v[1] * s;
        m_v[1] = m_v[0] * s + m_v[1] * c;
        m_v[0] = t;
}

CartesianVector3D::operator PolarVector3D( void ) const
{
        float r = length();
        float lon = atan2(m_v[1], m_v[0]);
        float lat = 0;
        if (r >= 0.1)
                lat = asin(m_v[2] / r);
        return PolarVector3D(lat, lon, r);
}

PolarVector3D::operator CartesianVector3D( void ) const
{
        float r = m_r;
        float z = r * sin(m_lat);
        r *= cos(m_lat);
        float x = r * cos(m_lon);
        float y = r * sin(m_lon);
        return CartesianVector3D(x, y, z);
}

