//
// C++ Interface: geom
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef GEOM_H
#define GEOM_H

//#include <cstdint>
#include <stdint.h>
#include <cmath>
#include <iostream>
#include <vector>
#include <glibmm.h>

#include "sysdeps.h"

namespace sqlite3x {
        class sqlite3_cursor;
        class sqlite3_command;
};

class Rect;
class Polygon;

class Point {
        public:
                typedef int32_t coord_t;
                static const coord_t lon_min = 0x80000000;
                static const coord_t lon_max = 0x7FFFFFFF;
                static const coord_t lat_min = 0xC0000000;
                static const coord_t lat_max = 0x40000000;
                static const Point invalid;
                Point(coord_t lo = 0, coord_t la = 0) : lon(lo), lat(la) {}
                coord_t get_lon(void) const { return lon; }
                coord_t get_lat(void) const { return lat; }
                void set_lon(coord_t lo) { lon = lo; }
                void set_lat(coord_t la) { lat = la; }
                bool operator==(const Point &p) const { return lon == p.lon && lat == p.lat; }
                bool operator!=(const Point &p) const { return !operator==(p); }
                Point operator+(const Point &p) const { return Point(lon + (uint32_t)p.lon, lat + (uint32_t)p.lat); }
                Point operator-(const Point &p) const { return Point(lon - (uint32_t)p.lon, lat - (uint32_t)p.lat); }
                Point operator-() const { return Point(-lon, -lat); }
                Point& operator+=(const Point &p) { lon += (uint32_t)p.lon; lat += (uint32_t)p.lat; return *this; }
                Point& operator-=(const Point &p) { lon -= (uint32_t)p.lon; lat -= (uint32_t)p.lat; return *this; }
                bool operator<(const Point &p) const { return lon < p.lon && lat < p.lat; }
                bool operator<=(const Point &p) const { return lon <= p.lon && lat <= p.lat; }
                bool operator>(const Point &p) const { return lon > p.lon && lat > p.lat; }
                bool operator>=(const Point &p) const { return lon >= p.lon && lat >= p.lat; }
                uint64_t length2(void) const { return (uint64_t)(get_lon() * (int64_t)get_lon()) + (uint64_t)(get_lat() * (int64_t)get_lat()); }
                double length(void) const { return sqrt(length2()); }
                bool left_of(const Point& p0, const Point& p1) const;
                bool intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);

                bool is_strictly_left(const Point& p0, const Point& p1) const { return area2(p0, p1, *this) > 0; }
                bool is_left_or_on(const Point& p0, const Point& p1) const { return area2(p0, p1, *this) >= 0; }
                bool is_collinear(const Point& p0, const Point& p1) const { return area2(p0, p1, *this) == 0; }
                bool is_between(const Point& p0, const Point& p1) const;
                bool is_convex(const Point& vprev, const Point& vnelont) const;
                bool is_in_cone(const Point& vprev, const Point& v, const Point& vnelont) const;
                bool is_left_of(const Point& p0, const Point& p1) const;
                bool in_box(const Point& p0, const Point& p1) const;
                bool is_invalid(void) const { return lat < lat_min || lat > lat_max; }

                std::ostream& print(std::ostream& os) const;

                static int64_t area2(const Point& p0, const Point& p1, const Point& p2);
                static bool is_proper_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);
                static bool is_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);

        private:
                Point simple_unwrap(const Point& p2) const;

        public:
                static const float to_deg = 90.0 / (1UL << 30);
                static const float to_rad = M_PI / (1UL << 31);
                static const float from_deg = (1UL << 30) / 90.0;
                static const float from_rad = (1UL << 31) / M_PI;

                static const double to_deg_dbl = 90.0 / (1UL << 30);
                static const double to_rad_dbl = M_PI / (1UL << 31);
                static const double from_deg_dbl = (1UL << 30) / 90.0;
                static const double from_rad_dbl = (1UL << 31) / M_PI;

                static const int32_t pole_lat = (1UL << 30);

                static const float earth_radius = 6378.155;
                static const float km_to_nmi = 0.5399568;
                static const float m_to_ft = 3.2808399;
                static const float ft_to_m = 0.3048;
                static const float mps_to_kts = 3.6 * 0.5399568;

                static const double earth_radius_dbl = 6378.155;
                static const double km_to_nmi_dbl = 0.5399568;
                static const double m_to_ft_dbl = 3.2808399;
                static const double ft_to_m_dbl = 0.3048;
                static const double mps_to_kts_dbl = 3.6 * 0.5399568;

                template<typename I, typename F> static I round(F x) { return static_cast<I>((x < 0) ? (x - 0.5) : (x + 0.5)); }

                float get_lon_rad(void) const { return lon * to_rad; }
                float get_lat_rad(void) const { return lat * to_rad; }
                void set_lon_rad(float lo) { lon = round<coord_t,float>(lo * from_rad); }
                void set_lat_rad(float la) { lat = round<coord_t,float>(la * from_rad); }

                double get_lon_rad_dbl(void) const { return lon * to_rad_dbl; }
                double get_lat_rad_dbl(void) const { return lat * to_rad_dbl; }
                void set_lon_rad_dbl(double lo) { lon = round<coord_t,double>(lo * from_rad_dbl); }
                void set_lat_rad_dbl(double la) { lat = round<coord_t,double>(la * from_rad_dbl); }

                float get_lon_deg(void) const { return lon * to_deg; }
                float get_lat_deg(void) const { return lat * to_deg; }
                void set_lon_deg(float lo) { lon = round<coord_t,float>(lo * from_deg); }
                void set_lat_deg(float la) { lat = round<coord_t,float>(la * from_deg); }

                double get_lon_deg_dbl(void) const { return lon * to_deg_dbl; }
                double get_lat_deg_dbl(void) const { return lat * to_deg_dbl; }
                void set_lon_deg_dbl(double lo) { lon = round<coord_t,double>(lo * from_deg_dbl); }
                void set_lat_deg_dbl(double la) { lat = round<coord_t,double>(la * from_deg_dbl); }

                float simple_true_course(const Point& p2) const;
                uint64_t simple_distance_rel(const Point& p2) const;
                float simple_distance_km(const Point& p2) const;
                float simple_distance_nmi(const Point& p2) const { return simple_distance_km(p2) * km_to_nmi; }
                Point simple_course_distance_nmi(float course, float dist) const;
                Rect simple_box_km(float centerdist) const;
                Rect simple_box_nmi(float centerdist) const;

                float spheric_true_course(const Point& p2) const;
                float spheric_distance_km(const Point& p2) const;
                float spheric_distance_nmi(const Point& p2) const { return spheric_distance_km(p2) * km_to_nmi; }
                Point spheric_course_distance_nmi(float course, float dist) const;

                double spheric_distance_km_dbl(const Point& p2) const;
                double spheric_distance_nmi_dbl(const Point& p2) const { return spheric_distance_km_dbl(p2) * km_to_nmi_dbl; }

                Point halfway(const Point& pt2) const;

                Glib::ustring get_lon_str(void) const;
                Glib::ustring get_lat_str(void) const;
                Glib::ustring get_lon_str2(void) const;
                Glib::ustring get_lat_str2(void) const;

                bool set_lon_str(const Glib::ustring& str);
                bool set_lat_str(const Glib::ustring& str);

        private:
                coord_t lon, lat;

                bool is_between_helper(const Point& p0, const Point& p1) const;

                static bool intersect_precheck(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);

                friend class Polygon;
};

class Rect {
        public:
                Rect(void) : southwest(Point(0, -Point::pole_lat)), northeast(Point(-1, Point::pole_lat)) {}
                Rect(const Point& sw, const Point& ne) : southwest(sw), northeast(ne) {}
                Point get_southwest(void) const { return southwest; }
                Point get_northeast(void) const { return northeast; }
                Point get_southeast(void) const { return Point(northeast.get_lon(), southwest.get_lat()); }
                Point get_northwest(void) const { return Point(southwest.get_lon(), northeast.get_lat()); }
                Point::coord_t get_north(void) const { return northeast.get_lat(); }
                Point::coord_t get_south(void) const { return southwest.get_lat(); }
                Point::coord_t get_east(void) const { return northeast.get_lon(); }
                int64_t get_east_unwrapped(void) const;
                Point::coord_t get_west(void) const { return southwest.get_lon(); }
                bool operator==(const Rect &r) const { return get_southwest() == r.get_southwest() && get_northeast() == r.get_northeast(); }
                bool operator!=(const Rect &r) const { return !operator==(r); }
                Rect& operator+=(const Point &p) { southwest += p; northeast += p; return *this; }
                Rect& operator-=(const Point &p) { southwest -= p; northeast -= p; return *this; }
                Rect operator+(const Point &p) { Rect r(*this); r += p; return r; }
                Rect operator-(const Point &p) { Rect r(*this); r -= p; return r; }
                bool is_inside(const Point& pt) const;
                bool is_inside(const Rect& r) const { return is_inside(r.get_southwest()) && is_inside(r.get_northeast()); }
                bool is_lon_inside(const Point& pt) const;
                bool is_lat_inside(const Point& pt) const;
                uint64_t simple_distance_rel(const Point& pt) const;
                float simple_distance_km(const Point& pt) const;
                float simple_distance_nmi(const Point& pt) const { return simple_distance_km(pt) * Point::km_to_nmi; }
                float spheric_distance_km(const Point& pt) const;
                float spheric_distance_nmi(const Point& pt) const { return spheric_distance_km(pt) * Point::km_to_nmi; }
                bool is_intersect(const Rect& r) const;
                uint64_t simple_distance_rel(const Rect& r) const;
                float simple_distance_km(const Rect& r) const;
                float simple_distance_nmi(const Rect& r) const { return simple_distance_km(r) * Point::km_to_nmi; }
                float spheric_distance_km(const Rect& r) const;
                float spheric_distance_nmi(const Rect& r) const { return spheric_distance_km(r) * Point::km_to_nmi; }
                Point boxcenter(void) const { return southwest.halfway(northeast); }

                std::ostream& print(std::ostream& os) const;

        private:
                Point southwest;
                Point northeast;
};

class Polygon : public std::vector<Point> {
        private:
                typedef std::vector<Point> base_t;

        public:
                reference operator[](size_type x);
                const_reference operator[](size_type x) const;
                bool is_on_boundary(const Point& pt) const;
                bool is_intersection(const Point& la, const Point& lb) const;
                bool is_intersection(const Rect& r) const;
                int windingnumber(const Point& pt) const;
                int64_t area2(void) const;
                void remove_redundant_polypoints(void);
                Rect get_bbox(void) const;
                void bindblob(sqlite3x::sqlite3_command& cmd, int index) const;
                void getblob(sqlite3x::sqlite3_cursor& cursor, int index);
};

class PolygonHole {
        public:
                PolygonHole(const Polygon& extr = Polygon(), const std::vector<Polygon>& intr = std::vector<Polygon>()) : m_exterior(extr), m_interior(intr) {}
                void clear(void) { m_exterior.clear(); m_interior.clear(); }
                const Polygon& get_exterior(void) const { return m_exterior; }
                void set_exterior(const Polygon& extr) { m_exterior = extr; }
                unsigned int get_nrinterior(void) const { return m_interior.size(); }
                Polygon& operator[](unsigned int x) { return m_interior[x]; }
                const Polygon& operator[](unsigned int x) const { return m_interior[x]; }
                void add_interior(const Polygon& intr) { m_interior.push_back(intr); }
                void insert_interior(const Polygon& intr, int index) { m_interior.insert(m_interior.begin() + index, intr); }
                void erase_interior(int index) { m_interior.erase(m_interior.begin() + index); }
                void clear_interior(void) { m_interior.clear(); }
                bool is_on_boundary(const Point& pt) const;
                bool is_intersection(const Point& la, const Point& lb) const;
                bool is_intersection(const Rect& r) const;
                int windingnumber(const Point& pt) const;
                int64_t area2(void) const;
                void remove_redundant_polypoints(void);
                Rect get_bbox(void) const { return m_exterior.get_bbox(); }
                void bindblob(sqlite3x::sqlite3_command& cmd, int index) const;
                void getblob(sqlite3x::sqlite3_cursor& cursor, int index);
                void normalize(void);

        private:
                Polygon m_exterior;
                typedef std::vector<Polygon> interior_t;
                interior_t m_interior;

                static unsigned int bloblength(const Polygon& p);
                static uint8_t *blobencode(const Polygon& p, uint8_t *data);
                static uint8_t const *blobdecode(Polygon& p, uint8_t const *data, uint8_t const *end);
};

class PolarVector3D;

class CartesianVector3D {
        public:
                static const float from_deg     = M_PI / 180.0;

                CartesianVector3D(float x = 0.0, float y = 0.0, float z = 0.0) { m_v[0] = x; m_v[1] = y; m_v[2] = z; }
                float getx(void) const { return m_v[0]; }
                float gety(void) const { return m_v[1]; }
                float getz(void) const { return m_v[2]; }
                void setx(float v) { m_v[0] = v; }
                void sety(float v) { m_v[1] = v; }
                void setz(float v) { m_v[2] = v; }
                CartesianVector3D& operator+=(const CartesianVector3D& v);
                CartesianVector3D& operator-=(const CartesianVector3D& v);
                CartesianVector3D& operator*=(float v);
                CartesianVector3D operator+(const CartesianVector3D& v) const { CartesianVector3D pv(*this); pv += v; return pv; }
                CartesianVector3D operator-(const CartesianVector3D& v) const { CartesianVector3D pv(*this); pv -= v; return pv; }
                CartesianVector3D operator-(void) const { CartesianVector3D pv; pv -= *this; return pv; }
                CartesianVector3D operator*(float v) const { CartesianVector3D pv(*this); pv *= v; return pv; }
                float scalarprod(const CartesianVector3D& v) const;
                float length2(void) const;
                float length(void) const { return sqrt(length2()); }
                float nearest_fraction(const CartesianVector3D& v0, const CartesianVector3D& v1);
                void rotate_xz(float ang);
                void rotate_xy(float ang);
                operator PolarVector3D(void) const;

        private:
                float m_v[3];
};

class PolarVector3D {
        public:
                static const float earth_radius = 6378.155 * 1000.0; /* earth_radius in m */
                static const float km_to_nmi    = 0.5399568;
                static const float m_to_ft      = 3.2808399;
                static const float ft_to_m      = 0.3048;
                static const float from_deg     = M_PI / 180.0;
                static const float to_deg       = 180.0 / M_PI;

                PolarVector3D(float lat = 0, float lon = 0, float r = 0) : m_lat(lat), m_lon(lon), m_r(r) {}
                PolarVector3D(const Point& pt, int16_t altft) : m_lat(pt.get_lat_rad()), m_lon(pt.get_lon_rad()), m_r(altft * ft_to_m + earth_radius) {}
                float getlat(void) const { return m_lat; }
                float getlatdeg(void) const { return m_lat * to_deg; }
                float getlon(void) const { return m_lon; }
                float getlondeg(void) const { return m_lon * to_deg; }
                Point getcoord(void) const { return Point((int32_t)(m_lon * Point::from_rad), (int32_t)(m_lat * Point::from_rad)); }
                float getalt(void) const { return m_r - earth_radius; }
                float getaltft(void) const { return (m_r - earth_radius) * m_to_ft; }
                float getr(void) const { return m_r; }
                void setlat(float v) { m_lat = v; }
                void setlatdeg(float v) { m_lat = v * from_deg; }
                void setlon(float v) { m_lon = v; }
                void setlondeg(float v) { m_lon = v * from_deg; }
                void setcoord(const Point& pt) { m_lon = pt.get_lon_rad(); m_lat = pt.get_lat_rad(); }
                void setr(float v) { m_r = v; }
                void setalt(float v) { m_r = v + earth_radius; }
                void setaltft(float v) { m_r = v * ft_to_m + earth_radius; }
                PolarVector3D& operator*=(float s) { m_r *= s; return *this; }
                PolarVector3D operator*(float s) const { PolarVector3D pv(*this); pv *= s; return pv; }
                operator CartesianVector3D(void) const;

        private:
                float m_lat, m_lon, m_r;
};

std::ostream& operator<<(std::ostream& os, const Point& p);
std::ostream& operator<<(std::ostream& os, const Rect& r);

#endif /* GEOM_H */
