//
// C++ Implementation: fplan
//
// Description: Flight Plan XML Import/Export Utility (PalmPilot Compatible Format)
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include <libxml++/libxml++.h>
#include <unistd.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <sqlite3.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdexcept>
#include <stdarg.h>
#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>

#include "fplan.h"

class FPlanXmlImporter : public xmlpp::SaxParser {
        public:
                FPlanXmlImporter(FPlan& dbx);
                virtual ~FPlanXmlImporter();

        protected:
                virtual void on_start_document();
                virtual void on_end_document();
                virtual void on_start_element(const Glib::ustring& name,
                                              const AttributeList& properties);
                virtual void on_end_element(const Glib::ustring& name);
                virtual void on_characters(const Glib::ustring& characters);
                virtual void on_comment(const Glib::ustring& text);
                virtual void on_warning(const Glib::ustring& text);
                virtual void on_error(const Glib::ustring& text);
                virtual void on_fatal_error(const Glib::ustring& text);

        private:
                FPlan& db;
                FPlanRoute route;
                bool indoc;
                uint32_t plannr, wptnr;
};

FPlanXmlImporter::FPlanXmlImporter(FPlan& dbx)
        : db(dbx), route(db), indoc(false), plannr(0), wptnr(0)
{
}

FPlanXmlImporter::~FPlanXmlImporter()
{
}

void FPlanXmlImporter::on_start_document( )
{
}

void FPlanXmlImporter::on_end_document( )
{
}

void FPlanXmlImporter::on_start_element( const Glib::ustring & name, const AttributeList & properties )
{
        if (name == "PilotFlightPlans") {
                indoc = true;
                for (AttributeList::const_iterator ai(properties.begin()), ae(properties.end()); ai != ae; ai++) {
                        Glib::ustring pname((*ai).name);
                        Glib::ustring pvalue((*ai).value);
                        if (pname == "CurFP") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                // if (is.good()) engine.SetCurPlan(t);
                                continue;
                        }
                }
                return;
        }
        if (!indoc)
                return;
        if (name == "NavFlightPlan") {
                route.new_fp();
                for (AttributeList::const_iterator ai(properties.begin()), ae(properties.end()); ai != ae; ai++) {
                        Glib::ustring pname((*ai).name);
                        Glib::ustring pvalue((*ai).value);
                        if (pname == "Note") {
                                route.set_note(pvalue);
                                continue;
                        }
                        if (pname == "TimeOffBlock") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                route.set_time_offblock(t);
                                continue;
                        }
                        if (pname == "SavedTimeOffBlock") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                route.set_save_time_offblock(t);
                                continue;
                        }
                        if (pname == "TimeOnBlock") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                route.set_time_onblock(t);
                                continue;
                        }
                        if (pname == "SavedTimeOnBlock") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                route.set_save_time_onblock(t);
                                continue;
                        }
                        if (pname == "CurrentWaypoint") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                route.set_curwpt(t);
                                continue;
                        }
                }
                std::cerr << "new flight plan" << std::endl;
                return;
        }
        if (name == "WayPoint") {
                FPlanWaypoint wpt;
                for (AttributeList::const_iterator ai(properties.begin()), ae(properties.end()); ai != ae; ai++) {
                        Glib::ustring pname((*ai).name);
                        Glib::ustring pvalue((*ai).value);
                        if (pname == "Note") {
                                wpt.set_note(pvalue);
                                continue;
                        }
                        if (pname == "ICAO") {
                                wpt.set_icao(pvalue);
                                continue;
                        }
                        if (pname == "name") {
                                wpt.set_name(pvalue);
                                continue;
                        }
                        if (pname == "Time") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_time(t);
                                continue;
                        }
                        if (pname == "SavedTime") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_save_time(t);
                                continue;
                        }
                        if (pname == "frequency") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_frequency(t);
                                continue;
                        }
                        if (pname == "lon") {
                                std::istringstream is(pvalue);
                                int32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_lon(t);
                                continue;
                        }
                        if (pname == "lat") {
                                std::istringstream is(pvalue);
                                int32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_lat(t);
                                continue;
                        }
                        if (pname == "alt") {
                                std::istringstream is(pvalue);
                                int32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_altitude(t);
                                continue;
                        }
                        if (pname == "flags") {
                                std::istringstream is(pvalue);
                                uint32_t t;
                                is >> t;
                                if (is.fail())
                                        continue;
                                wpt.set_flags(t);
                                continue;
                        }
                }
                route.insert_wpt(~0, wpt);
                std::cerr << "new waypoint: " << (route.get_nrwpt()-1) << ": " << wpt.get_icao().raw() << ' ' << wpt.get_name().raw() << std::endl;
                return;
        }
}

void FPlanXmlImporter::on_end_element( const Glib::ustring & name )
{
        if (!indoc)
                return;
        if (name == "NavFlightPlan") {
                route.save_fp();
                route.new_fp();
        }
}

void FPlanXmlImporter::on_characters( const Glib::ustring & characters )
{
}

void FPlanXmlImporter::on_comment( const Glib::ustring & text )
{
}

void FPlanXmlImporter::on_warning( const Glib::ustring & text )
{
}

void FPlanXmlImporter::on_error( const Glib::ustring & text )
{
}

void FPlanXmlImporter::on_fatal_error( const Glib::ustring & text )
{
}

class XmlWriter {
        public:
                XmlWriter(void);
                ~XmlWriter();
                void open(const Glib::ustring& fn, bool compress = false);
                void write(const FPlanRoute& route);
                void close(void);


        private:
                xmlTextWriterPtr writer;
};

XmlWriter::XmlWriter(void)
        : writer(0)
{
}

void XmlWriter::close(void)
{
        if (writer) {
                int rc = xmlTextWriterEndElement(writer);
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterEndElement");
                rc = xmlTextWriterEndDocument(writer);
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterEndDocument");
                xmlFreeTextWriter(writer);
        }
        writer = 0;
}

void XmlWriter::open(const Glib::ustring & fn, bool compress)
{
        close();
        writer = xmlNewTextWriterFilename(fn.c_str(), compress);
        int rc = xmlTextWriterStartDocument(writer, NULL, "UTF-8", NULL);
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterStartDocument");
        rc = xmlTextWriterStartElement(writer, (const xmlChar *)"PilotFlightPlans");
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterStartElement");
        rc = xmlTextWriterSetIndent(writer, 1);
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterSetIndent");

}

void XmlWriter::write(const FPlanRoute& route)
{
        int rc = xmlTextWriterStartElement(writer, (const xmlChar *)"NavFlightPlan");
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterStartElement");
        rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"TimeOffBlock", "%u", route.get_time_offblock());
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
        rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"SavedTimeOffBlock", "%u", route.get_save_time_offblock());
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
        rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"TimeOnBlock", "%u", route.get_time_onblock());
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
        rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"SavedTimeOnBlock", "%u", route.get_save_time_onblock());
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
        rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"CurrentWaypoint", "%u", route.get_curwpt());
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
        rc = xmlTextWriterWriteAttribute(writer, (const xmlChar *)"Note", (const xmlChar *)route.get_note().c_str());
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteAttribute");
        for (unsigned int nr = 0; nr < route.get_nrwpt(); nr++) {
                const FPlanWaypoint& wpt(route[nr]);
                rc = xmlTextWriterStartElement(writer, (const xmlChar *)"WayPoint");
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterStartElement");
                rc = xmlTextWriterWriteAttribute(writer, (const xmlChar *)"ICAO", (const xmlChar *)wpt.get_icao().c_str());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteAttribute");
                rc = xmlTextWriterWriteAttribute(writer, (const xmlChar *)"name", (const xmlChar *)wpt.get_name().c_str());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"Time", "%u", wpt.get_time());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"SavedTime", "%u", wpt.get_save_time());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"frequency", "%u", wpt.get_frequency());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"lon", "%d", wpt.get_lon());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"lat", "%d", wpt.get_lat());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"alt", "%d", wpt.get_altitude());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteFormatAttribute(writer, (const xmlChar *)"flags", "%u", wpt.get_flags());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteFormatAttribute");
                rc = xmlTextWriterWriteAttribute(writer, (const xmlChar *)"Note", (const xmlChar *)wpt.get_note().c_str());
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterWriteAttribute");
                rc = xmlTextWriterEndElement(writer);
                if (rc < 0)
                        throw std::runtime_error("XmlWriter: Error at xmlTextWriterEndElement");
        }
        rc = xmlTextWriterEndElement(writer);
        if (rc < 0)
                throw std::runtime_error("XmlWriter: Error at xmlTextWriterEndElement");
}

XmlWriter::~XmlWriter()
{
        //close();
        if (writer)
                xmlFreeTextWriter(writer);
}

int main(int argc, char *argv[])
{
        static struct option long_options[] = {
                { "import", optional_argument, 0, 'i' },
                { "export", optional_argument, 0, 'e' },
                { "clear", no_argument, 0, 'c' },
                {0, 0, 0, 0}
        };
        Glib::ustring filename;
        int mode(0), doclear(0), c, err(0);

        while ((c = getopt_long(argc, argv, "i:e:c", long_options, 0)) != EOF) {
                switch (c) {
                        case 'c':
                                doclear = 1;
                                break;

                        case 'i':
                                if (optarg)
                                        filename = optarg;
                                mode = 1;
                                break;

                        case 'e':
                                if (optarg)
                                        filename = optarg;
                                mode = 2;
                                break;

                        default:
                                err++;
                                break;
                }
        }
        if (err) {
                std::cerr << "usage: vfrnavfplan [-i <fn>] [-e <fn>]" << std::endl;
                return 1;
        }
        FPlan db;
        if (doclear)
                db.delete_database();
        switch (mode) {
                case 1:
                {
                        try {
                                FPlanXmlImporter parser(db);
                                parser.set_substitute_entities(true);
                                parser.parse_file(filename);
                        } catch (const xmlpp::exception& ex) {
                                std::cerr << "libxml++ exception: " << ex.what() << std::endl;
                        } catch (const std::exception& ex) {
                                std::cerr << "exception: " << ex.what() << std::endl;
                        }
                        break;
                }

                case 2:
                {
                        try {
                                XmlWriter xw;
                                xw.open(filename, false);
                                FPlanRoute route(db);
                                bool work = route.load_first_fp();
                                while (work) {
                                        std::cerr << "writing route id " << route.get_id() << std::endl;
                                        xw.write(route);
                                        work = route.load_next_fp();
                                }
                                xw.close();
                        } catch (const std::exception& ex) {
                                std::cerr << "exception: " << ex.what() << std::endl;
                        }
                        break;
                }

                default:
                {
                        std::cerr << "must select import or export" << std::endl;
                        return 3;
                }
        }
        return 0;
}
