//
// C++ Implementation: dbobj
//
// Description: Database Objects
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include <limits>
#include <cstring>
#include <sstream>
#include <locale>

#include <zlib.h>

#include <sqlite3.h>

#include "dbobj.h"
#include "fplan.h"

Glib::ustring Conversions::time_str(time_t t)
{
        struct tm tm;
        if (!gmtime_r(&t, &tm))
                return "?";
        char buf[16];
        strftime(buf, sizeof(buf), "%H:%M:%S", &tm);
        buf[sizeof(buf)-1] = 0;
        return buf;
}

Glib::ustring Conversions::time_str(const Glib::ustring& fmt, time_t t)
{
        struct tm tm;
        if (!gmtime_r(&t, &tm))
                return "?";
        char buf[32];
        strftime(buf, sizeof(buf), fmt.c_str(), &tm);
        buf[sizeof(buf)-1] = 0;
        return buf;
}

time_t Conversions::time_parse(const Glib::ustring & str)
{
        time_t t(0);
        struct tm tm;
        memset(&tm, 0, sizeof(tm));
        //std::cerr << "Conversions::time_parse: text " << str << std::endl;
        if (sscanf(str.c_str(), "%u-%u-%u %u:%u:%u", &tm.tm_year, &tm.tm_mon, &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec) >= 3) {
                //std::cerr << "Conversions::time_parse: after sscanf " << tm.tm_year << '-' << tm.tm_mon << '-' << tm.tm_mday
                //          << ' ' << tm.tm_hour << ':' << tm.tm_min << ':' << tm.tm_sec << std::endl;
                tm.tm_year -= 1900;
                tm.tm_mon--;
                time_t tm0(mktime(&tm));
                gmtime_r(&tm0, &tm);
                time_t tm1(mktime(&tm));
                t = 2 * tm0 - tm1;
        }
        //std::cerr << "Conversions::time_parse: text " << str << " val " << t << std::endl;
        return t;
}

Glib::ustring Conversions::timefrac8_str(uint64_t t)
{
        struct tm tm;
        time_t t1(t >> 8);
        if (!gmtime_r(&t1, &tm))
                return "?";
        char buf[32];
        strftime(buf, sizeof(buf), "%H:%M:%S", &tm);
        buf[sizeof(buf)-1] = 0;
        uint32_t t2(((t & 0xff) * 1000 + 0x80) >> 8);
        size_t blen(strlen(buf));
        snprintf(buf + blen, sizeof(buf) - blen, ".%03u", t2);
        buf[sizeof(buf)-1] = 0;
        return buf;
}

Glib::ustring Conversions::timefrac8_str(const Glib::ustring & fmt, uint64_t t)
{
        struct tm tm;
        time_t t1(t >> 8);
        if (!gmtime_r(&t1, &tm))
                return "?";
        char buf[64];
        strftime(buf, sizeof(buf), fmt.c_str(), &tm);
        buf[sizeof(buf)-1] = 0;
        Glib::ustring buf1(buf);
        char buf2[8];
        uint32_t t2(((t & 0xff) * 1000 + 0x80) >> 8);
        snprintf(buf2, sizeof(buf2), "%03u", t2);
        Glib::ustring::size_type p;
        while ((p = buf1.find("%f")) != Glib::ustring::npos)
                buf1.replace(p, 2, buf2);
        return buf1;
}

uint64_t Conversions::timefrac8_parse(const Glib::ustring & str)
{
        uint64_t t(0);
        struct tm tm;
        memset(&tm, 0, sizeof(tm));
        unsigned int tfrac(0);
        //std::cerr << "Conversions::timefrac8_parse: text " << str << std::endl;
        if (sscanf(str.c_str(), "%u-%u-%u %u:%u:%u.%u", &tm.tm_year, &tm.tm_mon, &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec, &tfrac) >= 3) {
                //std::cerr << "Conversions::timefrac8_parse: after sscanf " << tm.tm_year << '-' << tm.tm_mon << '-' << tm.tm_mday
                //          << ' ' << tm.tm_hour << ':' << tm.tm_min << ':' << tm.tm_sec << '.' << tfrac << std::endl;
                tm.tm_year -= 1900;
                tm.tm_mon--;
                time_t tm0(mktime(&tm));
                gmtime_r(&tm0, &tm);
                time_t tm1(mktime(&tm));
                t = 2 * tm0 - tm1;
                t <<= 8;
                t += std::min((tfrac * 256 + 500) / 1000, 0xFFU);
        }
        //std::cerr << "Conversions::timefrac8_parse: text " << str << " val " << t << std::endl;
        return t;
}

Glib::ustring Conversions::freq_str(uint32_t f)
{
        char buf[16];
        if (f >= 1000000 && f <= 999999999) {
                f = (f + 500) / 1000;
                uint32_t f1 = f / 1000;
                f -= f1 * 1000;
                const std::numpunct<char>& np = std::use_facet<std::numpunct<char> >(std::locale());
                snprintf(buf, sizeof(buf), "%03u%c%03u", f1, np.decimal_point(), f);
                return buf;
        }
        if (f >= 1000 && f <= 999999) {
                f = (f + 500) / 1000;
                snprintf(buf, sizeof(buf), "%03u", f);
                return buf;
        }
        if (!f)
                return "";
        snprintf(buf, sizeof(buf), "%uHz", f);
        return buf;
}

Glib::ustring Conversions::alt_str(int32_t a, uint16_t flg)
{
        char buf[16];
        if (flg & FPlanWaypoint::altflag_standard) {
                a = (a + 50) / 100;
                snprintf(buf, sizeof(buf), "FL %d", a);
                return buf;
        }
        snprintf(buf, sizeof(buf), "%d", a);
        return buf;
}

Glib::ustring Conversions::dist_str(float d)
{
        char buf[16];
        snprintf(buf, sizeof(buf), "%5.1f", d);
        return buf;
}

Glib::ustring Conversions::track_str(float t)
{
        char buf[16];
        int tt = (int)(t);
        tt %= 360;
        if (tt < 0)
                tt += 360;
        snprintf(buf, sizeof(buf), "%03d", tt);
        return buf;
}

Glib::ustring Conversions::velocity_str(float v)
{
        char buf[16];
        snprintf(buf, sizeof(buf), "%5.1f", v);
        return buf;
}

Glib::ustring Conversions::verticalspeed_str(float v)
{
        char buf[16];
        snprintf(buf, sizeof(buf), "%4.0f", v);
        return buf;
}

uint32_t Conversions::unsigned_feet_parse(const Glib::ustring & str, bool *valid)
{
        char *cp;
        double v = strtod(str.c_str(), &cp);
        if (!strcmp(cp, "m")) {
                v *= Point::m_to_ft;
                cp++;
        } else if (!strcmp(cp, "km")) {
                v *= Point::m_to_ft * 1000.0;
                cp += 2;
        } else if (!strcmp(cp, "nmi") || !strcmp(cp, "nm")) {
                v *= Point::m_to_ft * 1000.0 / Point::km_to_nmi;
                cp += 2;
                if (*cp == 'i')
                        cp++;
        }
        if (valid)
                *valid = ((*cp == ' ') || (!*cp)) && !str.empty();
        return Point::round<int,double>(v);
}

int32_t Conversions::signed_feet_parse(const Glib::ustring & str, bool *valid)
{
        char *cp;
        double v = strtod(str.c_str(), &cp);
        if (!strcmp(cp, "m")) {
                v *= Point::m_to_ft;
                cp++;
        } else if (!strcmp(cp, "km")) {
                v *= Point::m_to_ft * 1000.0;
                cp += 2;
        } else if (!strcmp(cp, "nmi") || !strcmp(cp, "nm")) {
                v *= Point::m_to_ft * 1000.0 / Point::km_to_nmi;
                cp += 2;
                if (*cp == 'i')
                        cp++;
        }
        if (valid)
                *valid = ((*cp == ' ') || (!*cp)) && !str.empty();
        return Point::round<int,double>(v);
}

std::ostream & operator <<(std::ostream & os, const DbBaseElements::ElementAddress & addr)
{
        os << addr.get_id();
        switch (addr.get_table()) {
                case DbBaseElements::ElementAddress::table_main:
                        os << 'M';
                        break;

                case DbBaseElements::ElementAddress::table_aux:
                        os << 'A';
                        break;

                default:
                        os << '(' << (int)addr.get_table() << ')';
                        break;
        }
        return os;
}

DbBaseElements::TileNumber::tile_t DbBaseElements::TileNumber::to_tilenumber(const Point & pt)
{
        Point::coord_t lat(pt.get_lat());
        Point::coord_t lon(pt.get_lon() + lon_offset);
        tile_t ret = 1;
        for (unsigned int i = 0; i < tile_bits; i += 2) {
                ret <<= 2;
                ret |= ((lon >> (8*sizeof(lon)-2)) & 2) | ((lat >> (8*sizeof(lat)-1)) & 1);
                lon <<= 1;
                lat <<= 1;
        }
        return ret;
}

DbBaseElements::TileNumber::tile_t DbBaseElements::TileNumber::to_tilenumber(const Rect & r)
{
        tile_t t1(to_tilenumber(r.get_southwest())), t2(to_tilenumber(r.get_northeast()));
        while (t1 != t2) {
                t1 >>= 1;
                t2 >>= 1;
        }
        return t1;
}

DbBaseElements::TileNumber::tilerange_t DbBaseElements::TileNumber::to_tilerange(const Rect & r)
{
        tilerange_t t(to_tilenumber(r.get_southwest()), to_tilenumber(r.get_northeast()));
        unsigned int sh(0);
        while (t.first != t.second) {
                t.first >>= 1;
                t.second >>= 1;
                sh++;
        }
        t.first <<= sh;
        t.second <<= sh;
        t.second |= (1 << sh) - 1;
        return t;
}

DbBaseElements::TileNumber::Iterator::Iterator(const Rect & r)
        : m_tile1(to_tilenumber(r.get_southwest())),
          m_tile2(to_tilenumber(r.get_northeast())),
          m_bits(0)
{
        m_tile = m_tile1;
}

DbBaseElements::TileNumber::tile_t DbBaseElements::TileNumber::Iterator::operator ( )(void) const
{
        return m_tile >> m_bits;
}

bool DbBaseElements::TileNumber::Iterator::next(void)
{
        tile_t tile_mask = (1 << tile_bits) - 1;
        tile_t lat_mask = 0x55555555 & tile_mask;
        tile_t lon_mask = 0xAAAAAAAA & tile_mask;
        if ((m_tile ^ m_tile2) & lon_mask) {
                // lon != lon2: increment lon
                m_tile = (m_tile & ~lon_mask) | (((m_tile & lon_mask) + ((3 << m_bits) & lon_mask) + ~lon_mask) & lon_mask);
                return m_bits <= tile_bits;
        }
        // lon = lon1
        m_tile = (m_tile & ~lon_mask) | (m_tile1 & lon_mask);
        if ((m_tile ^ m_tile2) & lat_mask) {
                // lat != lat2: increment lat
                m_tile = (m_tile & ~lat_mask) | (((m_tile & lat_mask) + ((3 << m_bits) & lat_mask) + ~lat_mask) & lat_mask);
                return m_bits <= tile_bits;
        }
        m_bits++;
        m_tile1 &= ~((1 << m_bits) - 1);
        m_tile2 &= ~((1 << m_bits) - 1);
        m_tile = m_tile1;
        return m_bits <= tile_bits;
}

DbBaseCommon::DbBaseCommon(void)
        : m_open(openstate_closed), m_has_rtree(false), m_has_aux_rtree(false)
{
}

DbBaseCommon::~DbBaseCommon(void)
{
        close();
}

void DbBaseCommon::close(void)
{
        if (m_open != openstate_closed)
                m_db.close();
        m_open = openstate_closed;
}

void DbBaseCommon::detach(void)
{
        if (m_open != openstate_auxopen)
                return;
        sqlite3x::sqlite3_command cmd(m_db, "DETACH DATABASE aux;");
        cmd.executenonquery();
        m_open = openstate_mainopen;
}

void DbBaseCommon::dbfunc_simpledist( sqlite3_context * ctxt, int , sqlite3_value ** values )
{
        Point pt1(sqlite3_value_int(values[0]), sqlite3_value_int(values[1]));
        Point pt2(sqlite3_value_int(values[2]), sqlite3_value_int(values[3]));
        sqlite3_result_int64(ctxt, pt1.simple_distance_rel(pt2));
}

void DbBaseCommon::dbfunc_simplerectdist(sqlite3_context * ctxt, int , sqlite3_value ** values)
{
        Rect r(Point(sqlite3_value_int(values[0]), sqlite3_value_int(values[1])),
               Point(sqlite3_value_int(values[2]), sqlite3_value_int(values[3])));
        Point pt(sqlite3_value_int(values[4]), sqlite3_value_int(values[5]));
        sqlite3_result_int64(ctxt, r.simple_distance_rel(pt));
}

void DbBaseCommon::dbfunc_upperbound(sqlite3_context * ctxt, int , sqlite3_value ** values)
{
        Glib::ustring s((const char *)sqlite3_value_text(values[0]));
        //std::cerr << "dbfunc_upperbound: in \"" << s << "\"" << std::endl;
        if (!s.empty())
                s.replace(--s.end(), s.end(), 1, (*(--s.end())) + 1);
        //std::cerr << "dbfunc_upperbound: out \"" << s << "\"" << std::endl;
        sqlite3_result_text(ctxt, s.data(), s.bytes(), SQLITE_TRANSIENT);
}

Glib::ustring DbBaseCommon::quote_text_for_like(const Glib::ustring & s, char e)
{
        Glib::ustring r;
        for (Glib::ustring::const_iterator si(s.begin()), se(s.end()); si != se; si++) {
                if (*si == '%' || *si == '_' || (e && *si == e))
                        r.push_back(e ? e : '%');
                r.push_back(*si);
        }
        return r;
}

void DbBaseCommon::sync_off(void)
{
        sqlite3x::sqlite3_command cmd(m_db, "PRAGMA synchronous = OFF;");
        cmd.executenonquery();
}

void DbBaseCommon::analyze(void)
{
        sqlite3x::sqlite3_command cmd(m_db, "ANALYZE main;");
        cmd.executenonquery();
}

void DbBaseCommon::vacuum(void)
{
        sqlite3x::sqlite3_command cmd(m_db, "VACUUM;");
        cmd.executenonquery();
}

unsigned int DbBaseCommon::query_cache_size(void)
{
        sqlite3x::sqlite3_command cmd(m_db, "PRAGMA cache_size;");
        return cmd.executeint();
}

void DbBaseCommon::set_cache_size(unsigned int cs)
{
        sqlite3x::sqlite3_command cmd(m_db, "PRAGMA cache_size = ?;");
        cmd.bind(1, (long long int)cs);
        cmd.executenonquery();
}

void DbBaseCommon::interrupt(void)
{
        sqlite3_interrupt(m_db.db());
}

template<class T> DbBase<T>::DbBase()
        : DbBaseCommon()
{
}

template<class T> void DbBase<T>::open(const Glib::ustring & path)
{
        close();
        Glib::ustring dbname(path);
        if (dbname.empty())
                dbname = FPlan::get_userdbpath();
        dbname = Glib::build_filename(dbname, database_file_name);
        //std::cerr << "DbBase: opening database " << dbname << std::endl;
        m_db.open(dbname);
        if (sqlite3_create_function(m_db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("DbBase: cannot add user function simpledist");
        if (sqlite3_create_function(m_db.db(), "simplerectdist", 6, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simplerectdist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("DbBase: cannot add user function simplerectdist");
        if (sqlite3_create_function(m_db.db(), "upperbound", 1, SQLITE_UTF8, 0, &dbfunc_upperbound, 0, 0) != SQLITE_OK)
                throw std::runtime_error("DbBase: cannot add user function upperbound");
        create_tables();
        create_indices();
        m_has_rtree = m_has_aux_rtree = false;
        m_open = openstate_mainopen;
        if (area_data) {
                try {
                        sqlite3x::sqlite3_command cmd1(m_db, std::string("SELECT ID FROM ") + main_table_name + " ORDER BY ID LIMIT 1;");
                        int nr1 = cmd1.executeint();
                        sqlite3x::sqlite3_command cmd2(m_db, std::string("SELECT ID FROM ") + main_table_name + "_rtree WHERE ID=?;");
                        cmd2.bind(1, nr1);
                        int nr2 = cmd2.executeint();
                        m_has_rtree = (nr2 == nr1);
                } catch (...) {
                }
        }
        if (true)
                std::cerr << main_table_name << ": rtree: " << (m_has_rtree ? "yes" : "no") << std::endl;
}

template<class T> void DbBase<T>::attach(const Glib::ustring & path)
{
        if (m_open == openstate_auxopen)
                detach();
        if (m_open != openstate_mainopen)
                return;
        Glib::ustring dbname(path);
        if (dbname.empty())
                dbname = PACKAGE_DATA_DIR;
        dbname += "/";
        dbname += database_file_name;
        {
                sqlite3x::sqlite3_command cmd(m_db, "ATTACH DATABASE ?1 AS aux;");
                cmd.bind(1, dbname);
                cmd.executenonquery();
        }
        m_has_aux_rtree = false;
        if (area_data) {
                try {
                        sqlite3x::sqlite3_command cmd1(m_db, std::string("SELECT ID FROM aux.") + main_table_name + " ORDER BY ID LIMIT 1;");
                        int nr1 = cmd1.executeint();
                        sqlite3x::sqlite3_command cmd2(m_db, std::string("SELECT ID FROM aux.") + main_table_name + "_rtree WHERE ID=?;");
                        cmd2.bind(1, nr1);
                        int nr2 = cmd2.executeint();
                        m_has_aux_rtree = (nr2 == nr1);
                } catch (...) {
                }
        }
        m_open = openstate_auxopen;
        if (true)
                std::cerr << "aux." << main_table_name << ": rtree: " << (m_has_aux_rtree ? "yes" : "no") << std::endl;
}

template<class T> bool DbBase<T>::is_dbfile_exists(const Glib::ustring & path)
{
        Glib::ustring dbname(path);
        if (dbname.empty())
                dbname = FPlan::get_userdbpath();
        dbname = Glib::build_filename(dbname, database_file_name);
        return Glib::file_test(dbname, Glib::FILE_TEST_IS_REGULAR);
}

template<class T> void DbBase<T>::take(sqlite3 * dbh)
{
        close();
        m_db.take(dbh);
        if (!m_db.db())
                return;
        if (sqlite3_create_function(m_db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("DbBase: cannot add user function simpledist");
        if (sqlite3_create_function(m_db.db(), "simplerectdist", 6, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simplerectdist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("DbBase: cannot add user function simplerectdist");
        if (sqlite3_create_function(m_db.db(), "upperbound", 1, SQLITE_UTF8, 0, &dbfunc_upperbound, 0, 0) != SQLITE_OK)
                throw std::runtime_error("DbBase: cannot add user function upperbound");
        create_tables();
        create_indices();
        m_has_rtree = m_has_aux_rtree = false;
        m_open = openstate_mainopen;
        if (area_data) {
                try {
                        sqlite3x::sqlite3_command cmd1(m_db, std::string("SELECT ID FROM ") + main_table_name + " ORDER BY ID LIMIT 1;");
                        int nr1 = cmd1.executeint();
                        sqlite3x::sqlite3_command cmd2(m_db, std::string("SELECT ID FROM ") + main_table_name + "_rtree WHERE ID=?;");
                        cmd2.bind(1, nr1);
                        int nr2 = cmd2.executeint();
                        m_has_rtree = (nr2 == nr1);
                } catch (...) {
                }
        }
        if (true)
                std::cerr << main_table_name << ": rtree: " << (m_has_rtree ? "yes" : "no") << std::endl;
}

template<class T> void DbBase<T>::purgedb(void)
{
        drop_indices();
        drop_tables();
        create_tables();
        create_indices();
}

template <class T> typename DbBase<T>::element_t DbBase<T>::operator()( int64_t id, table_t tbl, unsigned int loadsubtables )
{
        element_t e;
        std::string qs;
        if (tbl == element_t::table_aux) {
                if (m_open != openstate_auxopen)
                        return e;
                qs = "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name + " WHERE ID=? " +
                     " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted);";
        } else {
                qs = "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE ID=?;";
        }
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, (long long int)id);
        sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
        e.load(cursor, m_db, loadsubtables);
        return e;
}

template<class T> void DbBase<T>::loadfirst( element_t & e, bool include_aux, unsigned int loadsubtables )
{
        e = element_t();
        {
                sqlite3x::sqlite3_command cmd(m_db, "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " ORDER BY ID ASC LIMIT 1;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                e.load(cursor, m_db, loadsubtables);
        }
        if (e.is_valid() || m_open != openstate_auxopen || !include_aux)
                return;
        {
                sqlite3x::sqlite3_command cmd(m_db, "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name +
                                              " WHERE " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) ORDER BY ID ASC LIMIT 1;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                e.load(cursor, m_db, loadsubtables);
        }
}

template<class T> void DbBase<T>::loadnext( element_t & e, bool include_aux, unsigned int loadsubtables )
{
        uint64_t id(e.get_id());
        table_t tbl(e.get_table());
        e = element_t();
        if (tbl == element_t::table_aux) {
                if (m_open != openstate_auxopen)
                        return;
                sqlite3x::sqlite3_command cmd(m_db, "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name + " WHERE ID > ? " +
                                              " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) ORDER BY ID ASC LIMIT 1;");
                cmd.bind(1, (long long int)id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                e.load(cursor, m_db, loadsubtables);
                return;
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE ID > ? ORDER BY ID ASC LIMIT 1;");
                cmd.bind(1, (long long int)id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                e.load(cursor, m_db, loadsubtables);
        }
        if (e.is_valid() || m_open != openstate_auxopen || !include_aux)
                return;
        {
                sqlite3x::sqlite3_command cmd(m_db, "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name +
                                                    " WHERE " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) ORDER BY ID ASC LIMIT 1;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                e.load(cursor, m_db, loadsubtables);
        }
}

template <class T> void DbBase<T>::for_each(ForEach& cb, bool include_aux, unsigned int loadsubtables)
{
        if (!include_aux) {
                std::string qs("SELECT " + std::string(delete_field) + " FROM " + main_table_name + "_deleted ORDER BY " + delete_field + ";");
                sqlite3x::sqlite3_command cmd(m_db, qs);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        if (!cb(cursor.getstring(0)))
                                break;
        }
        std::string qs;
        if (include_aux && m_open == openstate_auxopen) {
                qs += "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name +
                      " WHERE " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " ORDER BY " + order_field;
        sqlite3x::sqlite3_command cmd(m_db, qs);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                element_t e;
                e.load(cursor, m_db, loadsubtables);
                if (!e.is_valid())
                        break;
                if (!cb(e))
                        break;
        }
}

template <class T> void DbBase<T>::for_each_by_rect(ForEach & cb, const Rect & r, bool include_aux, unsigned int loadsubtables)
{
        std::string qs;
        if (include_aux && m_open == openstate_auxopen) {
                qs += "SELECT " + get_area_select_string(true, r) +
                      " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + get_area_select_string(false, r);
        if (order_field) {
                qs += " ORDER BY ";
                qs += order_field;
        }
        qs += ";";
        if (false) {
                //std::cerr << "sql: " << qs << std::endl;
                sqlite3x::sqlite3_command cmd(m_db, "EXPLAIN " + qs);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step()) {
                        std::cerr << "vm:";
                        for (int col = 0; col < cursor.colcount(); col++) {
                                std::cerr << " " << cursor.getstring(col);
                        }
                        std::cerr << std::endl;
                }
        }
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, (long long int)r.get_west());
        cmd.bind(2, (long long int)r.get_south());
        cmd.bind(3, (long long int)r.get_east_unwrapped());
        cmd.bind(4, (long long int)r.get_north());
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                element_t e;
                e.load(cursor, m_db, loadsubtables);
                if (!e.is_valid())
                        break;
                if (!cb(e))
                        break;
        }
}

template <class T> typename DbBase<T>::elementvector_t DbBase<T>::find( const std::string & fldname, const std::string & pattern, char escape, comp_t comp, unsigned int limit, unsigned int loadsubtables )
{
        std::string qsfld;
        Glib::ustring pat(pattern);
        switch (comp) {
                default:
                case comp_startswith:
                        escape = 0;
                        qsfld += std::string("(") + fldname + ">=?1 COLLATE NOCASE) AND (" + fldname + "<upperbound(?1) COLLATE NOCASE)";
                        break;

                case comp_exact:
                        escape = 0;
                        qsfld += std::string("(") + fldname + "=?1 COLLATE NOCASE)";
                        break;

                case comp_exact_casesensitive:
                        escape = 0;
                        qsfld += std::string("(") + fldname + "=?1)";
                        break;

                case comp_contains:
                        escape = '!';
                        pat = std::string("%") + quote_text_for_like(pat, escape) + "%";
                        // fall through

                case comp_like:
                        qsfld += std::string("(") + fldname + " LIKE ?1";
                        if (escape)
                                qsfld += " ESCAPE ?2";
                        qsfld += ')';
                        break;
        }
        std::string qs;
        if (m_open == openstate_auxopen) {
                qs += "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name + " WHERE " + qsfld +
                      " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE " + qsfld;
        if (order_field) {
                qs += " ORDER BY ";
                qs += order_field;
        }
        if (limit)
                qs += " LIMIT ?3";
        qs += ";";
        //std::cerr << "sql: " << qs << std::endl;
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, pat);
        if (escape)
                cmd.bind(2, std::string(1, escape));
        if (limit)
                cmd.bind(3, (long long int)limit);
        std::vector<T> ret;
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                ret.push_back(T());
                ret.back().load(cursor, m_db, loadsubtables);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        return ret;
}

template <class T> typename DbBase<T>::elementvector_t DbBase<T>::find_by_text( const std::string & pattern, char escape, comp_t comp,  unsigned int limit, unsigned int loadsubtables )
{
        std::string qsfld;
        Glib::ustring pat(pattern);
        switch (comp) {
                default:
                case comp_startswith:
                        escape = 0;
                        break;

                case comp_contains:
                        escape = '!';
                        pat = std::string("%") + quote_text_for_like(pat, escape) + "%";
                        break;

                case comp_exact:
                        escape = 0;
                        break;

                case comp_like:
                        break;
        }
        {
                bool needor(false);
                for (const char **fld = element_t::db_text_fields; *fld; fld++) {
                        if (needor)
                                qsfld += " OR ";
                        const char *fldname(*fld);
                        switch (comp) {
                                default:
                                case comp_startswith:
                                        qsfld += std::string("(") + fldname + ">=?1 COLLATE NOCASE) AND (" + fldname + "<upperbound(?1) COLLATE NOCASE)";
                                        break;

                                case comp_exact:
                                        qsfld += std::string("(") + fldname + "=?1 COLLATE NOCASE)";
                                        break;

                                case comp_exact_casesensitive:
                                        qsfld += std::string("(") + fldname + "=?1)";
                                        break;

                                case comp_contains:
                                case comp_like:
                                        qsfld += std::string("(") + fldname + " LIKE ?1";
                                        if (escape)
                                                qsfld += " ESCAPE ?2";
                                        qsfld += ')';
                                        break;
                        }
                        needor = true;
                }
        }
        std::string qs;
        if (m_open == openstate_auxopen) {
                qs += "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name + " WHERE (" + qsfld +
                      ") AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE (" + qsfld + ')';
        if (order_field) {
                qs += " ORDER BY ";
                qs += order_field;
        }
        if (limit)
                qs += " LIMIT ?3";
        qs += ";";
        //std::cerr << "sql: " << qs << " pat: " << pat << std::endl;
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, pat);
        if (escape)
                cmd.bind(2, std::string(1, escape));
        if (limit)
                cmd.bind(3, (long long int)limit);
        std::vector<T> ret;
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                ret.push_back(T());
                ret.back().load(cursor, m_db, loadsubtables);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        return ret;
}

template <class T> typename DbBase<T>::elementvector_t DbBase<T>::find_by_time( time_t fromtime, time_t totime, unsigned int limit, unsigned int loadsubtables )
{
        std::string qsfld;
        {
                bool needor(false);
                for (const char **fld = element_t::db_time_fields; *fld; fld++) {
                        if (needor)
                                qsfld += " OR ";
                        qsfld += std::string("(") + (*fld) + " BETWEEN ?1 AND ?2)";
                        needor = true;
                }
        }
        std::string qs;
        if (m_open == openstate_auxopen) {
                qs += "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name + " WHERE (" + qsfld +
                      ") AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE (" + qsfld + ')';
        if (order_field) {
                qs += " ORDER BY ";
                qs += order_field;
        }
        if (limit)
                qs += " LIMIT ?3";
        qs += ";";
        //std::cerr << "sql: " << qs << std::endl;
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, (long long int)fromtime);
        cmd.bind(2, (long long int)totime);
        if (limit)
                cmd.bind(3, (long long int)limit);
        std::vector<T> ret;
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                ret.push_back(T());
                ret.back().load(cursor, m_db, loadsubtables);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        return ret;
}

template <class T> typename DbBase<T>::elementvector_t DbBase<T>::loadid( uint64_t startid, table_t table, const std::string & order, unsigned int limit, unsigned int loadsubtables )
{
        std::string qs;
        std::vector<T> ret;
        if (table == element_t::table_aux) {
                if (m_open != openstate_auxopen)
                        return ret;
                qs = "SELECT " + std::string(element_t::db_aux_query_string) + " FROM aux." + main_table_name + " WHERE ID>=?1" +
                     " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted)";
        } else {
                qs = "SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE ID>=?1";
        }
        if (!order.empty())
                qs += " ORDER BY " + order;
        if (limit)
                qs += " LIMIT ?2";
        qs += ";";
        //std::cerr << "sql: " << qs << std::endl;
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, (long long int)startid);
        if (limit)
                cmd.bind(2, (long long int)limit);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                ret.push_back(T());
                ret.back().load(cursor, m_db, loadsubtables);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        return ret;
}

template<class T> std::string DbBase<T>::get_area_select_string(bool auxtable, const Rect& r, const char *sortcol)
{
        std::ostringstream oss;
        if (auxtable)
                oss << element_t::db_aux_query_string;
        else
                oss << element_t::db_query_string;
        if (sortcol)
                oss << ',' << sortcol;
        oss << " FROM ";
        std::string table_name(main_table_name);
        if (auxtable)
                table_name = "aux." + table_name;
        oss << table_name;
        bool rtree(auxtable ? m_has_aux_rtree : m_has_rtree);
        oss << " WHERE ";
        if (rtree) {
                oss << "ID IN (SELECT ID FROM " << table_name << "_rtree WHERE "
                                << table_name << "_rtree.NELAT >= ?2 AND " << table_name << "_rtree.SWLAT <= ?4 AND (("
                                //<< table_name << "_rtree.SWLON <= ?3-4294967296 AND " << table_name << "_rtree.NELON >= ?1-4294967296) OR ("
                                //<< table_name << "_rtree.SWLON <= ?3 AND " << table_name << "_rtree.NELON >= ?1) OR ("
                                //<< table_name << "_rtree.SWLON <= ?3+4294967296 AND " << table_name << "_rtree.NELON >= ?1+4294967296))";
                                << table_name << "_rtree.SWLON <= ?3 AND " << table_name << "_rtree.NELON >= ?1)))";
                                return oss.str();
        }
        if (false) {
                oss << '(';
                typename element_t::tilerange_t tr(element_t::TileNumber::to_tilerange(r));
                bool needor(false);
                for (typename element_t::TileNumber::tile_t mask = (1 << element_t::TileNumber::tile_bits) - 1; mask; mask >>= 1, tr.first >>= 1, tr.second >>= 1) {
                        if (needor)
                                oss << " OR ";
                        needor = true;
                        if (tr.second < tr.first) {
                                oss << "(TILE BETWEEN " << tr.first << " AND " << (tr.first | mask) 
                                    << ") OR (TILE BETWEEN " << (tr.second & ~mask) << " AND " << tr.second << ')';
                        } else {
                                oss << "(TILE BETWEEN " << tr.first << " AND " << tr.second << ')';
                        }
                }
                oss << ") AND ";
        } else if (false) {
                oss << "TILE IN (";
                typename element_t::tilerange_t tr(element_t::TileNumber::to_tilerange(r));
                bool needcomma(false);
                for (typename element_t::TileNumber::tile_t mask = (1 << element_t::TileNumber::tile_bits) - 1; mask; mask >>= 1, tr.first >>= 1, tr.second >>= 1) {
                        for (typename element_t::TileNumber::tile_t tile = tr.first;; tile = (tile & ~mask) | ((tile + 1) & mask)) {
                                if (needcomma)
                                        oss << ',';
                                needcomma = true;
                                oss << tile;
                                if (tile == tr.second)
                                        break;
                        }
                }
                oss << ") AND ";
        } else if (true) {
                oss << "TILE IN (";
                typename element_t::TileNumber::Iterator ti(r);
                bool needcomma(false);
                do {
                        if (needcomma)
                                oss << ',';
                        needcomma = true;
                        oss << ti();
                } while (ti.next());
                oss << ") AND ";
        }
        if (area_data) {
                if (r.get_north() + 0x40000000 < 0x40000000 - r.get_south())
                        oss << "(SWLAT <= ?4 AND NELAT >= ?2)";
                else
                        oss << "(NELAT >= ?2 AND SWLAT <= ?4)";
                oss << " AND ((SWLON <= ?3-4294967296 AND NELON >= ?1-4294967296) OR "
                             "(SWLON <= ?3 AND NELON >= ?1) OR "
                             "(SWLON <= ?3+4294967296 AND NELON >= ?1+4294967296))";
        } else {
                oss << "(LAT BETWEEN ?2 AND ?4) AND ((LON BETWEEN ?1-4294967296 AND ?3-4294967296) OR (LON BETWEEN ?1 AND ?3))";
        }
        return oss.str();
}

template <class T> typename DbBase<T>::elementvector_t DbBase<T>::find_by_rect( const Rect & r, unsigned int limit, unsigned int loadsubtables )
{
        //std::cerr << "find_by_rect: loadsubtables 0x" << std::hex << loadsubtables << std::endl;
        std::string qs;
        if (m_open == openstate_auxopen) {
                qs += "SELECT " + get_area_select_string(true, r) +
                      " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + get_area_select_string(false, r);
        if (order_field) {
                qs += " ORDER BY ";
                qs += order_field;
        }
        if (limit)
                qs += " LIMIT ?5";
        qs += ";";
        if (false) {
                std::cerr << "sql: " << qs << std::endl;
                sqlite3x::sqlite3_command cmd(m_db, "EXPLAIN " + qs);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step()) {
                        std::cerr << "vm:";
                        for (int col = 0; col < cursor.colcount(); col++) {
                                std::cerr << " " << cursor.getstring(col);
                        }
                        std::cerr << std::endl;
                }
        }
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, (long long int)r.get_west());
        cmd.bind(2, (long long int)r.get_south());
        cmd.bind(3, (long long int)r.get_east_unwrapped());
        cmd.bind(4, (long long int)r.get_north());
        if (limit)
                cmd.bind(5, (long long int)limit);
        //std::cerr << "sql: " << qs << " W " << r.get_west() << " S " << r.get_south() << " E " << r.get_east() << " N " << r.get_north() << std::endl;
        std::vector<T> ret;
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                ret.push_back(T());
                ret.back().load(cursor, m_db, loadsubtables);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        //std::cerr << "find_by_rect: " << main_table_name << " rect " << r << " retsize " << ret.size() << " qs " << qs << std::endl;
        return ret;
}

template <class T> typename DbBase<T>::elementvector_t DbBase<T>::find_nearest( const Point & pt, const Rect & r, unsigned int limit, unsigned int loadsubtables )
{
        const char *sortcol(area_data ? "(simplerectdist(SWLON,SWLAT,NELON,NELAT,?5,?6)) AS 'SORTCOL'" : "(simpledist(LON,LAT,?5,?6)) AS 'SORTCOL'");
        std::string qs;
        if (m_open == openstate_auxopen) {
                qs += "SELECT " + get_area_select_string(true, r, sortcol) +
                      " AND " + delete_field + " NOT IN (SELECT " + delete_field + " FROM " + main_table_name + "_deleted) UNION ";
        }
        qs += "SELECT " + get_area_select_string(false, r, sortcol);
        qs += " ORDER BY SORTCOL";
        if (limit)
                qs += " LIMIT ?7";
        qs += ";";
        if (false) {
                //std::cerr << "sql: " << qs << std::endl;
                sqlite3x::sqlite3_command cmd(m_db, "EXPLAIN " + qs);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step()) {
                        std::cerr << "vm:";
                        for (int col = 0; col < cursor.colcount(); col++) {
                                std::cerr << " " << cursor.getstring(col);
                        }
                        std::cerr << std::endl;
                }
        }
        sqlite3x::sqlite3_command cmd(m_db, qs);
        cmd.bind(1, (long long int)r.get_west());
        cmd.bind(2, (long long int)r.get_south());
        cmd.bind(3, (long long int)r.get_east_unwrapped());
        cmd.bind(4, (long long int)r.get_north());
        cmd.bind(5, (long long int)pt.get_lon());
        cmd.bind(6, (long long int)pt.get_lat());
        if (limit)
                cmd.bind(7, (long long int)limit);
        std::vector<T> ret;
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        if (false) {
                for (int i = 0; i < cursor.colcount(); i++)
                        std::cerr << "Column " << i << " Name " << cursor.getcolname(i) << std::endl;
        }
        for (;;) {
                ret.push_back(T());
                ret.back().load(cursor, m_db, loadsubtables);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        return ret;
}

template class DbBase<NavaidsDb::Navaid>;
template class DbBase<WaypointsDb::Waypoint>;
template class DbBase<AirwaysDb::Airway>;
template class DbBase<AirportsDb::Airport>;
template class DbBase<AirspacesDb::Airspace>;
template class DbBase<MapelementsDb::Mapelement>;
template class DbBase<TracksDb::Track>;
template class DbBase<LabelsDb::Label>;

const char *DbBaseElements::Navaid::db_query_string = "navaids.ID AS 'ID',0 AS 'TABLE',navaids.SRCID AS 'SRCID',navaids.ICAO,navaids.NAME,navaids.FREQ,navaids.LAT,navaids.LON,navaids.ELEVATION,navaids.RANGE,navaids.TYPE,navaids.LABELPLACEMENT,navaids.STATUS,navaids.MODIFIED";
const char *DbBaseElements::Navaid::db_aux_query_string = "aux.navaids.ID AS 'ID',1 AS 'TABLE',aux.navaids.SRCID AS 'SRCID',aux.navaids.ICAO,aux.navaids.NAME,aux.navaids.FREQ,aux.navaids.LAT,aux.navaids.LON,aux.navaids.ELEVATION,aux.navaids.RANGE,aux.navaids.TYPE,aux.navaids.LABELPLACEMENT,aux.navaids.STATUS,aux.navaids.MODIFIED";
const char *DbBaseElements::Navaid::db_text_fields[] = { "SRCID", "ICAO", "NAME", 0 };
const char *DbBaseElements::Navaid::db_time_fields[] = { "MODIFIED", 0 };

DbBaseElements::Navaid::Navaid(void)
        : m_id(0), m_coord(), m_elev(0), m_range(0), m_freq(0), m_sourceid(), m_icao(), m_name(),
          m_navaid_type(navaid_invalid), m_label_placement(label_off), m_inservice(false), m_table(table_main), m_modtime(0)
{
}

void DbBaseElements::Navaid::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_navaid_type = navaid_invalid;
        m_table = table_main;
        if (!cursor.step())
                return;
        m_id = cursor.getint(0);
        m_table = (table_t)cursor.getint(1);
        m_sourceid = cursor.getstring(2);
        m_icao = cursor.getstring(3);
        m_name = cursor.getstring(4);
        m_freq = cursor.getint(5);
        m_coord = Point(cursor.getint(7), cursor.getint(6));
        m_elev = cursor.getint(8);
        m_range = cursor.getint(9);
        m_navaid_type = (navaid_type_t)cursor.getint(10);
        m_label_placement = (label_placement_t)cursor.getint(11);
        m_inservice = cursor.getint(12);
        m_modtime = cursor.getint(13);
}

void DbBaseElements::Navaid::update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE navaids SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(get_coord()));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE navaids_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Navaid::save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO navaids_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM navaids;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO navaids (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO navaids_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_coord.get_lat());
                        cmd.bind(3, m_coord.get_lat());
                        cmd.bind(4, m_coord.get_lon());
                        cmd.bind(5, m_coord.get_lon());
                        cmd.executenonquery();
                }
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE navaids SET SRCID=?,ICAO=?,NAME=?,FREQ=?,LAT=?,LON=?,ELEVATION=?,RANGE=?,TYPE=?,LABELPLACEMENT=?,STATUS=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_icao);
                cmd.bind(3, m_name);
                cmd.bind(4, (long long int)m_freq);
                cmd.bind(5, m_coord.get_lat());
                cmd.bind(6, m_coord.get_lon());
                cmd.bind(7, m_elev);
                cmd.bind(8, m_range);
                cmd.bind(9, m_navaid_type);
                cmd.bind(10, m_label_placement);
                cmd.bind(11, m_inservice);
                cmd.bind(12, (long long int)m_modtime);
                cmd.bind(13, (int)TileNumber::to_tilenumber(m_coord));
                cmd.bind(14, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE navaids_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Navaid::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO navaids_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM navaids WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM navaids_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

Glib::ustring DbBaseElements::Navaid::get_icao_name(void) const
{
        Glib::ustring r(m_icao);
        if (!(m_icao.empty() || m_name.empty()))
                r += " ";
        return r + m_name;
}

Glib::ustring DbBaseElements::Navaid::get_navaid_typename( navaid_type_t nt )
{
        switch (nt) {
                default:
                        return "Invalid";

                case navaid_vor:
                        return "VOR";

                case navaid_vordme:
                        return "VORDME";

                case navaid_dme:
                        return "DME";

                case navaid_ndb:
                        return "NDB";
        }
}

Glib::ustring DbBaseElements::Navaid::get_label_placement_name( label_placement_t lp )
{
        switch (lp) {
                default:
                        return "Invalid";

                case label_n:
                        return "N";

                case label_e:
                        return "E";

                case label_s:
                        return "S";

                case label_w:
                        return "W";

                case label_center:
                        return "Center";

                case label_any:
                        return "Any";

                case label_off:
                        return "Off";
        }
}

template<> void DbBase<DbBaseElements::Navaid>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS navaids;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS navaids_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Navaid>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS navaids (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "ICAO TEXT,"
                                              "NAME TEXT,"
                                              "FREQ INTEGER,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "ELEVATION INTEGER,"
                                              "RANGE INTEGER,"
                                              "TYPE INTEGER,"
                                              "LABELPLACEMENT INTEGER,"
                                              "STATUS INTEGER,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS navaids_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Navaid>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_icao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_freq;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS navaids_lon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Navaid>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_srcid ON navaids(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_icao ON navaids(ICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_name ON navaids(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_freq ON navaids(FREQ);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_latlon ON navaids(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_modified ON navaids(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_tile ON navaids(TILE);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_lat ON navaids(LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS navaids_lon ON navaids(LON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Navaid>::main_table_name = "navaids";
template<> const char *DbBase<DbBaseElements::Navaid>::database_file_name = "navaids.db";
template<> const char *DbBase<DbBaseElements::Navaid>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Navaid>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Navaid>::area_data = false;

const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_n;
const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_e;
const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_s;
const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_w;
const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_center;
const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_any;
const DbBaseElements::Waypoint::label_placement_t DbBaseElements::Waypoint::label_off;

const char DbBaseElements::Waypoint::usage_invalid;
const char DbBaseElements::Waypoint::usage_highlevel;
const char DbBaseElements::Waypoint::usage_lowlevel;
const char DbBaseElements::Waypoint::usage_bothlevel;
const char DbBaseElements::Waypoint::usage_rnav;
const char DbBaseElements::Waypoint::usage_terminal;
const char DbBaseElements::Waypoint::usage_vfr;
const char DbBaseElements::Waypoint::usage_user;

const char *DbBaseElements::Waypoint::db_query_string = "waypoints.ID AS 'ID',0 AS 'TABLE',waypoints.SRCID AS 'SRCID',waypoints.ICAO,waypoints.NAME,waypoints.USAGE,waypoints.LAT,waypoints.LON,waypoints.LABELPLACEMENT,waypoints.MODIFIED";
const char *DbBaseElements::Waypoint::db_aux_query_string = "aux.waypoints.ID AS 'ID',1 AS 'TABLE',aux.waypoints.SRCID AS 'SRCID',aux.waypoints.ICAO,aux.waypoints.NAME,aux.waypoints.USAGE,aux.waypoints.LAT,aux.waypoints.LON,aux.waypoints.LABELPLACEMENT,aux.waypoints.MODIFIED";
const char *DbBaseElements::Waypoint::db_text_fields[] = { "SRCID", "ICAO", "NAME", 0 };
const char *DbBaseElements::Waypoint::db_time_fields[] = { "MODIFIED", 0 };

DbBaseElements::Waypoint::Waypoint(void)
        : m_id(0), m_coord(), m_sourceid(), m_icao(), m_name(), m_usage(0), m_label_placement(label_off), m_table(table_main), m_modtime(0)
{
}

void DbBaseElements::Waypoint::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_table = table_main;
        m_usage = 0;
        if (!cursor.step())
                return;
        m_id = cursor.getint(0);
        m_table = (table_t)cursor.getint(1);
        m_sourceid = cursor.getstring(2);
        m_icao = cursor.getstring(3);
        m_name = cursor.getstring(4);
        m_usage = cursor.getint(5);
        m_coord = Point(cursor.getint(7), cursor.getint(6));
        m_label_placement = (label_placement_t)cursor.getint(8);
        m_modtime = cursor.getint(9);
}

void DbBaseElements::Waypoint::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(get_coord()));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Waypoint::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO waypoints_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM waypoints;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO waypoints (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO waypoints_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_coord.get_lat());
                        cmd.bind(3, m_coord.get_lat());
                        cmd.bind(4, m_coord.get_lon());
                        cmd.bind(5, m_coord.get_lon());
                        cmd.executenonquery();
                }
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints SET SRCID=?,ICAO=?,NAME=?,USAGE=?,LAT=?,LON=?,LABELPLACEMENT=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_icao);
                cmd.bind(3, m_name);
                cmd.bind(4, (int)m_usage);
                cmd.bind(5, m_coord.get_lat());
                cmd.bind(6, m_coord.get_lon());
                cmd.bind(7, m_label_placement);
                cmd.bind(8, (long long int)m_modtime);
                cmd.bind(9, (int)TileNumber::to_tilenumber(m_coord));
                cmd.bind(10, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Waypoint::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO waypoints_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM waypoints WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM waypoints_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

Glib::ustring DbBaseElements::Waypoint::get_icao_name(void) const
{
        Glib::ustring r(m_icao);
        if (!(m_icao.empty() || m_name.empty()))
                r += " ";
        return r + m_name;
}

Glib::ustring DbBaseElements::Waypoint::get_usage_name( char usage )
{
        switch (usage) {
                default:
                case usage_invalid:
                        return "Invalid";

                case usage_highlevel:
                        return "High Level";

                case usage_lowlevel:
                        return "Low Level";

                case usage_bothlevel:
                        return "Low and High Level";

                case usage_rnav:
                        return "RNAV";

                case usage_terminal:
                        return "Terminal";

                case usage_vfr:
                        return "VFR";

                case usage_user:
                        return "User";
        }
}

template<> void DbBase<DbBaseElements::Waypoint>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS waypoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS waypoints_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Waypoint>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS waypoints (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "ICAO TEXT,"
                                              "NAME TEXT,"
                                              "USAGE CHAR,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "LABELPLACEMENT INTEGER,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS waypoints_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Waypoint>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_icao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_lon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Waypoint>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_srcid ON waypoints(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_icao ON waypoints(ICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_name ON waypoints(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_latlon ON waypoints(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_modified ON waypoints(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_tile ON waypoints(TILE);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_lat ON waypoints(LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_lon ON waypoints(LON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Waypoint>::main_table_name = "waypoints";
template<> const char *DbBase<DbBaseElements::Waypoint>::database_file_name = "waypoints.db";
template<> const char *DbBase<DbBaseElements::Waypoint>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Waypoint>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Waypoint>::area_data = false;

const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_n;
const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_e;
const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_s;
const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_w;
const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_center;
const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_any;
const DbBaseElements::Airway::label_placement_t DbBaseElements::Airway::label_off;

const char *DbBaseElements::Airway::db_query_string = "airways.ID AS 'ID',0 AS 'TABLE',airways.SRCID AS 'SRCID',airways.BNAME,airways.ENAME,airways.NAME,airways.TYPE,airways.BLAT,airways.BLON,airways.ELAT,airways.ELON,airways.SWLAT,airways.NELAT,airways.SWLON,airways.NELON,airways.BLEVEL,airways.TLEVEL,airways.LABELPLACEMENT,airways.MODIFIED";
const char *DbBaseElements::Airway::db_aux_query_string = "aux.airways.ID AS 'ID',1 AS 'TABLE',aux.airways.SRCID AS 'SRCID',aux.airways.BNAME,aux.airways.ENAME,aux.airways.NAME,aux.airways.TYPE,aux.airways.BLAT,aux.airways.BLON,aux.airways.ELAT,aux.airways.ELON,aux.airways.SWLAT,aux.airways.NELAT,aux.airways.SWLON,aux.airways.NELON,aux.airways.BLEVEL,aux.airways.TLEVEL,aux.airways.LABELPLACEMENT,aux.airways.MODIFIED";
const char *DbBaseElements::Airway::db_text_fields[] = { "SRCID", "BNAME", "ENAME", "NAME", 0 };
const char *DbBaseElements::Airway::db_time_fields[] = { "MODIFIED", 0 };

DbBaseElements::Airway::Airway(void)
        : m_id(0), m_begin_coord(), m_end_coord(), m_sourceid(), m_begin_name(), m_end_name(), m_name(), m_base_level(0), m_top_level(0), m_type(airway_invalid), m_label_placement(label_off), m_table(table_main), m_modtime(0)
{
}

Rect DbBaseElements::Airway::get_bbox(void) const
{
        Polygon p;
        p.push_back(m_begin_coord);
        p.push_back(m_end_coord);
        return p.get_bbox();
}

void DbBaseElements::Airway::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_table = table_main;
        m_type = airway_invalid;
        if (!cursor.step())
                return;
        m_id = cursor.getint(0);
        m_table = (table_t)cursor.getint(1);
        m_sourceid = cursor.getstring(2);
        m_begin_name = cursor.getstring(3);
        m_end_name = cursor.getstring(4);
        m_name = cursor.getstring(5);
        m_type = (airway_type_t)cursor.getint(6);
        m_begin_coord = Point(cursor.getint(8), cursor.getint(7));
        m_end_coord = Point(cursor.getint(10), cursor.getint(9));
        m_base_level = cursor.getint(15);
        m_top_level = cursor.getint(16);
        m_label_placement = (label_placement_t)cursor.getint(17);
        m_modtime = cursor.getint(18);
}

void DbBaseElements::Airway::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        Rect bbox(get_bbox());
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airways SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(bbox));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airways_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, bbox.get_south());
                cmd.bind(2, bbox.get_north());
                cmd.bind(3, bbox.get_west());
                cmd.bind(4, (long long int)bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Airway::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO airways_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        Rect bbox(get_bbox());
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM airways;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airways (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airways_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, bbox.get_south());
                        cmd.bind(3, bbox.get_north());
                        cmd.bind(4, bbox.get_west());
                        cmd.bind(5, (long long int)bbox.get_east_unwrapped());
                        cmd.executenonquery();
                }
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airways SET SRCID=?,BNAME=?,ENAME=?,NAME=?,TYPE=?,BLAT=?,BLON=?,ELAT=?,ELON=?,SWLAT=?,SWLON=?,NELAT=?,NELON=?,BLEVEL=?,TLEVEL=?,LABELPLACEMENT=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_begin_name);
                cmd.bind(3, m_end_name);
                cmd.bind(4, m_name);
                cmd.bind(5, (int)m_type);
                cmd.bind(6, m_begin_coord.get_lat());
                cmd.bind(7, m_begin_coord.get_lon());
                cmd.bind(8, m_end_coord.get_lat());
                cmd.bind(9, m_end_coord.get_lon());
                cmd.bind(10, bbox.get_south());
                cmd.bind(11, bbox.get_west());
                cmd.bind(12, bbox.get_north());
                cmd.bind(13, (long long int)bbox.get_east_unwrapped());
                cmd.bind(14, (int)m_base_level);
                cmd.bind(15, (int)m_top_level);
                cmd.bind(16, m_label_placement);
                cmd.bind(17, (long long int)m_modtime);
                cmd.bind(18, (int)TileNumber::to_tilenumber(get_bbox()));
                cmd.bind(19, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airways_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, bbox.get_south());
                cmd.bind(2, bbox.get_north());
                cmd.bind(3, bbox.get_west());
                cmd.bind(4, (long long int)bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Airway::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO airways_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airways WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airways_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

Glib::ustring DbBaseElements::Airway::get_type_name(airway_type_t t)
{
        switch (t) {
                default:
                case airway_invalid:
                        return "Invalid";

                case airway_low:
                        return "Low Level";

                case airway_high:
                        return "High Level";

                case airway_both:
                        return "Both Levels";

                case airway_user:
                        return "User";
        }
}

template<> void DbBase<DbBaseElements::Airway>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airways;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airways_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Airway>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airways (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "BNAME TEXT,"
                                              "ENAME TEXT,"
                                              "NAME TEXT,"
                                              "TYPE INTEGER,"
                                              "BLAT INTEGER,"
                                              "BLON INTEGER,"
                                              "ELAT INTEGER,"
                                              "ELON INTEGER,"
                                              "SWLAT INTEGER,"
                                              "SWLON INTEGER,"
                                              "NELAT INTEGER,"
                                              "NELON INTEGER,"
                                              "BLEVEL INTEGER,"
                                              "TLEVEL INTEGER,"
                                              "LABELPLACEMENT INTEGER,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airways_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Airway>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_bname;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_ename;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_bbox;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airways_tile;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Airway>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_srcid ON airways(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_bname ON airways(BNAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_ename ON airways(ENAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_name ON airways(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_latlon ON airways(BLAT,BLON,ELAT,ELON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_bbox ON airways(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_modified ON airways(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airways_tile ON airways(TILE);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Airway>::main_table_name = "airways";
template<> const char *DbBase<DbBaseElements::Airway>::database_file_name = "airways.db";
template<> const char *DbBase<DbBaseElements::Airway>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Airway>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Airway>::area_data = true;

const uint16_t DbBaseElements::Airport::Runway::flag_le_usable;
const uint16_t DbBaseElements::Airport::Runway::flag_he_usable;

const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_off;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_pcl;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_sf;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_tdzl;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_cl;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_hirl;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_mirl;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_lirl;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_rail;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_reil;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_alsf2;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_alsf1;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_sals_salsf;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_ssalr;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_mals_malsf_ssals_ssalsf;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_malsr;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_overruncenterline;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_center;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_usconfiguration;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_hongkongcurve;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerrow;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_leftsinglerow;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_formernatostandard;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerrow2;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_natostandard;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerdoublerow;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_portableapproach;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerrow3;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_hals;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_2parallelrow;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_leftrow;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_airforceoverrun;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_calvert;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_singlerowcenterline;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_narrowmulticross;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerlinehighintensity;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_alternatecenverline;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_cross;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerrow4;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_singaporecenterline;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_centerline2crossbars;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_odals;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_vasi;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_tvasi;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_pvasi;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_jumbo;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_tricolor;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_apap;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_papi;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_ols;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_waveoff;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_portable;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_floods;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_lights;
const DbBaseElements::Airport::Runway::light_t DbBaseElements::Airport::Runway::light_lcvasi;

Glib::ustring DbBaseElements::Airport::Runway::get_light_string(light_t x)
{
        switch (x) {
                case light_off:
                        return "Off";
                case light_pcl:
                        return "Pilot Controlled Lighting";
                case light_sf:
                        return "Sequenced Flashing Lights";
                case light_tdzl:
                        return "Touchdown Zone Lighting";
                case light_cl:
                        return "Centerline Lighting System";
                case light_hirl:
                        return "High Intensity Runway Lights";
                case light_mirl:
                        return "Medium Intensity Runway Lighting System";
                case light_lirl:
                        return "Low Intensity Runway Lighting System";
                case light_rail:
                        return "Runway Alignment Lights";
                case light_reil:
                        return "Runway End Identifier Lights";
                case light_alsf2:
                        return "ALSF-2";
                case light_alsf1:
                        return "ALSF-1";
                case light_sals_salsf:
                        return "SALS or SALSF";
                case light_ssalr:
                        return "SSALR";
                case light_mals_malsf_ssals_ssalsf:
                        return "MALS and MALSF or SSALS and SSALF";
                case light_malsr:
                        return "MALSR";
                case light_overruncenterline:
                        return "Overrun Centerline";
                case light_center:
                        return "Centerline and Bar";
                case light_usconfiguration:
                        return "US Configuration";
                case light_hongkongcurve:
                        return "Hong Kong Curve";
                case light_centerrow:
                        return "Center Row";
                case light_leftsinglerow:
                        return "Left Single Row";
                case light_formernatostandard:
                        return "Former NATO Standard";
                case light_centerrow2:
                        return "Center Row (2)";
                case light_natostandard:
                        return "NATO Standard";
                case light_centerdoublerow:
                        return "Center and Double Row";
                case light_portableapproach:
                        return "Portable Approach";
                case light_centerrow3:
                        return "Center Row (3)";
                case light_hals:
                        return "Helicopter Approach Lighting System (HALS)";
                case light_2parallelrow:
                        return "Two Parallel Row";
                case light_leftrow:
                        return "Left Row (High Intensity)";
                case light_airforceoverrun:
                        return "Air Force Overrun";
                case light_calvert:
                        return "Calvert (British)";
                case light_singlerowcenterline:
                        return "Single Row Centerline";
                case light_narrowmulticross:
                        return "Narrow Multi-Cross";
                case light_centerlinehighintensity:
                        return "Centerline High Intensity";
                case light_alternatecenverline:
                        return "Alternate Centerline and Bar";
                case light_cross:
                        return "Cross";
                case light_centerrow4:
                        return "Center Row (4)";
                case light_singaporecenterline:
                        return "Singapore Centerline";
                case light_centerline2crossbars:
                        return "Centerline 2 Crossbars";
                case light_odals:
                        return "Omni-directional Approach Lighting System";
                case light_vasi:
                        return "Visual Approach Slope Indicator";
                case light_tvasi:
                        return "T-Visual Approach Slope Indicator";
                case light_pvasi:
                        return "Pulsating Visual Approach Slope Indicator";
                case light_jumbo:
                        return "VASI with a TCH to accommodate long bodied or jumbo aircraft";
                case light_tricolor:
                        return "Tri-color Arrival Approach";
                case light_apap:
                        return "Alignment of Elements System";
                case light_papi:
                        return "Precision Approach Path Indicator";
                case light_ols:
                        return "Optical landing System";
                case light_waveoff:
                        return "WAVEOFF";
                case light_portable:
                        return "Portable";
                case light_floods:
                        return "Floods";
                case light_lights:
                        return "Lights";
                case light_lcvasi:
                        return "Low Cost Visual Approach Slope Indicator";
        }
        return "";
}

DbBaseElements::Airport::Runway::Runway(const Glib::ustring & idhe, const Glib::ustring & idle, uint16_t len, uint16_t wid,
                                        const Glib::ustring & sfc, const Point & hecoord, const Point & lecoord,
                                        uint16_t hetda, uint16_t helda, uint16_t hedisp, uint16_t hehdg, int16_t heelev,
                                        uint16_t letda, uint16_t lelda, uint16_t ledisp, uint16_t lehdg, int16_t leelev,
                                        uint16_t flags)
        : m_ident_he(idhe), m_ident_le(idle), m_length(len), m_width(wid),
          m_surface(sfc), m_he_coord(hecoord), m_le_coord(lecoord),
          m_he_tda(hetda), m_he_lda(helda), m_he_disp(hedisp), m_he_hdg(hehdg), m_he_elev(heelev),
          m_le_tda(letda), m_le_lda(lelda), m_le_disp(ledisp), m_le_hdg(lehdg), m_le_elev(leelev),
          m_flags(flags)
{
        memset(m_he_lights, 0, sizeof(m_he_lights));
        memset(m_le_lights, 0, sizeof(m_le_lights));
}

void DbBaseElements::Airport::Runway::compute_default_hdg(void)
{
        char *cp;
        uint16_t ang = strtoul(m_ident_he.c_str(), &cp, 10);
        if (cp == m_ident_he.c_str())
                ang = 0;
        ang = Point::round<int,float>(ang * (10.f * FPlanLeg::from_deg));
        m_he_hdg = ang;
        m_le_hdg = ang + 0x8000;
}

void DbBaseElements::Airport::Runway::compute_default_coord(const Point & arptcoord)
{
        m_he_coord = arptcoord.spheric_course_distance_nmi((m_he_hdg + 0x8000) * FPlanLeg::to_deg, m_length * (0.5f * Point::ft_to_m * 0.001f * Point::km_to_nmi));
        m_le_coord = arptcoord.spheric_course_distance_nmi((m_le_hdg + 0x8000) * FPlanLeg::to_deg, m_length * (0.5f * Point::ft_to_m * 0.001f * Point::km_to_nmi));
}

bool DbBaseElements::Airport::Runway::is_concrete(void) const
{
        return (m_surface == "ASP") || (m_surface == "ASPH") || (m_surface == "CON") || (m_surface == "PEM");
}

bool DbBaseElements::Airport::Runway::is_water(void) const
{
        return (m_surface == "WTR") || (m_surface == "WATER");
}

void DbBaseElements::Airport::Runway::swap_he_le(void)
{
        std::swap(m_ident_he, m_ident_le);
        std::swap(m_he_coord, m_le_coord);
        std::swap(m_he_tda, m_le_tda);
        std::swap(m_he_lda, m_le_lda);
        std::swap(m_he_disp, m_le_disp);
        std::swap(m_he_hdg, m_le_hdg);
        std::swap(m_he_elev, m_le_elev);
        for (unsigned int i = 0; i < sizeof(m_he_lights)/sizeof(m_he_lights[0]); ++i)
                std::swap(m_he_lights[i], m_le_lights[i]);
        m_flags = (m_flags & ~(flag_he_usable | flag_le_usable))
                        | ((m_flags << 1) & flag_he_usable)
                        | ((m_flags >> 1) & flag_le_usable);
}

bool DbBaseElements::Airport::Runway::normalize_he_le(void)
{
#if 1
        if (m_ident_le.size() < 2)
                return false;
        if (!Glib::Unicode::isdigit(m_ident_le[0]) || !Glib::Unicode::isdigit(m_ident_le[1]))
                return false;
        unsigned int le((m_ident_le[0] - '0') * 10 + (m_ident_le[1] - '0'));
        if (le < 19)
                return false;
#else
        if (m_he_hdg >= 0x8000)
                return false;
#endif
        swap_he_le();
        return true;
}

DbBaseElements::Airport::Helipad::Helipad(const Glib::ustring & id, uint16_t len, uint16_t wid, const Glib::ustring & sfc, const Point & coord, uint16_t hdg, int16_t elev, uint16_t flags)
        : m_ident(id), m_length(len), m_width(wid), m_surface(sfc), m_coord(coord), m_hdg(hdg), m_elev(elev), m_flags(flags)
{
}

bool DbBaseElements::Airport::Helipad::is_concrete(void) const
{
        return (m_surface == "ASP") || (m_surface == "ASPH") || (m_surface == "CON") || (m_surface == "PEM");
}

DbBaseElements::Airport::Comm::Comm( const Glib::ustring & nm, const Glib::ustring & sect, const Glib::ustring & oprhr, uint32_t f0, uint32_t f1, uint32_t f2, uint32_t f3, uint32_t f4 )
        : m_name(nm), m_sector(sect), m_oprhours(oprhr)
{
        m_freq[0] = f0;
        m_freq[1] = f1;
        m_freq[2] = f2;
        m_freq[3] = f3;
        m_freq[4] = f4;
}

DbBaseElements::Airport::VFRRoute::VFRRoutePoint::VFRRoutePoint( const Glib::ustring & nm, const Point & pt, int16_t al, label_placement_t lp, char pts, pathcode_t pc, altcode_t ac, bool aa )
        : m_name(nm), m_coord(pt), m_alt(al), m_label_placement(lp), m_ptsym(pts), m_pathcode(pc), m_altcode(ac), m_at_airport(aa)
{
}

Glib::ustring DbBaseElements::Airport::VFRRoute::VFRRoutePoint::get_altitude_string(void) const
{
        char buf[16];
        snprintf(buf, sizeof(buf), "%d'", get_altitude());
        switch (get_altcode()) {
                case altcode_atorabove:
                        return Glib::ustring(">=") + buf;

                case altcode_atorbelow:
                        return Glib::ustring("<=") + buf;

                case altcode_assigned:
                        return Glib::ustring(":=") + buf;

                case altcode_altspecified:
                        return Glib::ustring("=") + buf;

                case altcode_recommendedalt:
                        return Glib::ustring("~") + buf;

                default:
                        return Glib::ustring("??") + buf;
        }
}

Glib::ustring DbBaseElements::Airport::VFRRoute::VFRRoutePoint::get_altcode_string(altcode_t ac)
{
        switch (ac) {
                case altcode_atorabove:
                        return Glib::ustring(">=");

                case altcode_atorbelow:
                        return Glib::ustring("<=");

                case altcode_assigned:
                        return Glib::ustring(":=");

                case altcode_altspecified:
                        return Glib::ustring("=");

                case altcode_recommendedalt:
                        return Glib::ustring("~");

                default:
                        return Glib::ustring("??");
        }
}

Glib::ustring DbBaseElements::Airport::VFRRoute::VFRRoutePoint::get_pathcode_string( pathcode_t pc )
{
        switch (pc) {
                case pathcode_arrival:
                        return "Arrival";

                case pathcode_departure:
                        return "Departure";

                case pathcode_holding:
                        return "Holding";

                case pathcode_trafficpattern:
                        return "Traffic Pattern";

                case pathcode_vfrtransition:
                        return "VFR Transition";

                case pathcode_other:
                        return "Other";

                default:
                        return "Invalid";
        }
}

DbBaseElements::Airport::VFRRoute::VFRRoute( const Glib::ustring & nm )
        : m_name(nm)
{
}

DbBaseElements::Airport::PolylineNode::PolylineNode(const Point & pt, const Point & pcpt, const Point & ncpt, uint16_t flags, feature_t f1, feature_t f2)
        : m_point(pt), m_prev_controlpoint(pcpt), m_next_controlpoint(ncpt), m_flags(flags)
{
        m_feature[0] = f1;
        m_feature[1] = f2;
}

unsigned int DbBaseElements::Airport::PolylineNode::blobsize(void) const
{
        unsigned int sz(12);
        if (!is_bezier())
                return sz;
        sz += 8;
        if (!(is_bezier_prev() && is_bezier_next()))
                return sz;
        Point pt(m_point - m_prev_controlpoint + m_point);
        if (pt == m_next_controlpoint)
                return sz;
        sz += 8;
        return sz;
}

uint8_t * DbBaseElements::Airport::PolylineNode::blobencode(uint8_t * ptr) const
{
        uint16_t flags(m_flags & ~flag_bezier_splitcontrol);
        Point ptprevmirror(m_point - m_prev_controlpoint + m_point);
        Point ptctrl(m_next_controlpoint);
        if (!is_bezier_next()) {
                ptctrl = ptprevmirror;
        } else if (is_bezier_next() && ptprevmirror != m_next_controlpoint) {
                flags |= flag_bezier_splitcontrol;
        }
        ptr[0] = m_point.get_lon() >> 0;
        ptr[1] = m_point.get_lon() >> 8;
        ptr[2] = m_point.get_lon() >> 16;
        ptr[3] = m_point.get_lon() >> 24;
        ptr[4] = m_point.get_lat() >> 0;
        ptr[5] = m_point.get_lat() >> 8;
        ptr[6] = m_point.get_lat() >> 16;
        ptr[7] = m_point.get_lat() >> 24;
        ptr += 8;
        ptr[0] = flags >> 0;
        ptr[1] = flags >> 8;
        ptr += 2;
        ptr[0] = m_feature[0];
        ptr[1] = m_feature[1];
        ptr += 2;
        if (is_bezier()) {
                ptr[0] = ptctrl.get_lon() >> 0;
                ptr[1] = ptctrl.get_lon() >> 8;
                ptr[2] = ptctrl.get_lon() >> 16;
                ptr[3] = ptctrl.get_lon() >> 24;
                ptr[4] = ptctrl.get_lat() >> 0;
                ptr[5] = ptctrl.get_lat() >> 8;
                ptr[6] = ptctrl.get_lat() >> 16;
                ptr[7] = ptctrl.get_lat() >> 24;
                ptr += 8;
                if (flags & flag_bezier_splitcontrol) {
                        ptr[0] = m_prev_controlpoint.get_lon() >> 0;
                        ptr[1] = m_prev_controlpoint.get_lon() >> 8;
                        ptr[2] = m_prev_controlpoint.get_lon() >> 16;
                        ptr[3] = m_prev_controlpoint.get_lon() >> 24;
                        ptr[4] = m_prev_controlpoint.get_lat() >> 0;
                        ptr[5] = m_prev_controlpoint.get_lat() >> 8;
                        ptr[6] = m_prev_controlpoint.get_lat() >> 16;
                        ptr[7] = m_prev_controlpoint.get_lat() >> 24;
                        ptr += 8;
                }
        }
        return ptr;
}

const uint8_t * DbBaseElements::Airport::PolylineNode::blobdecode(const uint8_t *ptr, uint8_t const *end)
{
        m_point = Point();
        m_prev_controlpoint = Point();
        m_next_controlpoint = Point();
        m_flags = 0;
        m_feature[0] = feature_off;
        m_feature[1] = feature_off;
        if (ptr + 12 > end)
                return end;
        uint32_t lon = ptr[0] | (ptr[1] << 8) | (ptr[2] << 16) | (ptr[3] << 24);
        uint32_t lat = ptr[4] | (ptr[5] << 8) | (ptr[6] << 16) | (ptr[7] << 24);
        ptr += 8;
        m_point = Point(lon, lat);
        m_prev_controlpoint = m_point;
        m_next_controlpoint = m_point;
        uint16_t flags = ptr[0] | (ptr[1] << 8);
        m_flags = flags & ~flag_bezier_splitcontrol;
        ptr += 2;
        m_feature[0] = (feature_t)ptr[0];
        m_feature[1] = (feature_t)ptr[1];
        ptr += 2;
        if (!is_bezier())
                return ptr;
        if (ptr + 8 > end) {
                m_flags &= ~(flag_bezier_prev | flag_bezier_next);
                return end;
        }
        lon = ptr[0] | (ptr[1] << 8) | (ptr[2] << 16) | (ptr[3] << 24);
        lat = ptr[4] | (ptr[5] << 8) | (ptr[6] << 16) | (ptr[7] << 24);
        ptr += 8;
        m_next_controlpoint = Point(lon, lat);
        if ((flags & flag_bezier_splitcontrol) && (ptr + 8 <= end)) {
                lon = ptr[0] | (ptr[1] << 8) | (ptr[2] << 16) | (ptr[3] << 24);
                lat = ptr[4] | (ptr[5] << 8) | (ptr[6] << 16) | (ptr[7] << 24);
                ptr += 8;
                m_prev_controlpoint = Point(lon, lat);
        } else {
                m_prev_controlpoint = m_point - m_next_controlpoint + m_point;
        }
        return ptr;
}

bool DbBaseElements::Airport::Polyline::is_concrete(void) const
{
        return (m_surface == "ASP") || (m_surface == "ASPH") || (m_surface == "CON") || (m_surface == "PEM");
}

unsigned int DbBaseElements::Airport::Polyline::blobsize_polygon(void) const
{
        unsigned int sz(2);
        for (const_iterator pi(begin()), pe(end()); pi != pe; ++pi)
                sz += pi->blobsize();
        return sz;
}

unsigned int DbBaseElements::Airport::Polyline::blobsize(void) const
{
        unsigned int sz(6 + m_name.bytes() + m_surface.bytes() + blobsize_polygon());
        return sz;
}

uint8_t * DbBaseElements::Airport::Polyline::blobencode_polygon(uint8_t * ptr) const
{
        uint16_t len = size();
        ptr[0] = len >> 0;
        ptr[1] = len >> 8;
        ptr += 2;
        for (const_iterator pi(begin()), pe(end()); pi != pe; ++pi)
                ptr = pi->blobencode(ptr);
        return ptr;
}

uint8_t *DbBaseElements::Airport::Polyline::blobencode(uint8_t *ptr) const
{
        ptr[0] = m_flags >> 0;
        ptr[1] = m_flags >> 8;
        ptr += 2;
        uint16_t len = m_name.bytes();
        ptr[0] = len >> 0;
        ptr[1] = len >> 8;
        ptr += 2;
        if (len) {
                memcpy(ptr, m_name.data(), len);
                ptr += len;
        }
        len = m_surface.bytes();
        ptr[0] = len >> 0;
        ptr[1] = len >> 8;
        ptr += 2;
        if (len) {
                memcpy(ptr, m_surface.data(), len);
                ptr += len;
        }
        return blobencode_polygon(ptr);
}

const uint8_t * DbBaseElements::Airport::Polyline::blobdecode_polygon(const uint8_t * ptr, uint8_t const * end)
{
        clear();
        if (ptr + 2 > end)
                return end;
        uint16_t len = ptr[0] | (ptr[1] << 8);
        ptr += 2;
        for (; len > 0 && ptr < end; --len) {
                push_back(PolylineNode());
                ptr = back().blobdecode(ptr, end);
        }
        return ptr;
}

const uint8_t *DbBaseElements::Airport::Polyline::blobdecode(const uint8_t *ptr, uint8_t const *end)
{
        m_name = "";
        m_surface = "";
        m_flags = 0;
        clear();
        if (ptr + 4 > end)
                return end;
        m_flags = ptr[0] | (ptr[1] << 8);
        ptr += 2;
        uint16_t len = ptr[0] | (ptr[1] << 8);
        ptr += 2;
        if (len && ptr + len <= end) {
                m_name = std::string((const char *)ptr, len);
                ptr += len;
        }
        if (ptr + 2 > end)
                return end;
        len = ptr[0] | (ptr[1] << 8);
        ptr += 2;
        if (len && ptr + len <= end) {
                m_surface = std::string((const char *)ptr, len);
                ptr += len;
        }
        return blobdecode_polygon(ptr, end);
}

void DbBaseElements::Airport::Polyline::bindblob(sqlite3x::sqlite3_command & cmd, int index) const
{
        unsigned int sz(blobsize());
        uint8_t data[sz];
        uint8_t *d = blobencode(data);
        cmd.bind(index, data, sz);
}

void DbBaseElements::Airport::Polyline::getblob(sqlite3x::sqlite3_cursor & cursor, int index)
{
        m_name = "";
        m_surface = "";
        m_flags = 0;
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        uint8_t const *end(data + sz);
        data = blobdecode(data, end);
}

void DbBaseElements::Airport::Polyline::bindblob_polygon(sqlite3x::sqlite3_command & cmd, int index) const
{
        unsigned int sz(blobsize_polygon());
        uint8_t data[sz];
        uint8_t *d = blobencode_polygon(data);
        cmd.bind(index, data, sz);
}

void DbBaseElements::Airport::Polyline::getblob_polygon(sqlite3x::sqlite3_cursor & cursor, int index)
{
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        uint8_t const *end(data + sz);
        data = blobdecode_polygon(data, end);
}

DbBaseElements::Airport::Polyline::operator PolygonHole(void) const
{
        static const int bdiv = 6;
        static const int ndiv = 1 << bdiv;
        PolygonHole ph;
        if (empty())
                return ph;
        Polygon p;
        bool first = true;
        const PolylineNode *pnfirst = 0;
        for (unsigned int i = 0;; ++i) {
                const PolylineNode& pn1(operator[](i));
                if (!pnfirst) {
                        pnfirst = &pn1;
                        p.push_back(pn1.get_point());
                        continue;
                }
                const PolylineNode& pn0(operator[](i - 1));
                if (pn0.is_bezier_next() || pn1.is_bezier_prev()) {
                        Point p0(pn0.get_point());
                        Point p1(pn0.get_next_controlpoint() - p0);
                        Point p2(pn1.get_prev_controlpoint() - p0);
                        Point p3(pn1.get_point() - p0);
                        for (int t = 1; t < ndiv; ++t) {
                                int tm = ndiv - t;
                                int m1 = 3 * t * tm * tm;
                                int m2 = 3 * t * t * tm;
                                int m3 = t * t * t;
                                int64_t lon(m1 * (int64_t)p1.get_lon() + m2 * (int64_t)p2.get_lon() + m3 * (int64_t)p3.get_lon());
                                int64_t lat(m1 * (int64_t)p1.get_lat() + m2 * (int64_t)p2.get_lat() + m3 * (int64_t)p3.get_lat());
                                p.push_back(Point(lon >> (3 * bdiv), lat >> (3 * bdiv)) + p0);
                        }
                }
                p.push_back(pn1.get_point());
                if (pn1.is_closepath() && (pn0.is_bezier_next() || pn1.is_bezier_prev())) {
                        Point p0(pn1.get_point());
                        Point p1(pn1.get_next_controlpoint() - p0);
                        Point p2(pnfirst->get_prev_controlpoint() - p0);
                        Point p3(pnfirst->get_point() - p0);
                        for (int t = 1; t < ndiv; ++t) {
                                int tm = ndiv - t;
                                int m1 = 3 * t * tm * tm;
                                int m2 = 3 * t * t * tm;
                                int m3 = t * t * t;
                                int64_t lon(m1 * (int64_t)p1.get_lon() + m2 * (int64_t)p2.get_lon() + m3 * (int64_t)p3.get_lon());
                                int64_t lat(m1 * (int64_t)p1.get_lat() + m2 * (int64_t)p2.get_lat() + m3 * (int64_t)p3.get_lat());
                                p.push_back(Point(lon >> (3 * bdiv), lat >> (3 * bdiv)) + p0);
                        }
                }
                if (pn1.is_closepath() || pn1.is_endpath()) {
                        pnfirst = 0;
                        if (first)
                                ph.set_exterior(p);
                        else
                                ph.add_interior(p);
                        first = false;
                        p.clear();
                }
        }
}

Rect DbBaseElements::Airport::Polyline::get_bbox(void) const
{
        if (empty())
                return Rect();
        Point poffs(begin()->get_point());
        Point sw(0, 0), ne(0, 0);
        for (const_iterator pi(begin()), pe(end()); pi != pe; ++pi) {
                Point p(pi->get_point() - poffs);
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
                if (!pi->is_bezier())
                        continue;
                // FIXME: this is pessimistic, should actually find max of bezier curve
                p = pi->get_point() - poffs;
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        return Rect(sw + poffs, ne + poffs);
}

DbBaseElements::Airport::PointFeature::PointFeature(feature_t ft, const Point & pt, uint16_t hdg, int16_t elev, unsigned int subtype, int attr1, int attr2, const Glib::ustring & name, const Glib::ustring & rwyident)
        : m_feature(ft), m_coord(pt), m_hdg(hdg), m_elev(elev), m_subtype(subtype), m_attr1(attr1), m_attr2(attr2), m_name(name), m_rwyident(rwyident)
{
}

const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_n;
const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_e;
const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_s;
const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_w;
const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_center;
const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_any;
const DbBaseElements::Airport::label_placement_t DbBaseElements::Airport::label_off;

const char *DbBaseElements::Airport::db_query_string = "airports.ID AS 'ID',0 AS 'TABLE',airports.SRCID AS 'SRCID',airports.ICAO,airports.NAME,airports.VFRRMK,airports.LAT,airports.LON,airports.SWLAT,airports.SWLON,airports.NELAT,airports.NELON,airports.ELEVATION,airports.TYPECODE,airports.LABELPLACEMENT,airports.MODIFIED";
const char *DbBaseElements::Airport::db_aux_query_string = "aux.airports.ID AS 'ID',1 AS 'TABLE',aux.airports.SRCID AS 'SRCID',aux.airports.ICAO,aux.airports.NAME,aux.airports.VFRRMK,aux.airports.LAT,aux.airports.LON,aux.airports.SWLAT,aux.airports.SWLON,aux.airports.NELAT,aux.airports.NELON,aux.airports.ELEVATION,aux.airports.TYPECODE,aux.airports.LABELPLACEMENT,aux.airports.MODIFIED";
const char *DbBaseElements::Airport::db_text_fields[] = { "SRCID", "ICAO", "NAME", "VFRRMK", 0 };
const char *DbBaseElements::Airport::db_time_fields[] = { "MODIFIED", 0 };

DbBaseElements::Airport::Airport(void)
        : m_id(0), m_rwy(), m_comm(), m_vfrrte(), m_sourceid(), m_icao(), m_name(), m_vfrrmk(),
          m_bbox(), m_coord(), m_elev(0), m_typecode(0), m_subtables(subtables_all), m_label_placement(label_off), m_table(table_main), m_modtime(0)
{
}

void DbBaseElements::Airport::load_subtables(sqlite3x::sqlite3_connection & db, unsigned int loadsubtables)
{
        //std::cerr << "Airport::load_subtables: loadsubtables 0x" << std::hex << loadsubtables << " m_subtables 0x" << std::hex << m_subtables << std::endl;
        loadsubtables &= subtables_all & ~m_subtables;
        if (!loadsubtables)
                return;
        std::string tablepfx;
        switch (m_table) {
                case 0:
                        break;
                case 1:
                        tablepfx = "aux.";
                        break;
                default:
                        throw std::runtime_error("invalid table ID");
        }
        //std::cerr << "Airport::load_subtables ID " << m_id << " table " << (int)m_table << ' ' << tablepfx + "airportrunways" << std::endl;
        if (loadsubtables & subtables_runways) {
                m_rwy.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT IDENTHE,IDENTLE,LENGTH,WIDTH,SURFACE,HELAT,HELON,LELAT,LELON,HETDA,HELDA,HEDISP,HEHDG,HEELEV,LETDA,LELDA,LEDISP,LEHDG,LEELEV,FLAGS,HELIGHT0,HELIGHT1,HELIGHT2,HELIGHT3,HELIGHT4,HELIGHT5,HELIGHT6,HELIGHT7,LELIGHT0,LELIGHT1,LELIGHT2,LELIGHT3,LELIGHT4,LELIGHT5,LELIGHT6,LELIGHT7 FROM " + tablepfx + "airportrunways WHERE ARPTID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        Runway rwy(cursor.getstring(0), cursor.getstring(1), cursor.getint(2), cursor.getint(3), cursor.getstring(4),
                                   Point(cursor.getint(6), cursor.getint(5)), Point(cursor.getint(8), cursor.getint(7)),
                                         cursor.getint(9), cursor.getint(10), cursor.getint(11), cursor.getint(12), cursor.getint(13),
                                                         cursor.getint(14), cursor.getint(15), cursor.getint(16), cursor.getint(17), cursor.getint(18), cursor.getint(19));
                        for (unsigned int i = 0; i < 8; ++i) {
                                rwy.set_he_light(i, cursor.getint(20 + i));
                                rwy.set_le_light(i, cursor.getint(28 + i));
                        }
                        add_rwy(rwy);
                }
                m_subtables |= subtables_runways;
        }
        if (loadsubtables & subtables_helipads) {
                m_helipad.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT IDENT,LENGTH,WIDTH,SURFACE,LAT,LON,HDG,ELEV,FLAGS FROM " + tablepfx + "airporthelipads WHERE ARPTID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        Helipad hp(cursor.getstring(0), cursor.getint(1), cursor.getint(2), cursor.getstring(3),
                                    Point(cursor.getint(5), cursor.getint(4)), cursor.getint(6), cursor.getint(7), cursor.getint(8));
                        add_helipad(hp);
                }
                m_subtables |= subtables_helipads;
        }
        if (loadsubtables & subtables_comms) {
                m_comm.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT NAME,SECTOR,OPRHOURS,FREQ0,FREQ1,FREQ2,FREQ3,FREQ4 FROM " + tablepfx + "airportcomms WHERE ARPTID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        add_comm(Comm(cursor.getstring(0), cursor.getstring(1), cursor.getstring(2), cursor.getint(3), cursor.getint(4), cursor.getint(5), cursor.getint(6), cursor.getint(7)));
                }
                m_subtables |= subtables_comms;
        }
        if (loadsubtables & subtables_vfrroutes) {
                m_vfrrte.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT ID,NAME FROM " + tablepfx + "airportvfrroutes WHERE ARPTID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        VFRRoute rte(cursor.getstring(1));
                        sqlite3x::sqlite3_command cmdp(db, "SELECT NAME,LAT,LON,ALT,ALTCODE,PATHCODE,SYM,ATAIRPORT,LABELPLACEMENT FROM " + tablepfx + "airportvfrroutepoints WHERE ARPTID=? AND ROUTEID=? ORDER BY ID;");
                        cmdp.bind(1, (long long int)m_id);
                        cmdp.bind(2, cursor.getint(0));
                        sqlite3x::sqlite3_cursor cursorp(cmdp.executecursor());
                        while (cursorp.step()) {
                                VFRRoute::VFRRoutePoint
                                        rtept(cursorp.getstring(0), Point(cursorp.getint(2), cursorp.getint(1)),
                                              cursorp.getint(3), (label_placement_t)cursorp.getint(8), cursorp.getint(6),
                                              (VFRRoute::VFRRoutePoint::pathcode_t)cursorp.getint(5),
                                              (VFRRoute::VFRRoutePoint::altcode_t)cursorp.getint(4), cursorp.getint(7));
                                rte.add_point(rtept);
                        }
                        add_vfrrte(rte);
                }
                m_subtables |= subtables_vfrroutes;
        }
        if (loadsubtables & subtables_linefeatures) {
                m_linefeature.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT NAME,SURFACE,FLAGS,POLY FROM " + tablepfx + "airportlinefeatures WHERE ARPTID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        Polyline pl;
                        pl.set_name(cursor.getstring(0));
                        pl.set_surface(cursor.getstring(1));
                        pl.set_flags(cursor.getint(2));
                        pl.getblob_polygon(cursor, 3);
                        add_linefeature(pl);
                }
                m_subtables |= subtables_linefeatures;
        }
        if (loadsubtables & subtables_pointfeatures) {
                m_pointfeature.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT FEATURE,LAT,LON,HDG,ELEV,SUBTYPE,ATTR1,ATTR2,NAME,RWYIDENT FROM " + tablepfx + "airportpointfeatures WHERE ARPTID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        PointFeature pf((PointFeature::feature_t)cursor.getint(0), Point(cursor.getint(2), cursor.getint(1)),
                                        cursor.getint(3), cursor.getint(4), cursor.getint(5), cursor.getint(6), cursor.getint(7),
                                        cursor.getstring(8), cursor.getstring(9));
                        add_pointfeature(pf);
                }
                m_subtables |= subtables_pointfeatures;
        }
}

void DbBaseElements::Airport::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_table = table_main;
        m_typecode = 0;
        m_subtables = subtables_none;
        {
                if (!cursor.step())
                        return;
                m_id = cursor.getint(0);
                m_table = (table_t)cursor.getint(1);
                m_sourceid = cursor.getstring(2);
                m_icao = cursor.getstring(3);
                m_name = cursor.getstring(4);
                m_vfrrmk = cursor.getstring(5);
                m_coord = Point(cursor.getint(7), cursor.getint(6));
                m_bbox = Rect(Point(cursor.getint(9), cursor.getint(8)), Point(cursor.getint(11), cursor.getint(10)));
                m_elev = cursor.getint(12);
                m_typecode = cursor.getint(13);
                m_label_placement = (label_placement_t)cursor.getint(14);
                m_modtime = cursor.getint(15);
        }
        if (!loadsubtables)
                return;
        load_subtables(db, loadsubtables);
}

void DbBaseElements::Airport::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airports SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airports_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Airport::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO airports_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM airports;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airports (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airports_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_bbox.get_south());
                        cmd.bind(3, m_bbox.get_north());
                        cmd.bind(4, m_bbox.get_west());
                        cmd.bind(5, (long long int)m_bbox.get_east_unwrapped());
                        cmd.executenonquery();
                }
        }
        // Update Airport Record
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airports SET SRCID=?,ICAO=?,NAME=?,VFRRMK=?,LAT=?,LON=?,SWLAT=?,SWLON=?,NELAT=?,NELON=?,ELEVATION=?,TYPECODE=?,LABELPLACEMENT=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_icao);
                cmd.bind(3, m_name);
                cmd.bind(4, m_vfrrmk);
                cmd.bind(5, m_coord.get_lat());
                cmd.bind(6, m_coord.get_lon());
                cmd.bind(7, m_bbox.get_south());
                cmd.bind(8, m_bbox.get_west());
                cmd.bind(9, m_bbox.get_north());
                cmd.bind(10, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(11, m_elev);
                cmd.bind(12, m_typecode);
                cmd.bind(13, m_label_placement);
                cmd.bind(14, (long long int)m_modtime);
                cmd.bind(15, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.bind(16, (long long int)m_id);
                cmd.executenonquery();
        }
        if (m_subtables & subtables_runways) {
                // Delete existing Runways
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportrunways WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write Runways
                for (unsigned int i = 0; i < get_nrrwy(); i++) {
                        const Runway& rwy(get_rwy(i));
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airportrunways (ARPTID,ID,IDENTHE,IDENTLE,LENGTH,WIDTH,SURFACE,HELAT,HELON,LELAT,LELON,HETDA,HELDA,HEDISP,HEHDG,HEELEV,LETDA,LELDA,LEDISP,LEHDG,LEELEV,FLAGS,HELIGHT0,HELIGHT1,HELIGHT2,HELIGHT3,HELIGHT4,HELIGHT5,HELIGHT6,HELIGHT7,LELIGHT0,LELIGHT1,LELIGHT2,LELIGHT3,LELIGHT4,LELIGHT5,LELIGHT6,LELIGHT7) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, rwy.get_ident_he());
                        cmd.bind(4, rwy.get_ident_le());
                        cmd.bind(5, rwy.get_length());
                        cmd.bind(6, rwy.get_width());
                        cmd.bind(7, rwy.get_surface());
                        cmd.bind(8, rwy.get_he_coord().get_lat());
                        cmd.bind(9, rwy.get_he_coord().get_lon());
                        cmd.bind(10, rwy.get_le_coord().get_lat());
                        cmd.bind(11, rwy.get_le_coord().get_lon());
                        cmd.bind(12, rwy.get_he_tda());
                        cmd.bind(13, rwy.get_he_lda());
                        cmd.bind(14, rwy.get_he_disp());
                        cmd.bind(15, rwy.get_he_hdg());
                        cmd.bind(16, rwy.get_he_elev());
                        cmd.bind(17, rwy.get_le_tda());
                        cmd.bind(18, rwy.get_le_lda());
                        cmd.bind(19, rwy.get_le_disp());
                        cmd.bind(20, rwy.get_le_hdg());
                        cmd.bind(21, rwy.get_le_elev());
                        cmd.bind(22, (long long int)rwy.get_flags());
                        for (unsigned int i = 0; i < 8; ++i) {
                                cmd.bind(23 + i, rwy.get_he_light(i));
                                cmd.bind(31 + i, rwy.get_le_light(i));
                        }
                        cmd.executenonquery();
                }
        }
        if (m_subtables & subtables_helipads) {
                // Delete existing Helipads
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airporthelipads WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write Helipads
                for (unsigned int i = 0; i < get_nrhelipads(); i++) {
                        const Helipad& hp(get_helipad(i));
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airporthelipads (ARPTID,ID,IDENT,LENGTH,WIDTH,SURFACE,LAT,LON,HDG,ELEV,FLAGS) VALUES(?,?,?,?,?,?,?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, hp.get_ident());
                        cmd.bind(4, hp.get_length());
                        cmd.bind(5, hp.get_width());
                        cmd.bind(6, hp.get_surface());
                        cmd.bind(7, hp.get_coord().get_lat());
                        cmd.bind(8, hp.get_coord().get_lon());
                        cmd.bind(9, hp.get_hdg());
                        cmd.bind(10, hp.get_elev());
                        cmd.bind(11, (long long int)hp.get_flags());
                        cmd.executenonquery();
                }
        }
        if (m_subtables & subtables_comms) {
                // Delete existing Comms
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportcomms WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write Comms
                for (unsigned int i = 0; i < get_nrcomm(); i++) {
                        const Comm& comm(get_comm(i));
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airportcomms (ARPTID,ID,NAME,SECTOR,OPRHOURS,FREQ0,FREQ1,FREQ2,FREQ3,FREQ4) VALUES(?,?,?,?,?,?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, comm.get_name());
                        cmd.bind(4, comm.get_sector());
                        cmd.bind(5, comm.get_oprhours());
                        for (unsigned int j = 0; j < 5; j++)
                                cmd.bind(6+j, (long long int)comm[j]);
                        cmd.executenonquery();
                }
        }
        if (m_subtables & subtables_vfrroutes) {
                // Delete existing VFR Routes and Points
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportvfrroutes WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportvfrroutepoints WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write VFR Routes
                for (unsigned int i = 0; i < get_nrvfrrte(); i++) {
                        const VFRRoute& rte(get_vfrrte(i));
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airportvfrroutes (ARPTID,ID,NAME) VALUES(?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, rte.get_name());
                        cmd.executenonquery();
                        for (unsigned int j = 0; j < rte.size(); j++) {
                                const VFRRoute::VFRRoutePoint& rtept(rte[j]);
                                sqlite3x::sqlite3_command cmd(db, "INSERT INTO airportvfrroutepoints (ARPTID,ROUTEID,ID,NAME,LAT,LON,ALT,ALTCODE,PATHCODE,SYM,ATAIRPORT,LABELPLACEMENT) VALUES(?,?,?,?,?,?,?,?,?,?,?,?);");
                                cmd.bind(1, (long long int)m_id);
                                cmd.bind(2, (int)i);
                                cmd.bind(3, (int)j);
                                cmd.bind(4, rtept.get_name());
                                cmd.bind(5, rtept.get_coord().get_lat());
                                cmd.bind(6, rtept.get_coord().get_lon());
                                cmd.bind(7, rtept.get_altitude());
                                cmd.bind(8, rtept.get_altcode());
                                cmd.bind(9, rtept.get_pathcode());
                                cmd.bind(10, rtept.get_ptsym());
                                cmd.bind(11, rtept.is_at_airport());
                                cmd.bind(12, rtept.get_label_placement());
                                cmd.executenonquery();
                        }
                }
        }
        if (m_subtables & subtables_linefeatures) {
                // Delete existing Line Features
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportlinefeatures WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write Line Features
                for (unsigned int i = 0; i < get_nrlinefeatures(); i++) {
                        const Polyline& pl(get_linefeature(i));
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airportlinefeatures (ARPTID,ID,NAME,SURFACE,FLAGS,POLY) VALUES(?,?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, pl.get_name());
                        cmd.bind(4, pl.get_surface());
                        cmd.bind(5, (long long int)pl.get_flags());
                        pl.bindblob_polygon(cmd, 6);
                        cmd.executenonquery();
                }
        }
        if (m_subtables & subtables_pointfeatures) {
                // Delete existing Point Features
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportpointfeatures WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write Point Features
                for (unsigned int i = 0; i < get_nrpointfeatures(); i++) {
                        const PointFeature& pf(get_pointfeature(i));
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airportpointfeatures (ARPTID,ID,FEATURE,LAT,LON,HDG,ELEV,SUBTYPE,ATTR1,ATTR2,NAME,RWYIDENT) VALUES(?,?,?,?,?,?,?,?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, (int)pf.get_feature());
                        cmd.bind(4, pf.get_coord().get_lat());
                        cmd.bind(5, pf.get_coord().get_lon());
                        cmd.bind(6, pf.get_hdg());
                        cmd.bind(7, pf.get_elev());
                        cmd.bind(8, (long long int)pf.get_subtype());
                        cmd.bind(9, pf.get_attr1());
                        cmd.bind(10, pf.get_attr2());
                        cmd.bind(11, pf.get_name());
                        cmd.bind(12, pf.get_rwyident());
                        cmd.executenonquery();
                }
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airports_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Airport::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO airports_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airports WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportrunways WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airporthelipads WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportcomms WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportvfrroutes WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportvfrroutepoints WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportlinefeatures WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airportpointfeatures WHERE ARPTID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airports_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

Glib::ustring DbBaseElements::Airport::get_icao_name(void) const
{
        Glib::ustring r(m_icao);
        if (!(m_icao.empty() || m_name.empty()))
                r += " ";
        return r + m_name;
}

Glib::ustring DbBaseElements::Airport::get_type_string( char t )
{
        switch (t) {
        case 'A':
                return "CIV";

        case 'B':
                return "CIV/MIL";

        case 'C':
                return "MIL";

        case 'D':
        default:
                return "other";
        }
}

void DbBaseElements::Airport::recompute_bbox(void)
{
        Point sw(0, 0), ne(0, 0);
        for (unsigned int i = 0; i < get_nrvfrrte(); i++) {
                const VFRRoute& rt(get_vfrrte(i));
                for (unsigned int j = 0; j < rt.size(); j++) {
                        const VFRRoute::VFRRoutePoint& rtp(rt[j]);
                        Point p(rtp.get_coord() - get_coord());
                        sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                        ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
                }
        }
        for (unsigned int i = 0; i < get_nrrwy(); i++) {
                const Runway& rwy(get_rwy(i));
                {
                        Point p(rwy.get_he_coord() - get_coord());
                        sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                        ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
                }
                {
                        Point p(rwy.get_le_coord() - get_coord());
                        sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                        ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
                }
        }
        for (unsigned int i = 0; i < get_nrhelipads(); i++) {
                const Helipad& hp(get_helipad(i));
                Point p(hp.get_coord() - get_coord());
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        for (unsigned int i = 0; i < get_nrpointfeatures(); i++) {
                const PointFeature& pf(get_pointfeature(i));
                Point p(pf.get_coord() - get_coord());
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        for (unsigned int i = 0; i < get_nrlinefeatures(); i++) {
                const Polyline& pl(get_linefeature(i));
                if (pl.empty())
                        continue;
                Rect bb(pl.get_bbox());
                sw = Point(std::min(sw.get_lon(), bb.get_west() - get_coord().get_lon()), std::min(sw.get_lat(), bb.get_south() - get_coord().get_lat()));
                ne = Point(std::max(ne.get_lon(), bb.get_east() - get_coord().get_lon()), std::max(ne.get_lat(), bb.get_north() - get_coord().get_lat()));
        }
        m_bbox = Rect(sw + get_coord(), ne + get_coord());
}

template<> void DbBase<DbBaseElements::Airport>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airports;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airportrunways;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airporthelipads;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airportcomms;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airportvfrroutes;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airportvfrroutepoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airportlinefeatures;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airportpointfeatures;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airports_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Airport>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airports (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "ICAO TEXT,"
                                              "NAME TEXT,"
                                              "VFRRMK TEXT,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "SWLAT INTEGER,"
                                              "SWLON INTEGER,"
                                              "NELAT INTEGER,"
                                              "NELON INTEGER,"
                                              "ELEVATION INTEGER,"
                                              "TYPECODE CHAR,"
                                              "LABELPLACEMENT CHAR,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airportrunways (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "IDENTHE TEXT,"
                                              "IDENTLE TEXT,"
                                              "LENGTH INTEGER,"
                                              "WIDTH INTEGER,"
                                              "SURFACE TEXT,"
                                              "HELAT INTEGER,"
                                              "HELON INTEGER,"
                                              "LELAT INTEGER,"
                                              "LELON INTEGER,"
                                              "HETDA INTEGER,"
                                              "HELDA INTEGER,"
                                              "HEDISP INTEGER,"
                                              "HEHDG INTEGER,"
                                              "HEELEV INTEGER,"
                                              "LETDA INTEGER,"
                                              "LELDA INTEGER,"
                                              "LEDISP INTEGER,"
                                              "LEHDG INTEGER,"
                                              "LEELEV INTEGER,"
                                              "FLAGS INTEGER,"
                                              "HELIGHT0 INTEGER,"
                                              "HELIGHT1 INTEGER,"
                                              "HELIGHT2 INTEGER,"
                                              "HELIGHT3 INTEGER,"
                                              "HELIGHT4 INTEGER,"
                                              "HELIGHT5 INTEGER,"
                                              "HELIGHT6 INTEGER,"
                                              "HELIGHT7 INTEGER,"
                                              "LELIGHT0 INTEGER,"
                                              "LELIGHT1 INTEGER,"
                                              "LELIGHT2 INTEGER,"
                                              "LELIGHT3 INTEGER,"
                                              "LELIGHT4 INTEGER,"
                                              "LELIGHT5 INTEGER,"
                                              "LELIGHT6 INTEGER,"
                                              "LELIGHT7 INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airporthelipads (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "IDENT TEXT,"
                                              "LENGTH INTEGER,"
                                              "WIDTH INTEGER,"
                                              "SURFACE TEXT,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "HDG INTEGER,"
                                              "ELEV INTEGER,"
                                              "FLAGS INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airportcomms (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "NAME TEXT,"
                                              "SECTOR TEXT,"
                                              "OPRHOURS TEXT,"
                                              "FREQ0 INTEGER,"
                                              "FREQ1 INTEGER,"
                                              "FREQ2 INTEGER,"
                                              "FREQ3 INTEGER,"
                                              "FREQ4 INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airportvfrroutes (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "NAME TEXT);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airportvfrroutepoints (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "ROUTEID INTEGER NOT NULL,"
                                              "NAME TEXT,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "ALT INTEGER,"
                                              "ALTCODE CHAR,"
                                              "PATHCODE CHAR,"
                                              "SYM CHAR,"
                                              "ATAIRPORT CHAR,"
                                              "LABELPLACEMENT CHAR);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airportlinefeatures (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "NAME TEXT,"
                                              "SURFACE TEXT,"
                                              "FLAGS INTEGER,"
                                              "POLY BLOB);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airportpointfeatures (ID INTEGER NOT NULL, "
                                              "ARPTID INTEGER NOT NULL,"
                                              "FEATURE INTEGER,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "HDG INTEGER,"
                                              "ELEV INTEGER,"
                                              "SUBTYPE INTEGER,"
                                              "ATTR1 INTEGER,"
                                              "ATTR2 INTEGER,"
                                              "NAME TEXT,"
                                              "RWYIDENT TEXT);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airports_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Airport>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_icao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_bbox;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportrunways_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airporthelipads_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportcomms_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportvfrroutes_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportvfrroutepoints_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportvfrroutepoints_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportvfrroutepoints_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_lon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_swlat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_swlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_nelat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airports_nelon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportvfrroutepoints_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airportvfrroutepoints_lon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Airport>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_srcid ON airports(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_icao ON airports(ICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_name ON airports(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_latlon ON airports(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_bbox ON airports(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_modified ON airports(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airportrunways_id ON airportrunways(ARPTID,ID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airporthelipads_id ON airporthelipads(ARPTID,ID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airportcomms_id ON airportcomms(ARPTID,ID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airportvfrroutes_id ON airportvfrroutes(ARPTID,ID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airportvfrroutepoints_id ON airportvfrroutepoints(ARPTID,ROUTEID,ID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airportvfrroutepoints_name ON airportvfrroutepoints(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airportvfrroutepoints_latlon ON airportvfrroutepoints(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_tile ON airports(TILE);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_swlat ON airports(SWLAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_swlon ON airports(SWLON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_nelat ON airports(NELAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airports_nelon ON airports(NELON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Airport>::main_table_name = "airports";
template<> const char *DbBase<DbBaseElements::Airport>::database_file_name = "airports.db";
template<> const char *DbBase<DbBaseElements::Airport>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Airport>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Airport>::area_data = true;

DbBaseElements::Airspace::Segment::Segment(void)
        : m_coord1(), m_coord2(), m_coord0(), m_radius1(0), m_radius2(0), m_shape(0)
{
}

DbBaseElements::Airspace::Segment::Segment( const Point & coord1, const Point & coord2, const Point & coord0, char shape, int32_t r1, int32_t r2 )
        : m_coord1(coord1), m_coord2(coord2), m_coord0(coord0), m_radius1(r1), m_radius2(r2), m_shape(shape)
{
}

const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_n;
const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_e;
const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_s;
const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_w;
const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_center;
const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_any;
const DbBaseElements::Airspace::label_placement_t DbBaseElements::Airspace::label_off;

const uint8_t DbBaseElements::Airspace::altflag_agl;
const uint8_t DbBaseElements::Airspace::altflag_fl;
const uint8_t DbBaseElements::Airspace::altflag_sfc;
const uint8_t DbBaseElements::Airspace::altflag_gnd;
const uint8_t DbBaseElements::Airspace::altflag_unlim;
const uint8_t DbBaseElements::Airspace::altflag_notam;
const uint8_t DbBaseElements::Airspace::altflag_unkn;

const char *DbBaseElements::Airspace::db_query_string = "airspaces.ID AS 'ID',0 AS 'TABLE',airspaces.SRCID AS 'SRCID',airspaces.ICAO,airspaces.NAME,airspaces.IDENT,airspaces.CLASSEXCEPT,airspaces.EFFTIMES,airspaces.NOTE,airspaces.SWLAT,airspaces.SWLON,airspaces.NELAT,airspaces.NELON,airspaces.COMMNAME,airspaces.COMMFREQ0,airspaces.COMMFREQ1,airspaces.TYPECODE,airspaces.CLASS,airspaces.ALTLOWER,airspaces.ALTLOWERFLAGS,airspaces.ALTUPPER,airspaces.ALTUPPERFLAGS,airspaces.GNDELEVMIN,airspaces.GNDELEVMAX,airspaces.LABELLAT,airspaces.LABELLON,airspaces.LABELPLACEMENT,airspaces.POLY,airspaces.MODIFIED";
const char *DbBaseElements::Airspace::db_aux_query_string = "aux.airspaces.ID AS 'ID',1 AS 'TABLE',aux.airspaces.SRCID AS 'SRCID',aux.airspaces.ICAO,aux.airspaces.NAME,aux.airspaces.IDENT,aux.airspaces.CLASSEXCEPT,aux.airspaces.EFFTIMES,aux.airspaces.NOTE,aux.airspaces.SWLAT,aux.airspaces.SWLON,aux.airspaces.NELAT,aux.airspaces.NELON,aux.airspaces.COMMNAME,aux.airspaces.COMMFREQ0,aux.airspaces.COMMFREQ1,aux.airspaces.TYPECODE,aux.airspaces.CLASS,aux.airspaces.ALTLOWER,aux.airspaces.ALTLOWERFLAGS,aux.airspaces.ALTUPPER,aux.airspaces.ALTUPPERFLAGS,aux.airspaces.GNDELEVMIN,aux.airspaces.GNDELEVMAX,aux.airspaces.LABELLAT,aux.airspaces.LABELLON,aux.airspaces.LABELPLACEMENT,aux.airspaces.POLY,aux.airspaces.MODIFIED";
const char *DbBaseElements::Airspace::db_text_fields[] = { "SRCID", "ICAO", "NAME", "IDENT", "CLASSEXCEPT", "EFFTIMES", "NOTE", 0 };
const char *DbBaseElements::Airspace::db_time_fields[] = { "MODIFIED", 0 };;
const DbBaseElements::Airspace::gndelev_t DbBaseElements::Airspace::gndelev_unknown = std::numeric_limits<gndelev_t>::min();

DbBaseElements::Airspace::Airspace(void)
        : m_id(0), m_seg(), m_sourceid(), m_icao(), m_name(), m_classexcept(), m_efftime(), m_note(), m_commname(),
          m_bbox(), m_labelcoord(), m_label_placement(label_off), m_typecode(0), m_bdryclass(0),
          m_altlwr(0), m_altupr(0), m_altlwrflags(0), m_altuprflags(0), m_gndelevmin(gndelev_unknown), m_gndelevmax(gndelev_unknown), m_modtime(0),
          m_table(table_main), m_subtables(subtables_all)
{
        m_commfreq[0] = m_commfreq[1] = 0;
}

void DbBaseElements::Airspace::load_subtables(sqlite3x::sqlite3_connection & db, unsigned int loadsubtables)
{
        loadsubtables &= subtables_all & ~m_subtables;
        if (!loadsubtables)
                return;
        std::string tablepfx;
        switch (m_table) {
                case 0:
                        break;
                case 1:
                        tablepfx = "aux.";
                        break;
                default:
                        throw std::runtime_error("invalid table ID");
        }
        if (loadsubtables & subtables_segments) {
                m_seg.clear();
                sqlite3x::sqlite3_command cmd(db, "SELECT LAT1,LON1,LAT2,LON2,LAT0,LON0,SHAPE,RADIUS1,RADIUS2 FROM " + tablepfx + "airspacesegments WHERE BDRYID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        m_seg.push_back(Segment(Point(cursor.getint(1), cursor.getint(0)),
                                                Point(cursor.getint(3), cursor.getint(2)),
                                                Point(cursor.getint(5), cursor.getint(4)),
                                                cursor.getint(6), cursor.getint(7), cursor.getint(8)));
                }
                m_subtables |= subtables_segments;
        }
}

void DbBaseElements::Airspace::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_table = table_main;
        m_typecode = 0;
        m_bdryclass = 0;
        m_subtables = subtables_none;
        {
                if (!cursor.step())
                        return;
                m_id = cursor.getint(0);
                m_table = (table_t)cursor.getint(1);
                m_sourceid = cursor.getstring(2);
                m_icao = cursor.getstring(3);
                m_name = cursor.getstring(4);
                m_ident = cursor.getstring(5);
                m_classexcept = cursor.getstring(6);
                m_efftime = cursor.getstring(7);
                m_note = cursor.getstring(8);
                m_bbox = Rect(Point(cursor.getint(10), cursor.getint(9)), Point(cursor.getint(12), cursor.getint(11)));
                m_commname = cursor.getstring(13);
                m_commfreq[0] = cursor.getint(14);
                m_commfreq[1] = cursor.getint(15);
                m_typecode = cursor.getint(16);
                m_bdryclass = cursor.getint(17);
                m_altlwr = cursor.getint(18);
                m_altlwrflags = cursor.getint(19);
                m_altupr = cursor.getint(20);
                m_altuprflags = cursor.getint(21);
                m_gndelevmin = cursor.getint(22);
                m_gndelevmax = cursor.getint(23);
                m_labelcoord = Point(cursor.getint(25), cursor.getint(24));
                m_label_placement = (label_placement_t)cursor.getint(26);
                m_polygon.getblob(cursor, 27);
                m_modtime = cursor.getint(28);
                //std::cerr << "Airspace::load: srcid " << m_sourceid << " bdryclass " << (int)m_bdryclass << std::endl;
        }
        if (!loadsubtables)
                return;
        load_subtables(db, loadsubtables);
}

void DbBaseElements::Airspace::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airspaces SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airspaces_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Airspace::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO airspaces_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM airspaces;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airspaces (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airspaces_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_bbox.get_south());
                        cmd.bind(3, m_bbox.get_north());
                        cmd.bind(4, m_bbox.get_west());
                        cmd.bind(5, (long long int)m_bbox.get_east_unwrapped());
                        cmd.executenonquery();
                }
        }
        // Update Airspace Record
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airspaces SET SRCID=?,ICAO=?,NAME=?,IDENT=?,CLASSEXCEPT=?,EFFTIMES=?,NOTE=?,SWLAT=?,SWLON=?,NELAT=?,NELON=?,COMMNAME=?,COMMFREQ0=?,COMMFREQ1=?,TYPECODE=?,CLASS=?,ALTLOWER=?,ALTLOWERFLAGS=?,ALTUPPER=?,ALTUPPERFLAGS=?,GNDELEVMIN=?,GNDELEVMAX=?,LABELLAT=?,LABELLON=?,LABELPLACEMENT=?,POLY=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_icao);
                cmd.bind(3, m_name);
                cmd.bind(4, m_ident);
                cmd.bind(5, m_classexcept);
                cmd.bind(6, m_efftime);
                cmd.bind(7, m_note);
                cmd.bind(8, m_bbox.get_south());
                cmd.bind(9, m_bbox.get_west());
                cmd.bind(10, m_bbox.get_north());
                cmd.bind(11, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(12, m_commname);
                cmd.bind(13, (long long int)m_commfreq[0]);
                cmd.bind(14, (long long int)m_commfreq[1]);
                cmd.bind(15, m_typecode);
                cmd.bind(16, m_bdryclass);
                cmd.bind(17, m_altlwr);
                cmd.bind(18, m_altlwrflags);
                cmd.bind(19, m_altupr);
                cmd.bind(20, m_altuprflags);
                cmd.bind(21, m_gndelevmin);
                cmd.bind(22, m_gndelevmax);
                cmd.bind(23, m_labelcoord.get_lat());
                cmd.bind(24, m_labelcoord.get_lon());
                cmd.bind(25, m_label_placement);
                m_polygon.bindblob(cmd, 26);
                cmd.bind(27, (long long int)m_modtime);
                cmd.bind(28, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.bind(29, (long long int)m_id);
                cmd.executenonquery();
        }
        //std::cerr << "airspace: updated main, subtables " << m_subtables << " segments " <<  get_nrsegments()<< std::endl;
        if (m_subtables & subtables_segments) {
                // Delete existing Segments
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airspacesegments WHERE BDRYID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                // Write Segments
                for (unsigned int i = 0; i < get_nrsegments(); i++) {
                        const Segment& seg((*this)[i]);
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO airspacesegments (BDRYID,ID,LAT1,LON1,LAT2,LON2,LAT0,LON0,SHAPE,RADIUS1,RADIUS2) VALUES(?,?,?,?,?,?,?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (int)i);
                        cmd.bind(3, seg.get_coord1().get_lat());
                        cmd.bind(4, seg.get_coord1().get_lon());
                        cmd.bind(5, seg.get_coord2().get_lat());
                        cmd.bind(6, seg.get_coord2().get_lon());
                        cmd.bind(7, seg.get_coord0().get_lat());
                        cmd.bind(8, seg.get_coord0().get_lon());
                        cmd.bind(9, seg.get_shape());
                        cmd.bind(10, seg.get_radius1());
                        cmd.bind(11, seg.get_radius2());
                        cmd.executenonquery();
                        //std::cerr << "airspace: save segment " << i << " id " << m_id << std::endl;
                }
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE airspaces_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Airspace::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO airspaces_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airspaces WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airspacesegments WHERE BDRYID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM airspaces_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

Glib::ustring DbBaseElements::Airspace::get_icao_name(void) const
{
        Glib::ustring r(m_icao);
        if (!(m_icao.empty() || m_name.empty()))
                r += " ";
        return r + m_name;
}

int32_t DbBaseElements::Airspace::get_altlwr_corr(void) const
{
        uint8_t fl(get_altlwrflags());
        int32_t a(get_altlwr());
        int32_t g(get_gndelevmin());

        if (fl & (altflag_sfc | altflag_gnd))
                return g;
        if (fl & altflag_agl)
                return a + g;
        if (fl & altflag_unlim)
                return std::numeric_limits<int32_t>::min();
        return a;
}

int32_t DbBaseElements::Airspace::get_altupr_corr(void) const
{
        uint8_t fl(get_altuprflags());
        int32_t a(get_altupr());
        int32_t g(get_gndelevmax());

        if (fl & (altflag_sfc | altflag_gnd))
                return g;
        if (fl & altflag_agl)
                return a + g;
        if (fl & altflag_unlim)
                return std::numeric_limits<int32_t>::max();
        return a;
}

Glib::ustring DbBaseElements::Airspace::get_altlwr_string(void) const
{
        uint8_t fl(get_altlwrflags());
        int32_t a(get_altlwr());
        if (fl & altflag_sfc)
                return "SFC";
        if (fl & altflag_gnd)
                return "GND";
        if (fl & altflag_unlim)
                return "UNLIM";
        if (fl & altflag_notam)
                return "NOTAM";
        if (fl & altflag_unkn)
                return "U";
        if (fl & altflag_fl) {
                char buf[16];
                snprintf(buf, sizeof(buf), "FL%d", a / 100);
                return buf;
        }
        char buf[16];
        snprintf(buf, sizeof(buf), "%d'", a);
        if (fl & altflag_agl)
                return Glib::ustring(buf) + "AGL";
        return buf;
}

Glib::ustring DbBaseElements::Airspace::get_altupr_string(void) const
{
        uint8_t fl(get_altuprflags());
        int32_t a(get_altupr());
        if (fl & altflag_sfc)
                return "SFC";
        if (fl & altflag_gnd)
                return "GND";
        if (fl & altflag_unlim)
                return "UNLIM";
        if (fl & altflag_notam)
                return "NOTAM";
        if (fl & altflag_unkn)
                return "U";
        if (fl & altflag_fl) {
                char buf[16];
                snprintf(buf, sizeof(buf), "FL%d", a / 100);
                return buf;
        }
        char buf[16];
        snprintf(buf, sizeof(buf), "%d'", a);
        if (fl & altflag_agl)
                return Glib::ustring(buf) + "AGL";
        return buf;
}

Gdk::Color DbBaseElements::Airspace::get_color( char bclass, uint8_t tc )
{
        /*
         * airspace colors
         */
        Gdk::Color col;
        col.set_rgb(255 << 8, 128 << 8, 0 << 8);
        if (tc == 255)  /* special use airspace */
                return col;
        switch (bclass) {
        default:
                col.set_rgb(125 << 8, 42 << 8, 107 << 8);
                break;

        case 'B':
                col.set_rgb(167 << 8, 101 << 8, 160 << 8);
                break;

        case 'C':
                col.set_rgb(131 << 8, 104 << 8, 174 << 8);
                break;

        case 'D':
                col.set_rgb(98 << 8, 111 << 8, 212 << 8);
                break;

        case 'E':
        case 'F':
                col.set_rgb(107 << 8, 142 << 8, 109 << 8);
                break;

        case 'G':
                col.set_rgb(175 << 8, 197 << 8, 177 << 8);
                break;
        }
        return col;
}

bool DbBaseElements::Airspace::triangle_center( Point & ret, Point p0, Point p1, Point p2 )
{
        ret = p0;
        p1 -= p0;
        p2 -= p0;
        Point c0 = p1 + p2;
        Point c1 = -p1 + (p2 - p1);
        float xn0 = c0.get_lat();
        float yn0 = -c0.get_lon();
        float xn1 = c1.get_lat();
        float yn1 = -c1.get_lon();
        int32_t e1 = (int32_t)(p1.get_lon() * xn1 + p1.get_lat() * yn1);
        float t = xn0 * yn1 - yn0 * xn1;
        if (fabs(t) < 0.0000001 || !c0.get_lat())
                return false;
        float y = xn0 * e1 / t;
        float x = - y * yn0 / xn0;
        ret = p0 + Point((int32_t)x, (int32_t)y);
        return true;
}

float DbBaseElements::Airspace::label_metric( const Point & p0, const Point & p1 )
{
        float l = ((p0.get_lat() >> 1) + (p1.get_lat() >> 1)) * Point::to_rad;
        float l2 = l * l;
        float l4 = l2 * l2;
        float cosapprox = 1 - (1.0 / 2.0) * l2 + (1.0 / 24.0) * l4;
        Point pt(p0 - p1);
        float d = (pt.get_lat() * (float)pt.get_lat() + pt.get_lon() * (float)pt.get_lon() * cosapprox);
        d *= (Point::to_rad * 60.0 * Point::to_rad * 60.0);
        if (d < 0.001)
                d = 1000;
        else
                d = 1.0 / d;
        return d;
}

void DbBaseElements::Airspace::compute_initial_label_placement(void)
{
        if (get_nrsegments() == 1 && ((*this)[0].get_shape() == 'C' || (*this)[0].get_shape() == 'A')) {
                m_labelcoord = (*this)[0].get_coord0();
                return;
        }
        if (m_polygon.empty()) {
                m_labelcoord = Point();
                return;
        }
        recompute_lineapprox();
        Point pt(m_bbox.boxcenter());
        m_labelcoord = pt;
        if (m_polygon.windingnumber(m_labelcoord))
                return;
        float m = 2000;
        for (unsigned int i = 0; i < m_polygon.size(); i++) {
                for (unsigned int j = i + 2; j < m_polygon.size(); j++) {
                        if (!i && (j + 1) >= m_polygon.size())
                                continue;
                        Point pt1(m_polygon[i].halfway(m_polygon[j]));
                        float m1 = label_metric(m_polygon[i], m_polygon[j]);
                        if (m1 > m)
                                continue;
                        if (!m_polygon.windingnumber(pt1))
                                continue;
                        pt = pt1;
                        m = m1;
                }
        }
        if (m < 1000) {
                m_labelcoord = pt;
                return;
        }
        for (unsigned int i = 2; i < m_polygon.size(); i++) {
                Point pt1;
                if (!triangle_center(pt1, m_polygon[i-2], m_polygon[i-1], m_polygon[i]))
                        continue;
                if (!m_polygon.windingnumber(pt1)) {
                        std::cerr << "DbBaseElements::Airspace::compute_initial_label_placement: triangle_center: triangle error!!" << std::endl;
                        continue;
                }
                m_labelcoord = pt1;
                return;
        }
        m_labelcoord = m_polygon[0];
}

void DbBaseElements::Airspace::recompute_lineapprox_circle(void)
{
        uint32_t anginc = (uint32_t)(boundarycircleapprox * Point::from_deg);
        uint32_t ang1 = 0;
        uint32_t ang2 = -anginc;
        const Segment& seg((*this)[0]);
        float r1 = Point::to_rad * seg.get_radius1();
        for (; ang1 < ang2; ang1 += anginc) {
                CartesianVector3D v(1.0, r1 * cos(ang1 * Point::to_rad), r1 * sin(ang1 * Point::to_rad));
                v.rotate_xz(seg.get_coord0().get_lat() * Point::to_rad);
                v.rotate_xy(seg.get_coord0().get_lon() * Point::to_rad);
                m_polygon.push_back(((PolarVector3D)v).getcoord());
        }
}

void DbBaseElements::Airspace::recompute_lineapprox(void)
{
        m_polygon.clear();
        if (!get_nrsegments())
                return;
        if (get_nrsegments() == 1) {
                char shape((*this)[0].get_shape());
                if (shape == 'C' || shape == 'A') {
                        if (shape == 'C' || (*this)[0].get_radius1() > 100)
                                recompute_lineapprox_circle();
                        recompute_bbox();
                        return;
                }
        }
        for (unsigned int i = 0; i < get_nrsegments(); i++) {
                const Segment& seg((*this)[i]);
                unsigned int j = i - 1;
                if (!i)
                        j = get_nrsegments() - 1;
                const Segment& segp((*this)[j]);
                if (seg.get_shape() == 'C') {
                        std::cerr << "DbBaseElements::Airspace::recompute_lineapprox: " << get_name() << " Circle within path" << std::endl;
                        continue;
                }
                if (seg.get_shape() == 'A') {
                        std::cerr << "DbBaseElements::Airspace::recompute_lineapprox: " << get_name() << " Point within path" << std::endl;
                        continue;
                }
                if (abs(seg.get_coord1().get_lat() - segp.get_coord2().get_lat()) > 100 ||
                    abs(seg.get_coord1().get_lon() - segp.get_coord2().get_lon()) > 100) {
                        std::cerr << "DbBaseElements::Airspace::recompute_lineapprox: " << get_name() << " Segments not adjacent" << std::endl;
                        continue;
                }
                m_polygon.push_back(seg.get_coord1());
                if (seg.get_shape() != 'L' && seg.get_shape() != 'R')
                        continue;
                /* convert circle segment to polygon edges */
                CartesianVector3D v1 = PolarVector3D(seg.get_coord1().get_lat() * Point::to_rad, seg.get_coord1().get_lon() * Point::to_rad, 1.0);
                CartesianVector3D v2 = PolarVector3D(seg.get_coord2().get_lat() * Point::to_rad, seg.get_coord2().get_lon() * Point::to_rad, 1.0);
                /* rotate back to near 1/0/0 */
                v1.rotate_xy(seg.get_coord0().get_lon() * (-Point::to_rad));
                v1.rotate_xz(seg.get_coord0().get_lat() * (-Point::to_rad));
                v2.rotate_xy(seg.get_coord0().get_lon() * (-Point::to_rad));
                v2.rotate_xz(seg.get_coord0().get_lat() * (-Point::to_rad));
                uint32_t ang1 = (uint32_t)(atan2(v1.getz(), v1.gety()) * Point::from_rad);
                float r1 = sqrt(v1.gety() * v1.gety() + v1.getz() * v1.getz());
                uint32_t ang2 = (uint32_t)(atan2(v2.getz(), v2.gety()) * Point::from_rad);
                float r2 = sqrt(v2.gety() * v2.gety() + v2.getz() * v2.getz());
                uint32_t anginc = (uint32_t)(boundarycircleapprox * Point::from_deg);
                if (seg.get_shape() == 'R')
                        anginc = -anginc;
                for (; abs((int32_t)(ang1 - ang2)) >= abs((int32_t)(anginc));) {
                        ang1 += anginc;
                        CartesianVector3D v(1.0, r1 * cos(ang1 * Point::to_rad), r1 * sin(ang1 * Point::to_rad));
                        v.rotate_xz(seg.get_coord0().get_lat() * Point::to_rad);
                        v.rotate_xy(seg.get_coord0().get_lon() * Point::to_rad);
                        m_polygon.push_back(((PolarVector3D)v).getcoord());
                }
        }
        if (m_polygon.size() < 2) {
                std::cerr << "DbBaseElements::Airspace::recompute_lineapprox: " << get_name() << " less than 2 poly points" << std::endl;
                m_polygon.clear();
        }
        recompute_bbox();
}

void DbBaseElements::Airspace::recompute_bbox(void)
{
        if (m_polygon.empty()) {
                if (get_nrsegments() == 1 && (*this)[0].get_shape() == 'A') {
                        m_bbox = Rect((*this)[0].get_coord2(), (*this)[0].get_coord2());
                        return;
                }
                m_bbox = Rect();
                return;
        }
        m_bbox = m_polygon.get_bbox();
}

void DbBaseElements::Airspace::set_polygon(const Polygon & poly)
{
        m_subtables &= ~subtables_segments;
        m_polygon = poly;
        m_seg.clear();
        recompute_bbox();
}

void DbBaseElements::Airspace::compute_segments_from_poly(void)
{
        m_subtables |= subtables_segments;
        m_seg.clear();
        for (unsigned int i = 0, j = m_polygon.size() - 1; i < m_polygon.size(); j = i, i++)
                m_seg.push_back(Segment(m_polygon[j], m_polygon[i], Point(), '-', 0, 0));
}

template<> void DbBase<DbBaseElements::Airspace>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airspaces;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airspacesegments;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airspacepolypoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS airspaces_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Airspace>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airspaces (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "ICAO TEXT,"
                                              "NAME TEXT,"
                                              "IDENT TEXT,"
                                              "CLASSEXCEPT TEXT,"
                                              "EFFTIMES TEXT,"
                                              "NOTE TEXT,"
                                              "SWLAT INTEGER,"
                                              "SWLON INTEGER,"
                                              "NELAT INTEGER,"
                                              "NELON INTEGER,"
                                              "COMMNAME TEXT,"
                                              "COMMFREQ0 INTEGER,"
                                              "COMMFREQ1 INTEGER,"
                                              "TYPECODE INTEGER,"
                                              "CLASS CHAR,"
                                              "ALTLOWER INTEGER,"
                                              "ALTLOWERFLAGS INTEGER,"
                                              "ALTUPPER INTEGER,"
                                              "ALTUPPERFLAGS INTEGER,"
                                              "GNDELEVMIN INTEGER,"
                                              "GNDELEVMAX INTEGER,"
                                              "LABELLAT INTEGER,"
                                              "LABELLON INTEGER,"
                                              "LABELPLACEMENT CHAR,"
                                              "POLY BLOB,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airspacesegments (ID INTEGER NOT NULL, "
                                              "BDRYID INTEGER NOT NULL,"
                                              "LAT1 INTEGER,"
                                              "LON1 INTEGER,"
                                              "LAT2 INTEGER,"
                                              "LON2 INTEGER,"
                                              "LAT0 INTEGER,"
                                              "LON0 INTEGER,"
                                              "SHAPE CHAR,"
                                              "RADIUS1 INTEGER,"
                                              "RADIUS2 INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS airspaces_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Airspace>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_icao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_ident;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_bbox;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspacesegments_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_lon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_swlat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_swlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_nelat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspaces_nelon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspacepolypoints_id;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspacepolypoints_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspacepolypoints_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS airspacepolypoints_lon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Airspace>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_srcid ON airspaces(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_icao ON airspaces(ICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_name ON airspaces(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_ident ON airspaces(IDENT COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_latlon ON airspaces(LABELLAT,LABELLON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_bbox ON airspaces(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_modified ON airspaces(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_tile ON airspaces(TILE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspacesegments_id ON airspacesegments(BDRYID,ID);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_swlat ON airspaces(SWLAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_swlon ON airspaces(SWLON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_nelat ON airspaces(NELAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS airspaces_nelon ON airspaces(NELON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Airspace>::main_table_name = "airspaces";
template<> const char *DbBase<DbBaseElements::Airspace>::database_file_name = "airspaces.db";
template<> const char *DbBase<DbBaseElements::Airspace>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Airspace>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Airspace>::area_data = true;

Glib::ustring DbBaseElements::Airspace::get_class_string( char bc, uint8_t tc )
{
        if (tc != 255) {
                if (bc < 'A' || bc > 'G')
                        return Glib::ustring();
                return Glib::ustring(1, bc);
        }
        switch (bc) {
        case 'A':
                return "ALERT";

        case 'D':
                return "DANGER";

        case 'M':
                return "MIL OPS AREA";

        case 'P':
                return "PROHIBITED";

        case 'R':
                return "RESTRICTED";

        case 'T':
                return "TEMP RSVD";

        case 'W':
                return "WARNING";

        default:
                return Glib::ustring();
        }
}

Glib::ustring DbBaseElements::Airspace::get_type_string( uint8_t tc )
{
        switch (tc) {
                case 255:
                        return "Special Use Airspace";

                case 'A':
                        return "CIV";

                case 'B':
                        return "CIV/MIL";

                case 'C':
                        return "MIL";

                case 'D':
                default:
                        return "other";
        }
}

const char *DbBaseElements::Mapelement::db_query_string = "mapelements.ID AS 'ID',0 AS 'TABLE',mapelements.TYPECODE,mapelements.NAME,mapelements.LAT,mapelements.LON,mapelements.SWLAT,mapelements.SWLON,mapelements.NELAT,mapelements.NELON,mapelements.POLY";
const char *DbBaseElements::Mapelement::db_aux_query_string = "aux.mapelements.ID AS 'ID',1 AS 'TABLE',aux.mapelements.TYPECODE,aux.mapelements.NAME,aux.mapelements.LAT,aux.mapelements.LON,aux.mapelements.SWLAT,aux.mapelements.SWLON,aux.mapelements.NELAT,aux.mapelements.NELON,aux.mapelements.POLY";
const char *DbBaseElements::Mapelement::db_text_fields[] = { "NAME", 0 };
const char *DbBaseElements::Mapelement::db_time_fields[] = { 0 };

DbBaseElements::Mapelement::Mapelement(void)
        : m_id(0), m_bbox(), m_typecode(maptyp_invalid), m_table(table_main), m_name(), m_labelcoord(), m_polygon()
{
}

DbBaseElements::Mapelement::Mapelement(uint8_t tc, const PolygonHole & po, const Glib::ustring nm)
        : m_id(0), m_bbox(), m_typecode(tc), m_table(table_main), m_name(nm), m_labelcoord(), m_polygon(po)
{
        m_bbox = m_polygon.get_bbox();
        recompute_label_coord();
}

DbBaseElements::Mapelement::Mapelement(uint8_t tc, const PolygonHole & po, const Point & lblcoord, const Glib::ustring nm)
        : m_id(0), m_bbox(), m_typecode(tc), m_table(table_main), m_name(nm), m_labelcoord(lblcoord), m_polygon(po)
{
        m_bbox = m_polygon.get_bbox();
}

void DbBaseElements::Mapelement::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_table = table_main;
        m_typecode = maptyp_invalid;
        if (!cursor.step())
                return;
        m_id = cursor.getint(0);
        m_table = (table_t)cursor.getint(1);
        m_typecode = cursor.getint(2);
        m_name = cursor.getstring(3);
        m_labelcoord = Point(cursor.getint(5), cursor.getint(4));
        m_bbox = Rect(Point(cursor.getint(7), cursor.getint(6)), Point(cursor.getint(9), cursor.getint(8)));
        m_polygon.getblob(cursor, 10);
        recompute_bbox_label();
}

void DbBaseElements::Mapelement::update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        //std::cerr << "Start transaction" << std::endl;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE mapelements SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE mapelements_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        //std::cerr << "Commit transaction" << std::endl;
        tran.commit();
}

void DbBaseElements::Mapelement::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO mapelements_deleted (ID) VALUES (?);");
                cmd.bind(1, (long long int)m_id);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM mapelements;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO mapelements (ID) VALUES(?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO mapelements_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_bbox.get_south());
                        cmd.bind(3, m_bbox.get_north());
                        cmd.bind(4, m_bbox.get_west());
                        cmd.bind(5, (long long int)m_bbox.get_east_unwrapped());
                        cmd.executenonquery();
                }
        }
        // Update Mapelement Record
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE mapelements SET TYPECODE=?,NAME=?,LAT=?,LON=?,SWLAT=?,SWLON=?,NELAT=?,NELON=?,POLY=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_typecode);
                cmd.bind(2, m_name);
                cmd.bind(3, m_labelcoord.get_lat());
                cmd.bind(4, m_labelcoord.get_lon());
                cmd.bind(5, m_bbox.get_south());
                cmd.bind(6, m_bbox.get_west());
                cmd.bind(7, m_bbox.get_north());
                cmd.bind(8, (long long int)m_bbox.get_east_unwrapped());
                m_polygon.bindblob(cmd, 9);
                cmd.bind(10, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.bind(11, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE mapelements_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Mapelement::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO mapelements_deleted (ID) VALUES (?);");
                cmd.bind(1, (long long int)m_id);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM mapelements WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM mapelements_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

bool DbBaseElements::Mapelement::is_area(void) const
{
        switch (get_typecode()) {
                case maptyp_lake:
                case maptyp_lake_t:
                case maptyp_pack_ice:
                case maptyp_glacier:
                case maptyp_forest:
                case maptyp_city:
                case maptyp_border:
                        return true;

                default:
                        return false;
        }
}

Glib::ustring DbBaseElements::Mapelement::get_typename(uint8_t tc)
{
        switch (tc) {
                default:
                        return "Invalid";

                case maptyp_railway:
                        return "Railway";

                case maptyp_railway_d:
                        return "Railway (multitrack)";

                case maptyp_aerial_cable:
                        return "Aerial Cable";

                case maptyp_highway:
                        return "Highway";

                case maptyp_road:
                        return "Road";

                case maptyp_trail:
                        return "Trail";

                case maptyp_river:
                case maptyp_river_t:
                        return "River";

                case maptyp_lake:
                case maptyp_lake_t:
                        return "Lake";

                case maptyp_canal:
                        return "Canal";

                case maptyp_pack_ice:
                        return "Ice";

                case maptyp_glacier:
                        return "Glacier";

                case maptyp_forest:
                        return "Forest";

                case maptyp_city:
                        return "City";

                case maptyp_village:
                        return "Village";

                case maptyp_spot:
                        return "Spot";

                case maptyp_landmark:
                        return "Landmark";

                case maptyp_powerline:
                        return "Powerline";

                case maptyp_border:
                        return "Border";
        }
}

void DbBaseElements::Mapelement::recompute_bbox_label(void)
{
        m_bbox = m_polygon.get_bbox();
        recompute_label_coord();
}

int DbBaseElements::Mapelement::lon_center_poly(int32_t lat)
{
        const Polygon& poly(get_exterior());
        if (!poly.size())
                return -1;
        lat -= m_bbox.get_south();
        std::vector<int32_t> isect;
        for (unsigned int j = poly.size()-1, i = 0; i < poly.size(); j = i, i++) {
                Point p1 = poly[j] - m_bbox.get_southwest();
                Point p2 = poly[i] - m_bbox.get_southwest();
                Point d = p2 - p1;
                if (!d.get_lat())
                        continue;
                float v = p1.get_lon() * (float)d.get_lat() - p1.get_lat() * (float)d.get_lon();
                if (p1.get_lat() > p2.get_lat()) {
                        p2.set_lon(p2.get_lat());
                        p2.set_lat(p1.get_lat());
                        p1.set_lat(p2.get_lon());
                }
                if (lat < p1.get_lat() || lat >= p2.get_lat())
                        continue;
                isect.push_back((int32_t)((v + lat * (float)d.get_lon()) / (float)d.get_lat()));
        }
        if (isect.size() < 2)
                return -1;
        sort(isect.begin(), isect.end());
        unsigned int i = (isect.size() >> 1) & ~1;
        m_labelcoord = m_bbox.get_southwest() + Point((isect[i] + isect[i+1]) / 2, lat);
        return 0;
}

void DbBaseElements::Mapelement::recompute_label_coord(void)
{
        Point center(m_bbox.boxcenter());
        if (lon_center_poly(center.get_lat()))
                m_labelcoord = center;
}

template<> void DbBase<DbBaseElements::Mapelement>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS mapelements;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS mapelements_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Mapelement>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS mapelements (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "TYPECODE INTEGER,"
                                              "NAME TEXT,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "SWLAT INTEGER,"
                                              "SWLON INTEGER,"
                                              "NELAT INTEGER,"
                                              "NELON INTEGER,"
                                              "POLY BLOB,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS mapelements_deleted(ID INTEGER PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Mapelement>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_bbox;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_swlat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_swlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_nelat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS mapelements_nelon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Mapelement>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_name ON mapelements(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_latlon ON mapelements(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_bbox ON mapelements(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_tile ON mapelements(TILE);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_swlat ON mapelements(SWLAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_swlon ON mapelements(SWLON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_nelat ON mapelements(NELAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS mapelements_nelon ON mapelements(NELON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Mapelement>::main_table_name = "mapelements";
template<> const char *DbBase<DbBaseElements::Mapelement>::database_file_name = "mapelements.db";
template<> const char *DbBase<DbBaseElements::Mapelement>::order_field = "ID";
template<> const char *DbBase<DbBaseElements::Mapelement>::delete_field = "ID";
template<> const bool DbBase<DbBaseElements::Mapelement>::area_data = true;

const DbBaseElements::Track::TrackPoint::alt_t DbBaseElements::Track::TrackPoint::altinvalid = std::numeric_limits<alt_t>::min();

const char *DbBaseElements::Track::db_query_string = "tracks.ID AS 'ID',0 AS 'TABLE',tracks.SRCID AS 'SRCID',tracks.FROMICAO,tracks.FROMNAME,tracks.TOICAO,tracks.TONAME,tracks.NOTES,tracks.OFFBLOCKTIME,tracks.TAKEOFFTIME,tracks.LANDINGTIME,tracks.ONBLOCKTIME,tracks.SWLAT,tracks.SWLON,tracks.NELAT,tracks.NELON,tracks.POINTS,tracks.MODIFIED";
const char *DbBaseElements::Track::db_aux_query_string = "aux.tracks.ID AS 'ID',1 AS 'TABLE',aux.tracks.SRCID AS 'SRCID',aux.tracks.FROMICAO,aux.tracks.FROMNAME,aux.tracks.TOICAO,aux.tracks.TONAME,aux.tracks.NOTES,aux.tracks.OFFBLOCKTIME,aux.tracks.TAKEOFFTIME,aux.tracks.LANDINGTIME,aux.tracks.ONBLOCKTIME,aux.tracks.SWLAT,aux.tracks.SWLON,aux.tracks.NELAT,aux.tracks.NELON,aux.tracks.POINTS,aux.tracks.MODIFIED";
const char *DbBaseElements::Track::db_text_fields[] = { "SRCID", "FROMICAO", "FROMNAME", "TOICAO", "TONAME", 0 };
const char *DbBaseElements::Track::db_time_fields[] = { "MODIFIED", "OFFBLOCKTIME", "TAKEOFFTIME", "LANDINGTIME", "ONBLOCKTIME", 0 };

DbBaseElements::Track::TrackPoint::TrackPoint(void)
        : m_coord(), m_time(0), m_alt(altinvalid)
{
}

void DbBaseElements::Track::TrackPoint::read_blob(const uint8_t * data)
{
        if (!data) {
                m_coord = Point();
                m_time = 0;
                m_alt = altinvalid;
                return;
        }
        m_time = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
        m_time <<= 8;
        m_time |= data[14];
        m_coord = Point(data[4] | (data[5] << 8) | (data[6] << 16) | (data[7] << 24),
                        data[8] | (data[9] << 8) | (data[10] << 16) | (data[11] << 24));
        m_alt = data[12] | (data[13] << 8);
}

void DbBaseElements::Track::TrackPoint::write_blob(uint8_t * data) const
{
        if (!data)
                return;
        data[14] = m_time;
        data[0] = m_time >> 8;
        data[1] = m_time >> 16;
        data[2] = m_time >> 24;
        data[3] = m_time >> 32;
        data[4] = m_coord.get_lon();
        data[5] = m_coord.get_lon() >> 8;
        data[6] = m_coord.get_lon() >> 16;
        data[7] = m_coord.get_lon() >> 24;
        data[8] = m_coord.get_lat();
        data[9] = m_coord.get_lat() >> 8;
        data[10] = m_coord.get_lat() >> 16;
        data[11] = m_coord.get_lat() >> 24;
        data[12] = m_alt;
        data[13] = m_alt >> 8;
        data[15] = 0;
}

DbBaseElements::Track::Track(void)
        : m_id(0), m_sourceid(), m_fromicao(), m_fromname(), m_toicao(), m_toname(), m_notes(),
          m_offblocktime(0), m_takeofftime(0), m_landingtime(0), m_onblocktime(0), m_bbox(), m_trackpoints(), m_savedtrackpoints(0),
          m_format(format_invalid), m_table(table_main), m_modtime(0)
{
}

void DbBaseElements::Track::load(sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables)
{
        m_id = 0;
        m_table = table_main;
        m_sourceid.clear();
        m_fromicao.clear();
        m_fromname.clear();
        m_toicao.clear();
        m_toname.clear();
        m_notes.clear();
        m_offblocktime = 0;
        m_takeofftime = 0;
        m_landingtime = 0;
        m_onblocktime = 0;
        m_bbox = Rect();
        m_trackpoints.clear();
        m_savedtrackpoints = 0;
        m_format = format_invalid;
        m_modtime = 0;
        {
                if (!cursor.step())
                        return;
                m_format = format_empty;
                m_id = cursor.getint(0);
                m_table = (table_t)cursor.getint(1);
                m_sourceid = cursor.getstring(2);
                m_fromicao = cursor.getstring(3);
                m_fromname = cursor.getstring(4);
                m_toicao = cursor.getstring(5);
                m_toname = cursor.getstring(6);
                m_notes = cursor.getstring(7);
                m_offblocktime = cursor.getint(8);
                m_takeofftime = cursor.getint(9);
                m_landingtime = cursor.getint(10);
                m_onblocktime = cursor.getint(11);
                m_bbox = Rect(Point(cursor.getint(13), cursor.getint(12)), Point(cursor.getint(15), cursor.getint(14)));
                if (!cursor.isnull(16)) {
                        int blobsize(0);
                        void const *blob(cursor.getblob(16, blobsize));
                        z_stream d_stream;
                        memset(&d_stream, 0, sizeof(d_stream));
                        d_stream.zalloc = (alloc_func)0;
                        d_stream.zfree = (free_func)0;
                        d_stream.opaque = (voidpf)0;
                        d_stream.next_in  = (Bytef *)blob;
                        d_stream.avail_in = blobsize;
                        int err = inflateInit(&d_stream);
                        if (err != Z_OK) {
                                inflateEnd(&d_stream);
                                std::cerr << "zlib: inflateInit error " << err << std::endl;
                                return;
                        }
                        uint8_t data[256];
                        d_stream.next_out = data;
                        d_stream.avail_out = sizeof(data);
                        for (;;) {
                                err = inflate(&d_stream, Z_SYNC_FLUSH);
                                if (err != Z_OK && err != Z_STREAM_END) {
                                        std::cerr << "zlib: inflate error " << err << std::endl;
                                        return;
                                }
                                uint8_t *data2;
                                for (data2 = data; data2 + TrackPoint::blob_size_v1 <= d_stream.next_out; data2 += TrackPoint::blob_size_v1) {
                                        m_trackpoints.push_back(TrackPoint());
                                        m_trackpoints.back().read_blob(data2);
                                }
                                if (err == Z_STREAM_END)
                                        break;
                                if (data2 < d_stream.next_out) {
                                        unsigned int len(d_stream.next_out - data2);
                                        memmove(data, data2, len);
                                        d_stream.next_out = data + len;
                                        d_stream.avail_out = sizeof(data) - len;
                                        continue;
                                }
                                d_stream.next_out = data;
                                d_stream.avail_out = sizeof(data);
                        }
                        inflateEnd(&d_stream);
                        m_format = format_gzip;
                }
                m_modtime = cursor.getint(17);
        }
        std::string tablepfx;
        switch (m_table) {
                case 0:
                        break;
                case 1:
                        tablepfx = "aux.";
                        break;
                default:
                        throw std::runtime_error("invalid table ID");
        }
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT POINTS FROM " + tablepfx + "tracks_appendlog WHERE TRACKID=? ORDER BY ID;");
                cmd.bind(1, (long long int)m_id);
                sqlite3x::sqlite3_cursor cursor(cmd.executecursor());
                while (cursor.step()) {
                        int blobsize(0);
                        uint8_t const *blob((uint8_t const *)cursor.getblob(1, blobsize));
                        uint8_t const *data = blob;
                        for (; data + TrackPoint::blob_size_v1 <= blob + blobsize; data += TrackPoint::blob_size_v1) {
                                m_trackpoints.push_back(TrackPoint());
                                m_trackpoints.back().read_blob(data);
                        }
                }
        }
        m_savedtrackpoints = m_trackpoints.size();
        if (!loadsubtables)
                return;
        load_subtables(db);
}

void DbBaseElements::Track::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        recompute_bbox();
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO tracks_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM tracks;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO tracks (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO tracks_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_bbox.get_south());
                        cmd.bind(3, m_bbox.get_north());
                        cmd.bind(4, m_bbox.get_west());
                        cmd.bind(5, (long long int)m_bbox.get_east_unwrapped());
                        cmd.executenonquery();
                }
        }
        // Update Airspace Record
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE tracks SET SRCID=?,FROMICAO=?,FROMNAME=?,TOICAO=?,TONAME=?,NOTES=?,OFFBLOCKTIME=?,TAKEOFFTIME=?,LANDINGTIME=?,ONBLOCKTIME=?,SWLAT=?,SWLON=?,NELAT=?,NELON=?,POINTS=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_fromicao);
                cmd.bind(3, m_fromname);
                cmd.bind(4, m_toicao);
                cmd.bind(5, m_toname);
                cmd.bind(6, m_notes);
                cmd.bind(7, (long long int)m_offblocktime);
                cmd.bind(8, (long long int)m_takeofftime);
                cmd.bind(9, (long long int)m_landingtime);
                cmd.bind(10, (long long int)m_onblocktime);
                cmd.bind(11, m_bbox.get_south());
                cmd.bind(12, m_bbox.get_west());
                cmd.bind(13, m_bbox.get_north());
                cmd.bind(14, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(15);
                if (!m_trackpoints.empty()) {
                        uint8_t data2[m_trackpoints.size() * TrackPoint::blob_size_v1];
                        uint8_t *data3 = data2;
                        for (trackpoints_t::const_iterator ti(m_trackpoints.begin()), te(m_trackpoints.end()); ti != te; ti++, data3 += TrackPoint::blob_size_v1)
                                ti->write_blob(data3);
                        char data[m_trackpoints.size() * TrackPoint::blob_size_v1 + 64];
                        z_stream c_stream;
                        memset(&c_stream, 0, sizeof(c_stream));
                        c_stream.zalloc = (alloc_func)0;
                        c_stream.zfree = (free_func)0;
                        c_stream.opaque = (voidpf)0;
                        int err = deflateInit(&c_stream, Z_BEST_COMPRESSION);
                        if (err) {
                                std::cerr << "deflateInit error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        c_stream.next_out = (Bytef *)data;
                        c_stream.avail_out = sizeof(data);
                        c_stream.next_in = (Bytef *)data2;
                        c_stream.avail_in = sizeof(data2);
                        err = deflate(&c_stream, Z_FINISH);
                        if (err != Z_STREAM_END) {
                                std::cerr << "deflate error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        if (c_stream.avail_in != 0 || c_stream.total_in != sizeof(data2)) {
                                std::cerr << "deflate did not consume all input" << std::endl;
                                goto outerr2;
                        }
                        cmd.bind(15, data, c_stream.total_out);
                        std::cerr << "Track compression: input: " << (sizeof(data2)) << " output: " << c_stream.total_out << std::endl;
                        err = deflateEnd(&c_stream);
                        if (err) {
                                std::cerr << "deflateEnd error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        outerr2:;

                }
                cmd.bind(16, (long long int)m_modtime);
                cmd.bind(17, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.bind(18, (long long int)m_id);
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(db, "DELETE FROM tracks_appendlog WHERE TRACKID=?;");
                cmd.bind(1, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE tracks_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
        m_savedtrackpoints = m_trackpoints.size();
}

void DbBaseElements::Track::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        //std::cerr << "Start transaction" << std::endl;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE tracks SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(m_bbox));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE tracks_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_bbox.get_south());
                cmd.bind(2, m_bbox.get_north());
                cmd.bind(3, m_bbox.get_west());
                cmd.bind(4, (long long int)m_bbox.get_east_unwrapped());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        //std::cerr << "Commit transaction" << std::endl;
        tran.commit();
}

void DbBaseElements::Track::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO tracks_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM tracks WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM tracks_appendlog WHERE TRACKID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM tracks_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

Glib::ustring DbBaseElements::Track::get_fromicaoname(void) const
{
        Glib::ustring r(get_fromicao());
        if (!r.empty() && !get_fromname().empty())
                r += " ";
        r += get_fromname();
        return r;
}

Glib::ustring DbBaseElements::Track::get_toicaoname(void) const
{
        Glib::ustring r(get_toicao());
        if (!r.empty() && !get_toname().empty())
                r += " ";
        r += get_toname();
        return r;
}

void DbBaseElements::Track::save_appendlog(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        if (m_savedtrackpoints >= m_trackpoints.size())
                return;
        if (m_table != table_main || !m_id) {
                save(db, rtree, aux_rtree);
                return;
        }
        {
                int64_t appid = 1;
                sqlite3x::sqlite3_transaction tran(db);
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM tracks_appendlog WHERE TRACKID=?;");
                        cmd.bind(1, (long long int)m_id);
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                appid = cursor.getint(0) + 1;
                }
                size_type nrel = m_trackpoints.size() - m_savedtrackpoints;
                uint8_t data[nrel * TrackPoint::blob_size_v1];
                for (size_type i = 0; i < nrel; i++)
                        m_trackpoints[i + m_savedtrackpoints].write_blob(data + i * TrackPoint::blob_size_v1);
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO tracks_appendlog (TRACKID,ID,POINTS) VALUES (?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, (long long int)appid);
                        cmd.bind(3, data, nrel * TrackPoint::blob_size_v1);
                        cmd.executenonquery();
                }
        }
        m_savedtrackpoints = m_trackpoints.size();
}

void DbBaseElements::Track::recompute_bbox(void)
{
        if (!size()) {
                m_bbox = Rect();
                return;
        }
        Point pt(operator[](0).get_coord());
        Point sw(0, 0), ne(0, 0);
        for (unsigned int i = 1; i < size(); i++) {
                Point p(operator[](i).get_coord() - pt);
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        m_bbox = Rect(sw + pt, ne + pt);
}

DbBaseElements::Track::TrackPoint& DbBaseElements::Track::append(const TrackPoint & tp)
{
        m_trackpoints.push_back(tp);
        return m_trackpoints.back();
}

DbBaseElements::Track::TrackPoint& DbBaseElements::Track::insert(const TrackPoint & tp, int index)
{
        m_trackpoints.insert(m_trackpoints.begin() + index, tp);
        return m_trackpoints[index];
}

void DbBaseElements::Track::erase(int index)
{
        m_trackpoints.erase(m_trackpoints.begin() + index);
}

void DbBaseElements::Track::make_valid(void)
{
        if (is_valid())
                return;
        m_format = format_gzip;
}

template<> void DbBase<DbBaseElements::Track>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS tracks;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS tracks_appendlog;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS tracks_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Track>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS tracks (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "FROMICAO TEXT,"
                                              "FROMNAME TEXT,"
                                              "TOICAO TEXT,"
                                              "TONAME TEXT,"
                                              "NOTES TEXT,"
                                              "OFFBLOCKTIME INTEGER,"
                                              "TAKEOFFTIME INTEGER,"
                                              "LANDINGTIME INTEGER,"
                                              "ONBLOCKTIME INTEGER,"
                                              "SWLAT INTEGER,"
                                              "SWLON INTEGER,"
                                              "NELAT INTEGER,"
                                              "NELON INTEGER,"
                                              "POINTS BLOB,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS tracks_appendlog (ID INTEGER NOT NULL, "
                                              "TRACKID INTEGER NOT NULL,"
                                              "POINTS BLOB);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS tracks_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Track>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_fromname;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_fromicao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_toname;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_toicao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_bbox;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_tile;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS tracks_appendlog_id;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Track>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_srcid ON tracks(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_fromname ON tracks(FROMNAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_fromicao ON tracks(FROMICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_toname ON tracks(TONAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_toicao ON tracks(TOICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_bbox ON tracks(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_modified ON tracks(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_tile ON tracks(TILE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS tracks_appendlog_id ON tracks_appendlog(TRACKID,ID);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Track>::main_table_name = "tracks";
template<> const char *DbBase<DbBaseElements::Track>::database_file_name = "tracks.db";
template<> const char *DbBase<DbBaseElements::Track>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Track>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Track>::area_data = true;

void TracksDb::open(const Glib::ustring & path)
{
        DbBase<DbBaseElements::Track>::open(path);
        if (m_open != openstate_mainopen)
                return;
        clear_appendlogs();
}

void TracksDb::clear_appendlogs(void)
{
        if (m_open != openstate_mainopen)
                return;
        std::list<Address> addrs;
        {
                sqlite3x::sqlite3_command cmd(m_db, "SELECT DISTINCT TRACKID FROM tracks_appendlog;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (cursor.step())
                        addrs.push_back(Address(cursor.getint(0), element_t::table_main));
        }
        while (!addrs.empty()) {
                element_t e(operator()(addrs.front()));
                save(e);
                addrs.pop_front();
        }
}

const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_n;
const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_e;
const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_s;
const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_w;
const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_center;
const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_any;
const DbBaseElements::Label::label_placement_t DbBaseElements::Label::label_off;

const char *DbBaseElements::Label::db_query_string = "labels.ID AS 'ID',0 AS 'TABLE',labels.MUTABLE,labels.LAT,labels.LON,labels.LABELPLACEMENT,labels.METRIC,labels.SUBID,labels.SUBTABLE";
const char *DbBaseElements::Label::db_aux_query_string = "aux.labels.ID AS 'ID',1 AS 'TABLE',aux.labels.MUTABLE,aux.labels.LAT,aux.labels.LON,aux.labels.LABELPLACEMENT,aux.labels.METRIC,aux.labels.SUBID,aux.labels.SUBTABLE";
const char *DbBaseElements::Label::db_text_fields[] = { 0 };
const char *DbBaseElements::Label::db_time_fields[] = { 0 };

DbBaseElements::Label::Label(void)
        : m_id(0), m_table(table_main), m_mutable(false), m_coord(), m_label_placement(label_off), m_metric(std::numeric_limits<double>::quiet_NaN()), m_subid(0)
{
}

void DbBaseElements::Label::load( sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables )
{
        m_id = 0;
        m_table = table_main;
        m_label_placement = label_off;
        if (!cursor.step())
                return;
        m_id = cursor.getint(0);
        m_table = (table_t)cursor.getint(1);
        m_mutable = !!cursor.getint(2);
        m_coord = Point(cursor.getint(4), cursor.getint(3));
        m_label_placement = (label_placement_t)cursor.getint(5);
        m_metric = cursor.getdouble(6);
        m_subid = cursor.getint(7);
        m_subtable = (table_t)cursor.getint(8);
}

void DbBaseElements::Label::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE labels SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(get_coord()));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE labels_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Label::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO labels_deleted (ID) VALUES (?);");
                cmd.bind(1, (long long int)m_id);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM labels;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO labels (ID) VALUES(?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO labels_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_coord.get_lat());
                        cmd.bind(3, m_coord.get_lat());
                        cmd.bind(4, m_coord.get_lon());
                        cmd.bind(5, m_coord.get_lon());
                        cmd.executenonquery();
                }
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE labels SET MUTABLE=?,LAT=?,LON=?,LABELPLACEMENT=?,METRIC=?,SUBID=?,SUBTABLE=?,TILE=? WHERE ID=?;");
                cmd.bind(1, (int)m_mutable);
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_label_placement);
                cmd.bind(5, m_metric);
                cmd.bind(6, (long long int)m_subid);
                cmd.bind(7, (int)m_subtable);
                cmd.bind(8, (int)TileNumber::to_tilenumber(get_coord()));
                cmd.bind(9, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE labels_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Label::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO labels_deleted (ID) VALUES (?);");
                cmd.bind(1, (long long int)m_id);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM labels WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM labels_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

template<> void DbBase<DbBaseElements::Label>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS labels;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS labels_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Label>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS labels (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "MUTABLE INTEGER,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "LABELPLACEMENT INTEGER,"
                                              "METRIC REAL,"
                                              "SUBID INTEGER,"
                                              "SUBTABLE INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS labels_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Label>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS labels_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS labels_metric;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS labels_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS labels_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS labels_lon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Label>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS labels_latlon ON labels(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS labels_metric ON labels(MUTABLE,METRIC);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS labels_tile ON labels(TILE);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS labels_lat ON labels(LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS labels_lon ON labels(LON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Label>::main_table_name = "labels";
template<> const char *DbBase<DbBaseElements::Label>::database_file_name = "labels.db";
template<> const char *DbBase<DbBaseElements::Label>::order_field = "ID";
template<> const char *DbBase<DbBaseElements::Label>::delete_field = "ID";
template<> const bool DbBase<DbBaseElements::Label>::area_data = false;

LabelsDb::elementvector_t LabelsDb::find_mutable_by_metric(unsigned int limit, unsigned int offset, unsigned int loadsubtables)
{
        std::string qs("SELECT " + std::string(element_t::db_query_string) + " FROM " + main_table_name + " WHERE MUTABLE==1 ORDER BY METRIC DESC");
        if (limit)
                qs += " LIMIT ?";
        if (offset)
                qs += " OFFSET ?";
        qs += ";";
        //std::cerr << "sql: " << qs << std::endl;
        sqlite3x::sqlite3_command cmd(m_db, qs);
        int nr = 1;
        if (limit)
                cmd.bind(nr++, (long long int)limit);
        if (offset)
                cmd.bind(nr++, (long long int)offset);
        std::vector<element_t> ret;
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        for (;;) {
                ret.push_back(element_t());
                ret.back().load(cursor, m_db);
                if (ret.back().is_valid())
                        continue;
                ret.pop_back();
                break;
        }
        return ret;
}

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::pixel_index_t TopoDbN<resolution,tilesize>::tile_size;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::nodata = std::numeric_limits<elev_t>::min();

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::ocean = std::numeric_limits<elev_t>::min() + 1;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::TopoCoordinate::coord_t TopoDbN<resolution,tilesize>::TopoCoordinate::from_point;

template<int resolution, int tilesize>
const uint64_t TopoDbN<resolution,tilesize>::TopoCoordinate::to_point;

template<int resolution, int tilesize>
const unsigned int TopoDbN<resolution,tilesize>::tile_cache_size;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::tile_index_t TopoDbN<resolution,tilesize>::lon_tiles;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::tile_index_t TopoDbN<resolution,tilesize>::lat_tiles;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::tile_index_t TopoDbN<resolution,tilesize>::nr_tiles;

template<int resolution, int tilesize>
const uint16_t TopoDbN<resolution,tilesize>::TopoTileCoordinate::lon_tiles;

template<int resolution, int tilesize>
const uint16_t TopoDbN<resolution,tilesize>::TopoTileCoordinate::lat_tiles;

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoTileCoordinate::TopoTileCoordinate(const TopoCoordinate & tc)
{
        m_lon_tile = tc.get_lon() / tilesize;
        m_lat_tile = tc.get_lat() / tilesize;
        m_lon_offs = tc.get_lon() - m_lon_tile * tilesize;
        m_lat_offs = tc.get_lat() - m_lat_tile * tilesize;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_east(void)
{
        m_lon_tile++;
        if (m_lon_tile >= lon_tiles)
                m_lon_tile = 0;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_north(void)
{
        if (m_lat_tile < lat_tiles)
                m_lat_tile++;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_west(void)
{
        if (!m_lon_tile)
                m_lon_tile = lon_tiles;
        m_lon_tile--;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_south(void)
{
        if (m_lat_tile)
                m_lat_tile--;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoCoordinate::TopoCoordinate(const Point & pt)
{
        int64_t lat1 = pt.get_lat();
        lat1 += 0x40000000;
        uint64_t lon = (uint32_t)pt.get_lon();
        uint64_t lat = std::min(std::max(lat1, (int64_t)0), (int64_t)0x80000000);
        lon *= from_point;
        lat *= from_point;
        m_lon = lon >> 32;
        m_lat = lat >> 32;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoCoordinate::operator Point(void) const
{
        uint64_t lon(m_lon), lat(m_lat);
        lon *= to_point;
        lat *= to_point;
        lon += (((uint64_t)1) << 31);
        lat += (((uint64_t)1) << 31);
        Point pt(lon >> 32, (lat >> 32) - 0x40000000);
        return pt;
}

template<int resolution, int tilesize>
std::ostream & TopoDbN<resolution,tilesize>::TopoCoordinate::print(std::ostream & os)
{
        return os << '(' << get_lon() << ',' << get_lat() << ')';
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_east(void)
{
        m_lon++;
        if (m_lon == from_point)
                m_lon = 0;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_west(void)
{
        if (!m_lon)
                m_lon = from_point;
        m_lon--;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_north(void)
{
        if (m_lat < from_point/2)
                m_lat++;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_south(void)
{
        if (m_lat)
                m_lat--;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::TopoCoordinate::coord_t TopoDbN<resolution,tilesize>::TopoCoordinate::lon_dist(const TopoCoordinate & tc) const
{
        coord_t r = tc.get_lon() + from_point - get_lon();
        if (r >= from_point)
                r -= from_point;
        return r;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::TopoCoordinate::coord_t TopoDbN<resolution,tilesize>::TopoCoordinate::lat_dist(const TopoCoordinate & tc) const
{
        return tc.get_lat() - get_lat();
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::Tile(tile_index_t index)
        : m_index(index), m_minelev(nodata), m_maxelev(nodata), m_dirty(false), m_loaded(false), m_minmaxinvalid(false)
{
        m_elev = new elev_t[tilesize * tilesize];
        for (pixel_index_t idx = 0; idx < tilesize * tilesize; idx++)
                m_elev[idx] = nodata;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::Tile(const Tile & tile)
        : m_index(tile.m_index), m_minelev(tile.m_minelev), m_maxelev(tile.m_maxelev), m_dirty(tile.m_dirty), m_loaded(tile.m_loaded), m_minmaxinvalid(tile.m_minmaxinvalid)
{
        m_elev = new elev_t[tilesize * tilesize];
        for (pixel_index_t idx = 0; idx < tilesize * tilesize; idx++)
                m_elev[idx] = tile.m_elev[idx];
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::Tile(sqlite3x::sqlite3_connection & db, tile_index_t index)
        : m_index(index), m_minelev(nodata), m_maxelev(nodata), m_dirty(false), m_loaded(false), m_minmaxinvalid(false)
{
        m_elev = new elev_t[tilesize * tilesize];
        for (pixel_index_t idx = 0; idx < tilesize * tilesize; idx++)
                m_elev[idx] = nodata;
        std::string qs("SELECT MINELEV,MAXELEV,ELEV FROM topo WHERE ID=?1;");
        //std::cerr << "sql: " << qs << std::endl;
        sqlite3x::sqlite3_command cmd(db, qs);
        int nr = 1;
        cmd.bind(nr++, (int)m_index);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        if (!cursor.step())
                return;
        m_loaded = true;
        elev_t minelev = m_minelev = cursor.getint(0);
        elev_t maxelev = m_maxelev = cursor.getint(1);
        if (m_minelev == ocean)
                m_minelev = 0;
        if (m_maxelev == ocean)
                m_maxelev = 0;
        if (cursor.isnull(2)) {
                elev_t e(nodata);
                if (minelev == maxelev)
                        e = minelev;
                for (pixel_index_t i = 0; i < tilesize * tilesize; i++)
                        m_elev[i] = e;
        } else {
                int blobsize(0);
                void const *blob(cursor.getblob(2, blobsize));
                z_stream d_stream;
                memset(&d_stream, 0, sizeof(d_stream));
                d_stream.zalloc = (alloc_func)0;
                d_stream.zfree = (free_func)0;
                d_stream.opaque = (voidpf)0;
                d_stream.next_in  = (Bytef *)blob;
                d_stream.avail_in = blobsize;
                int err = inflateInit(&d_stream);
                if (err != Z_OK) {
                        inflateEnd(&d_stream);
                        std::cerr << "zlib: inflateInit error " << err << std::endl;
                        return;
                }
                uint8_t *data = new uint8_t[2 * tilesize * tilesize];
                d_stream.next_out = data;
                d_stream.avail_out = 2 * tilesize * tilesize;
                err = inflate(&d_stream, Z_FINISH);
                inflateEnd(&d_stream);
                if (err != Z_STREAM_END) {
                        delete[] data;
                        std::cerr << "zlib: inflate error " << err << std::endl;
                        return;
                }
                for (pixel_index_t i = 0; i < tilesize * tilesize; i++)
                        m_elev[i] = data[2*i] + (data[2*i+1] << 8);
                delete[] data;
        }
        //std::cerr << "TopoDb::find: pt " << pt << " blobsize " << blobsize << std::endl;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::~Tile(void)
{
        delete[] m_elev;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Tile::save(sqlite3x::sqlite3_connection & db)
{
        if (!m_dirty)
                return;
        if (m_minmaxinvalid)
                update_minmax();
        std::cerr << "Saving tile: ID " << m_index << " min: " << m_minelev << " max: " << m_maxelev << std::endl;
        sqlite3x::sqlite3_transaction tran(db);
        if (!m_loaded) {
                sqlite3x::sqlite3_command cmd(db, "INSERT INTO topo (ID) VALUES(?);");
                cmd.bind(1, (int)m_index);
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE topo SET MINELEV=?2,MAXELEV=?3,ELEV=?4 WHERE ID=?1;");
                cmd.bind(1, (int)m_index);
                if (m_minelev == 0 && m_maxelev == 0 && m_elev[0] == ocean) {
                        cmd.bind(2, (long long int)ocean);
                        cmd.bind(3, (long long int)ocean);
                } else {
                        cmd.bind(2, (long long int)m_minelev);
                        cmd.bind(3, (long long int)m_maxelev);
                }
                cmd.bind(4);
                if (m_minelev != m_maxelev) {
                        unsigned int els(tilesize * tilesize);
                        uint8_t edata[2*els];
                        for (unsigned int i = 0; i < els; i++) {
                                edata[2*i] = m_elev[i];
                                edata[2*i+1] = m_elev[i] >> 8;
                        }
                        uint8_t cdata[2*els+64];
                        z_stream c_stream;
                        memset(&c_stream, 0, sizeof(c_stream));
                        c_stream.zalloc = (alloc_func)0;
                        c_stream.zfree = (free_func)0;
                        c_stream.opaque = (voidpf)0;
                        int err = deflateInit(&c_stream, Z_BEST_COMPRESSION);
                        if (err) {
                                std::cerr << "deflateInit error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        c_stream.next_out = cdata;
                        c_stream.avail_out = 2*els+64;
                        c_stream.next_in = (Bytef *)edata;
                        c_stream.avail_in = 2*els;
                        err = deflate(&c_stream, Z_FINISH);
                        if (err != Z_STREAM_END) {
                                std::cerr << "deflate error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        if (c_stream.avail_in != 0 || c_stream.total_in != 2*els) {
                                std::cerr << "deflate did not consume all input" << std::endl;
                                goto outerr2;
                        }
                        cmd.bind(4, cdata, c_stream.total_out);
                        std::cerr << "Tile compression: input: " << (2*els) << " output: " << c_stream.total_out << std::endl;
                        err = deflateEnd(&c_stream);
                        if (err) {
                                std::cerr << "deflateEnd error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                outerr2:;
                }
                cmd.executenonquery();
        }
        tran.commit();
        m_loaded = true;
        m_dirty = false;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::Tile::get_elev(pixel_index_t index) const
{
        if (index < 0 || index >= tilesize * tilesize)
                return nodata;
        return m_elev[index];
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Tile::set_elev(pixel_index_t index, elev_t elev)
{
        if (index < 0 || index >= tilesize * tilesize)
                return;
        if (elev != m_elev[index])
                m_dirty = true;
        m_elev[index] = elev;
        m_minmaxinvalid = true;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Tile::update_minmax(void)
{
        int16_t mi(std::numeric_limits<elev_t>::max()), ma(std::numeric_limits<elev_t>::min());
        unsigned int els(tilesize * tilesize);
        for (unsigned int i = 0; i < els; i++) {
                elev_t e(m_elev[i]);
                if (e == nodata)
                        continue;
                if (e == ocean)
                        e = 0;
                mi = std::min(mi, e);
                ma = std::max(ma, e);
        }
        m_minelev = mi;
        m_maxelev = ma;
        m_minmaxinvalid = false;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoDbN()
{
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::open( const Glib::ustring & path, const Glib::ustring & dbname )
{
        Glib::ustring dbname1(path);
        if (dbname1.empty())
                dbname1 = FPlan::get_userdbpath();
        dbname1 = Glib::build_filename(dbname1, dbname);
        m_db.open(dbname1);
        create_tables();
        create_indices();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::take(sqlite3 * dbh)
{
        m_db.take(dbh);
        if (!m_db.db())
                return;
        create_tables();
        create_indices();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::close(void)
{
        cache_commit();
        m_db.close();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::purgedb(void)
{
        m_tile_cache.clear();
        drop_indices();
        drop_tables();
        create_tables();
        create_indices();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::sync_off(void)
{
        sqlite3x::sqlite3_command cmd(m_db, "PRAGMA synchronous = OFF;");
        cmd.executenonquery();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::analyze(void)
{
        cache_commit();
        sqlite3x::sqlite3_command cmd(m_db, "ANALYZE;");
        cmd.executenonquery();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::vacuum(void)
{
        cache_commit();
        sqlite3x::sqlite3_command cmd(m_db, "VACUUM;");
        cmd.executenonquery();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::interrupt(void)
{
        sqlite3_interrupt(m_db.db());
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS topo;"); cmd.executenonquery(); }
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS topo (ID INTEGER PRIMARY KEY, "
                                              "MINELEV INTEGER,"
                                              "MAXELEV INTEGER,"
                                              "ELEV BLOB);");
                cmd.executenonquery();
        }
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::drop_indices(void)
{
        //{ sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS topo_bbox;"); cmd.executenonquery(); }
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::create_indices(void)
{
        //{ sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS topo_bbox ON topo(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::Tile * TopoDbN<resolution,tilesize>::find(tile_index_t index)
{
        if (index >= nr_tiles)
                return 0;
        unsigned int nrctiles(0);
        for (typename tile_cache_t::iterator ti(m_tile_cache.begin()), te(m_tile_cache.end()); ti != te; ti++, nrctiles++) {
                if (ti->get_index() == index) {
                        // move *ti to the beginning of the list without copying it
                        m_tile_cache.splice(m_tile_cache.begin(), m_tile_cache, ti);
                        return &*ti;
                }
        }
        if (nrctiles >= tile_cache_size) {
                typename tile_cache_t::iterator ti(--m_tile_cache.end());
                ti->save(m_db);
                m_tile_cache.erase(ti);
        }
        typename tile_cache_t::iterator ti(m_tile_cache.insert(m_tile_cache.begin(), Tile(m_db, index)));
        return &*ti;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::cache_commit(void)
{
        for (typename tile_cache_t::iterator ti(m_tile_cache.begin()), te(m_tile_cache.end()); ti != te; ti++) {
                ti->save(m_db);
        }
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::get_elev(const TopoTileCoordinate & tc)
{
        Tile *t(find(tc.get_tile_index()));
        if (!t) {
                std::cerr << "TopoDbN: cannot find tile " << tc.get_tile_index() << std::endl;
                return nodata;
        }
        return t->get_elev(tc.get_pixel_index());
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::set_elev(const TopoTileCoordinate & tc, elev_t elev)
{
        Tile *t(find(tc.get_tile_index()));
        if (!t) {
                std::cerr << "TopoDbN: cannot find tile " << tc.get_tile_index() << std::endl;
                return;
        }
        t->set_elev(tc.get_pixel_index(), elev);
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_tile_minmax_elev(tile_index_t index)
{
        Tile *t(find(index));
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        if (!t) {
                std::cerr << "TopoDbN: cannot find tile " << index << std::endl;
                return ret;
        }
        ret.first = t->get_minelev();
        ret.second = t->get_maxelev();
        return ret;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_minmax_elev(const Rect & r)
{
        Point pdim(TopoCoordinate::get_pointsize());
        TopoTileCoordinate tmin(r.get_southwest()), tmax(r.get_northeast());
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        for (TopoTileCoordinate tc(tmin);;) {
                pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : tile_size - 1);
                pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : tile_size - 1);
                Rect rtile((Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmin, ymin),
                           (Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmax, ymax) + pdim);
                if (r.is_inside(rtile)) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        ret.first = std::min(ret.first, mm.first);
                        ret.second = std::max(ret.second, mm.second);
                } else {
                        for (pixel_index_t y(ymin); y <= ymax; y++) {
                                tc.set_lat_offs(y);
                                for (pixel_index_t x(xmin); x <= xmax; x++) {
                                        tc.set_lon_offs(x);
                                        elev_t e(get_elev(tc));
                                        if (e == ocean)
                                                e = 0;
                                        if (e != nodata) {
                                                ret.first = std::min(ret.first, e);
                                                ret.second = std::max(ret.second, e);
                                        }
                                }
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                tc.advance_tile_north();
        }
        if (ret.first == std::numeric_limits<elev_t>::max())
                ret.first = nodata;
        if (ret.second == std::numeric_limits<elev_t>::min())
                ret.second = nodata;
        return ret;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_minmax_elev(const Polygon & p)
{
        if (p.empty())
                return minmax_elev_t(nodata, nodata);
        Point pdim(TopoCoordinate::get_pointsize()), pdim2(pdim.get_lon()/2, pdim.get_lat()/2);
        Rect r(p.get_bbox());
        TopoTileCoordinate tmin(r.get_southwest()), tmax(r.get_northeast());
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        for (TopoTileCoordinate tc(tmin);;) {
                pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : tile_size - 1);
                pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : tile_size - 1);
                Rect rtile((Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmin, ymin),
                           (Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmax, ymax) + pdim);
                bool intersect(p.is_intersection(rtile));
                bool poly_inside(!intersect && rtile.is_inside(p[0]));
                bool tile_inside(!intersect && !poly_inside && p.windingnumber(rtile.get_southwest()));
                if (false) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        std::cerr << "get_minmax_elev: bbox " << r << " tile " << rtile << " intersect " << (intersect ? "yes" : "no")
                                  << " poly_inside " << (poly_inside ? "yes" : "no") << " tile_inside " << (tile_inside ? "yes" : "no")
                                  << " tile elev " << mm.first << " / " << mm.second << std::endl;
                }
                if (tile_inside) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        ret.first = std::min(ret.first, mm.first);
                        ret.second = std::max(ret.second, mm.second);
                } else if (poly_inside || intersect) {
                        for (pixel_index_t y(ymin); y <= ymax; y++) {
                                tc.set_lat_offs(y);
                                for (pixel_index_t x(xmin); x <= xmax; x++) {
                                        tc.set_lon_offs(x);
                                        Point pt2(pdim2 + (Point)(TopoCoordinate)tc);
                                        if (r.is_inside(pt2) && p.windingnumber(pt2)) {
                                                elev_t e(get_elev(tc));
                                                //std::cerr << "get_minmax_elev: " << e << " @ " << pt2 << std::endl;
                                                if (e == ocean)
                                                        e = 0;
                                                if (e != nodata) {
                                                        ret.first = std::min(ret.first, e);
                                                        ret.second = std::max(ret.second, e);
                                                }
                                        }
                                }
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                tc.advance_tile_north();
        }
        if (ret.first == std::numeric_limits<elev_t>::max())
                ret.first = nodata;
        if (ret.second == std::numeric_limits<elev_t>::min())
                ret.second = nodata;
        return ret;
}

template<int resolution, int tilesize>
std::ostream & operator <<(std::ostream & os, const typename TopoDbN< resolution, tilesize >::TopoCoordinate & tc)
{
        return tc.print(os);
}

void TopoDb30::open(const Glib::ustring & path)
{
        TopoDbNN::open(path, "topo30.db");
}

template<int resolution, int tilesize>
TopoDbN<resolution, tilesize>::RGB::operator Gdk::Color(void) const
{
        Gdk::Color col;
        col.set_rgb((m_col >> 8) & 0xff00, m_col & 0xff00, (m_col << 8) & 0xff00);
        return col;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution, tilesize>::RGB TopoDbN<resolution, tilesize>::color(elev_t x)
{
#include "gtopocolor.h"
        return RGB(0, 0, 255);
}

template<int resolution, int tilesize>
int TopoDbN<resolution, tilesize>::m_to_ft(elev_t elev)
{
        return (elev << 2) - elev + (elev >> 2) + (elev >> 5) - (elev >> 11);
}

template class TopoDbN<30,480>;
//const TopoDb30::elev_t TopoDb30::nodata = TopoDbNN::nodata;

